!> @file advec_ws.f90
!------------------------------------------------------------------------------!
! This file is part of the PALM model system.
!
! PALM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2019 Leibniz Universitaet Hannover
!------------------------------------------------------------------------------!
!
! Current revisions:
! ------------------
! 
! 
! Former revisions:
! -----------------
! $Id: advec_ws.f90 3870 2019-04-08 13:44:34Z knoop $
! Moving initialization of bcm specific flux arrays into bulk_cloud_model_mod
! 
! 3864 2019-04-05 09:01:56Z monakurppa
! Remove tailing white spaces
! 
! 3696 2019-01-24 16:37:35Z suehring
! Bugfix in degradation height
! 
! 3661 2019-01-08 18:22:50Z suehring
! - Minor bugfix in divergence correction (only has implications at 
!   downward-facing wall surfaces)
! - Remove setting of Neumann condition for horizontal velocity variances
! - Split loops for tendency calculation and divergence correction in order to
!   reduce bit queries
! - Introduce new parameter nzb_max_l to better control order degradation at 
!   non-cyclic boundaries
! 
! 3655 2019-01-07 16:51:22Z knoop
! OpenACC port for SPEC
! 
! 3589 2018-11-30 15:09:51Z suehring
! Move the control parameter "salsa" from salsa_mod to control_parameters 
! (M. Kurppa)
! 
! 3582 2018-11-29 19:16:36Z suehring
! - Computation of vertical fluxes separated from computation of horizontal 
!   fluxes. Loops are splitted in order to avoid indirect indexing and allow 
!   for better vectorization. 
! - Accelerate code by remove type-conversions of ibits
! - replace pointer definition in scalar routine by simple explicit definition
!   of the passed array
! 
! 3547 2018-11-21 13:21:24Z suehring
! variables documented
! 
! 3467 2018-10-30 19:05:21Z suehring
! Implementation of a new aerosol module salsa.
! 
! 3302 2018-10-03 02:39:40Z raasch
! Stokes drift velocity added in scalar advection
! 
! 3298 2018-10-02 12:21:11Z kanani
! Add formatted note from ketelsen about statistics for chemistry
!
! 3294 2018-10-01 02:37:10Z raasch
! ocean renamed ocean_mode
! 
! 3274 2018-09-24 15:42:55Z knoop
! Modularization of all bulk cloud physics code components
! 
! 3241 2018-09-12 15:02:00Z raasch
! unused variables removed
! 
! 3183 2018-07-27 14:25:55Z suehring
! Rename variables for boundary-condition flags and for offline nesting mode 
! 
! 3182 2018-07-27 13:36:03Z suehring
! Bugfix in calculation of left-sided fluxes for u-component in OpenMP case.
! 
! 2731 2018-01-09 17:44:02Z suehring
! Enable loop vectorization by splitting the k-loop
!
! 2718 2018-01-02 08:49:38Z maronga
! Corrected "Former revisions" section
! 
! 2696 2017-12-14 17:12:51Z kanani
! Change in file header (GPL part)
! Implement advection for TKE-dissipation in case of RANS-TKE-e closure (TG)
! Allocate advc_flags_1/2 within ws_init_flags instead of init_grid
! Change argument list for exchange_horiz_2d_int (MS)
! 
! 2329 2017-08-03 14:24:56Z knoop
! Bugfix concerning density in divergence correction close to buildings
! 
! 2292 2017-06-20 09:51:42Z schwenkel
! Implementation of new microphysic scheme: cloud_scheme = 'morrison' 
! includes two more prognostic equations for cloud drop concentration (nc)  
! and cloud water content (qc). 
! 
! 2233 2017-05-30 18:08:54Z suehring
! 
! 2232 2017-05-30 17:47:52Z suehring
! Rename wall_flags_0 and wall_flags_00 into advc_flags_1 and advc_flags_2, 
! respectively.
! Set advc_flags_1/2 on basis of wall_flags_0/00 instead of nzb_s/u/v/w_inner.
! Setting advc_flags_1/2 also for downward-facing walls
! 
! 2200 2017-04-11 11:37:51Z suehring
! monotonic_adjustment removed
! 
! 2118 2017-01-17 16:38:49Z raasch
! OpenACC version of subroutines removed
! 
! 2037 2016-10-26 11:15:40Z knoop
! Anelastic approximation implemented
! 
! 2000 2016-08-20 18:09:15Z knoop
! Forced header and separation lines into 80 columns
! 
! 1996 2016-08-18 11:42:29Z suehring
! Bugfix concerning calculation of turbulent of turbulent fluxes
! 
! 1960 2016-07-12 16:34:24Z suehring
! Separate humidity and passive scalar
! 
! 1942 2016-06-14 12:18:18Z suehring
! Initialization of flags for ws-scheme moved from init_grid.
! 
! 1873 2016-04-18 14:50:06Z maronga
! Module renamed (removed _mod)
! 
! 
! 1850 2016-04-08 13:29:27Z maronga
! Module renamed
!
! 
! 1822 2016-04-07 07:49:42Z hoffmann
! icloud_scheme removed, microphysics_seifert added
!
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1630 2015-08-26 16:57:23Z suehring
! 
! 
! 1629 2015-08-26 16:56:11Z suehring
! Bugfix concerning wall_flags at left and south PE boundaries
!
! 1581 2015-04-10 13:45:59Z suehring
! 
! 
! 1580 2015-04-10 13:43:49Z suehring
! Bugfix: statistical evaluation of scalar fluxes in case of monotonic limiter
!
! 1567 2015-03-10 17:57:55Z suehring
! Bugfixes in monotonic limiter.
!
! 2015-03-09 13:10:37Z heinze
! Bugfix: REAL constants provided with KIND-attribute in call of 
! intrinsic functions like MAX and MIN 
! 
! 1557 2015-03-05 16:43:04Z suehring
! Enable monotone advection for scalars using monotonic limiter
!
! 1374 2014-04-25 12:55:07Z raasch
! missing variables added to ONLY list
! 
! 1361 2014-04-16 15:17:48Z hoffmann
! accelerator and vector version for qr and nr added
! 
! 1353 2014-04-08 15:21:23Z heinze
! REAL constants provided with KIND-attribute,
! module kinds added
! some formatting adjustments
!
! 1322 2014-03-20 16:38:49Z raasch
! REAL constants defined as wp-kind
!
! 1320 2014-03-20 08:40:49Z raasch
! ONLY-attribute added to USE-statements,
! kind-parameters added to all INTEGER and REAL declaration statements,
! kinds are defined in new module kinds,
! old module precision_kind is removed,
! revision history before 2012 removed,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements
!
! 1257 2013-11-08 15:18:40Z raasch
! accelerator loop directives removed
!
! 1221 2013-09-10 08:59:13Z raasch
! wall_flags_00 introduced, which holds bits 32-...
!
! 1128 2013-04-12 06:19:32Z raasch
! loop index bounds in accelerator version replaced by i_left, i_right, j_south,
! j_north
!
! 1115 2013-03-26 18:16:16Z hoffmann
! calculation of qr and nr is restricted to precipitation
!
! 1053 2012-11-13 17:11:03Z hoffmann
! necessary expansions according to the two new prognostic equations (nr, qr) 
! of the two-moment cloud physics scheme:
! +flux_l_*, flux_s_*, diss_l_*, diss_s_*, sums_ws*s_ws_l 
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 1027 2012-10-15 17:18:39Z suehring
! Bugfix in calculation indices k_mm, k_pp in accelerator version
!
! 1019 2012-09-28 06:46:45Z raasch
! small change in comment lines
!
! 1015 2012-09-27 09:23:24Z raasch
! accelerator versions (*_acc) added
!
! 1010 2012-09-20 07:59:54Z raasch
! cpp switch __nopointer added for pointer free version
!
! 888 2012-04-20 15:03:46Z suehring
! Number of REAL( IBITS() calls with identical arguments is reduced.
!
! 862 2012-03-26 14:21:38Z suehring
! ws-scheme also work with topography in combination with vector version.
! ws-scheme also work with outflow boundaries in combination with
! vector version.
! Degradation of the applied order of scheme is now steered by multiplying with
! Integer advc_flags_1. 2nd order scheme, WS3 and WS5 are calculated on each
! grid point and mulitplied with the appropriate flag.
! 2nd order numerical dissipation term changed. Now the appropriate 2nd order
! term derived according to the 4th and 6th order terms is applied. It turns
! out that diss_2nd does not provide sufficient dissipation near walls.
! Therefore, the function diss_2nd is removed.
! Near walls a divergence correction is necessary to overcome numerical
! instabilities due to too less divergence reduction of the velocity field.
! boundary_flags and logicals steering the degradation are removed.
! Empty SUBROUTINE local_diss removed.
! Further formatting adjustments.
!
! 801 2012-01-10 17:30:36Z suehring
! Bugfix concerning OpenMP parallelization. Summation of sums_wsus_ws_l,
! sums_wsvs_ws_l, sums_us2_ws_l, sums_vs2_ws_l, sums_ws2_ws_l, sums_wspts_ws_l,
! sums_wsqs_ws_l, sums_wssas_ws_l is now thread-safe by adding an additional
! dimension.
!
! Initial revision
!
! 411 2009-12-11 12:31:43 Z suehring
!
! Description:
! ------------
!> Advection scheme for scalars and momentum using the flux formulation of
!> Wicker and Skamarock 5th order. Additionally the module contains of a 
!> routine using for initialisation and steering of the statical evaluation. 
!> The computation of turbulent fluxes takes place inside the advection
!> routines.
!> Near non-cyclic boundaries the order of the applied advection scheme is
!> degraded.
!> A divergence correction is applied. It is necessary for topography, since
!> the divergence is not sufficiently reduced, resulting in erroneous fluxes 
!> and could lead to numerical instabilities. 
!-----------------------------------------------------------------------------!
 MODULE advec_ws

    USE kinds

    IMPLICIT NONE

    REAL(wp) ::  adv_mom_1            !< 1/4 - constant used in 5th-order advection scheme for momentum advection (1st-order part)
    REAL(wp) ::  adv_mom_3            !< 1/24 - constant used in 5th-order advection scheme for momentum advection (3rd-order part)
    REAL(wp) ::  adv_mom_5            !< 1/120 - constant used in 5th-order advection scheme for momentum advection (5th-order part)
    REAL(wp) ::  adv_sca_1            !< 1/2 - constant used in 5th-order advection scheme for scalar advection (1st-order part)
    REAL(wp) ::  adv_sca_3            !< 1/12 - constant used in 5th-order advection scheme for scalar advection (3rd-order part)
    REAL(wp) ::  adv_sca_5            !< 1/60 - constant used in 5th-order advection scheme for scalar advection (5th-order part)

    PRIVATE
    PUBLIC   advec_s_ws, advec_u_ws, advec_v_ws, advec_w_ws, ws_init,          &
             ws_init_flags, ws_statistics

    INTERFACE ws_init
       MODULE PROCEDURE ws_init
    END INTERFACE ws_init

    INTERFACE ws_init_flags
       MODULE PROCEDURE ws_init_flags
    END INTERFACE ws_init_flags

    INTERFACE ws_statistics
       MODULE PROCEDURE ws_statistics
    END INTERFACE ws_statistics

    INTERFACE advec_s_ws
       MODULE PROCEDURE advec_s_ws
       MODULE PROCEDURE advec_s_ws_ij
    END INTERFACE advec_s_ws

    INTERFACE advec_u_ws
       MODULE PROCEDURE advec_u_ws
       MODULE PROCEDURE advec_u_ws_ij
    END INTERFACE advec_u_ws

    INTERFACE advec_v_ws
       MODULE PROCEDURE advec_v_ws
       MODULE PROCEDURE advec_v_ws_ij
    END INTERFACE advec_v_ws

    INTERFACE advec_w_ws
       MODULE PROCEDURE advec_w_ws
       MODULE PROCEDURE advec_w_ws_ij
    END INTERFACE advec_w_ws

 CONTAINS


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Initialization of WS-scheme
!------------------------------------------------------------------------------!
    SUBROUTINE ws_init

       USE arrays_3d,                                                          &
           ONLY:  diss_l_diss, diss_l_e, diss_l_pt,      &
                  diss_l_q, diss_l_s, diss_l_sa,         &
                  diss_l_u, diss_l_v, diss_l_w, flux_l_diss, flux_l_e,         & 
                  flux_l_pt, flux_l_q,      &
                  flux_l_s, flux_l_sa, flux_l_u, flux_l_v,          &
                  flux_l_w, diss_s_diss, diss_s_e,      &
                  diss_s_pt, diss_s_q, diss_s_s,         &
                  diss_s_sa, diss_s_u, diss_s_v,  diss_s_w, flux_s_diss,       &
                  flux_s_e, flux_s_pt, flux_s_q,         &
                  flux_s_s, flux_s_sa, flux_s_u,         &
                  flux_s_v, flux_s_w

       USE control_parameters,                                                 &
           ONLY:  humidity, loop_optimization, passive_scalar, ocean_mode,     &
                  rans_tke_e, ws_scheme_mom, ws_scheme_sca, salsa

       USE indices,                                                            &
           ONLY:  nyn, nys, nzb, nzt

       USE kinds

       USE pegrid

       USE salsa_util_mod,                                                     &
           ONLY:  sums_salsa_ws_l

       USE statistics,                                                         &
           ONLY:  sums_us2_ws_l, sums_vs2_ws_l, sums_ws2_ws_l,&
                  sums_wspts_ws_l,            &
                  sums_wsqs_ws_l, sums_wsss_ws_l,             &
                  sums_wssas_ws_l,  sums_wsss_ws_l, sums_wsus_ws_l,            &
                  sums_wsvs_ws_l


!
!--    Set the appropriate factors for scalar and momentum advection.
       adv_sca_5 = 1.0_wp /  60.0_wp
       adv_sca_3 = 1.0_wp /  12.0_wp
       adv_sca_1 = 1.0_wp /   2.0_wp
       adv_mom_5 = 1.0_wp / 120.0_wp
       adv_mom_3 = 1.0_wp /  24.0_wp
       adv_mom_1 = 1.0_wp /   4.0_wp
!         
!--    Arrays needed for statical evaluation of fluxes.
       IF ( ws_scheme_mom )  THEN

          ALLOCATE( sums_wsus_ws_l(nzb:nzt+1,0:threads_per_task-1),            &
                    sums_wsvs_ws_l(nzb:nzt+1,0:threads_per_task-1),            &
                    sums_us2_ws_l(nzb:nzt+1,0:threads_per_task-1),             &
                    sums_vs2_ws_l(nzb:nzt+1,0:threads_per_task-1),             &
                    sums_ws2_ws_l(nzb:nzt+1,0:threads_per_task-1) )

          sums_wsus_ws_l = 0.0_wp
          sums_wsvs_ws_l = 0.0_wp
          sums_us2_ws_l  = 0.0_wp
          sums_vs2_ws_l  = 0.0_wp
          sums_ws2_ws_l  = 0.0_wp

       ENDIF

       IF ( ws_scheme_sca )  THEN

          ALLOCATE( sums_wspts_ws_l(nzb:nzt+1,0:threads_per_task-1) )
          sums_wspts_ws_l = 0.0_wp

          IF ( humidity  )  THEN
             ALLOCATE( sums_wsqs_ws_l(nzb:nzt+1,0:threads_per_task-1) )
             sums_wsqs_ws_l = 0.0_wp
          ENDIF

          IF ( passive_scalar )  THEN
             ALLOCATE( sums_wsss_ws_l(nzb:nzt+1,0:threads_per_task-1) )
             sums_wsss_ws_l = 0.0_wp
          ENDIF

          IF ( ocean_mode )  THEN
             ALLOCATE( sums_wssas_ws_l(nzb:nzt+1,0:threads_per_task-1) )
             sums_wssas_ws_l = 0.0_wp
          ENDIF        

          IF ( salsa )  THEN
             ALLOCATE( sums_salsa_ws_l(nzb:nzt+1,0:threads_per_task-1) )
             sums_salsa_ws_l = 0.0_wp
          ENDIF

       ENDIF

!
!--    Arrays needed for reasons of speed optimization for cache version.
!--    For the vector version the buffer arrays are not necessary,
!--    because the the fluxes can swapped directly inside the loops of the 
!--    advection routines.
       IF ( loop_optimization /= 'vector' )  THEN

          IF ( ws_scheme_mom )  THEN

             ALLOCATE( flux_s_u(nzb+1:nzt,0:threads_per_task-1),               &
                       flux_s_v(nzb+1:nzt,0:threads_per_task-1),               &
                       flux_s_w(nzb+1:nzt,0:threads_per_task-1),               &
                       diss_s_u(nzb+1:nzt,0:threads_per_task-1),               &
                       diss_s_v(nzb+1:nzt,0:threads_per_task-1),               &
                       diss_s_w(nzb+1:nzt,0:threads_per_task-1) )
             ALLOCATE( flux_l_u(nzb+1:nzt,nys:nyn,0:threads_per_task-1),       &
                       flux_l_v(nzb+1:nzt,nys:nyn,0:threads_per_task-1),       &
                       flux_l_w(nzb+1:nzt,nys:nyn,0:threads_per_task-1),       &
                       diss_l_u(nzb+1:nzt,nys:nyn,0:threads_per_task-1),       &
                       diss_l_v(nzb+1:nzt,nys:nyn,0:threads_per_task-1),       &
                       diss_l_w(nzb+1:nzt,nys:nyn,0:threads_per_task-1) )

          ENDIF

          IF ( ws_scheme_sca )  THEN

             ALLOCATE( flux_s_pt(nzb+1:nzt,0:threads_per_task-1),              &
                       flux_s_e(nzb+1:nzt,0:threads_per_task-1),               &
                       diss_s_pt(nzb+1:nzt,0:threads_per_task-1),              &
                       diss_s_e(nzb+1:nzt,0:threads_per_task-1) ) 
             ALLOCATE( flux_l_pt(nzb+1:nzt,nys:nyn,0:threads_per_task-1),      &
                       flux_l_e(nzb+1:nzt,nys:nyn,0:threads_per_task-1),       &
                       diss_l_pt(nzb+1:nzt,nys:nyn,0:threads_per_task-1),      &
                       diss_l_e(nzb+1:nzt,nys:nyn,0:threads_per_task-1) )

             IF ( rans_tke_e )  THEN
                ALLOCATE( flux_s_diss(nzb+1:nzt,0:threads_per_task-1),         &
                          diss_s_diss(nzb+1:nzt,0:threads_per_task-1) )
                ALLOCATE( flux_l_diss(nzb+1:nzt,nys:nyn,0:threads_per_task-1), &
                          diss_l_diss(nzb+1:nzt,nys:nyn,0:threads_per_task-1) )
             ENDIF

             IF ( humidity )  THEN
                ALLOCATE( flux_s_q(nzb+1:nzt,0:threads_per_task-1),            &
                          diss_s_q(nzb+1:nzt,0:threads_per_task-1) )
                ALLOCATE( flux_l_q(nzb+1:nzt,nys:nyn,0:threads_per_task-1),    &
                          diss_l_q(nzb+1:nzt,nys:nyn,0:threads_per_task-1) )
             ENDIF

             IF ( passive_scalar )  THEN
                ALLOCATE( flux_s_s(nzb+1:nzt,0:threads_per_task-1),            &
                          diss_s_s(nzb+1:nzt,0:threads_per_task-1) )
                ALLOCATE( flux_l_s(nzb+1:nzt,nys:nyn,0:threads_per_task-1),    &
                          diss_l_s(nzb+1:nzt,nys:nyn,0:threads_per_task-1) )
             ENDIF

             IF ( ocean_mode )  THEN
                ALLOCATE( flux_s_sa(nzb+1:nzt,0:threads_per_task-1),           &
                          diss_s_sa(nzb+1:nzt,0:threads_per_task-1) )
                ALLOCATE( flux_l_sa(nzb+1:nzt,nys:nyn,0:threads_per_task-1),   &
                          diss_l_sa(nzb+1:nzt,nys:nyn,0:threads_per_task-1) )
             ENDIF

          ENDIF

       ENDIF

    END SUBROUTINE ws_init

!------------------------------------------------------------------------------!
! Description:
! ------------
!> Initialization of flags for WS-scheme used to degrade the order of the scheme
!> near walls.
!------------------------------------------------------------------------------!
    SUBROUTINE ws_init_flags

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s, momentum_advec, scalar_advec

       USE indices,                                                            &
           ONLY:  advc_flags_1, advc_flags_2, nbgp, nxl, nxlg, nxlu, nxr, nxrg,&
                  nyn, nyng, nys, nysg, nysv, nzb, nzt, wall_flags_0

       USE kinds

       IMPLICIT NONE

       INTEGER(iwp) ::  i     !< index variable along x
       INTEGER(iwp) ::  j     !< index variable along y
       INTEGER(iwp) ::  k     !< index variable along z
       INTEGER(iwp) ::  k_mm  !< dummy index along z
       INTEGER(iwp) ::  k_pp  !< dummy index along z
       INTEGER(iwp) ::  k_ppp !< dummy index along z
       
       LOGICAL      ::  flag_set !< steering variable for advection flags
   
       ALLOCATE( advc_flags_1(nzb:nzt+1,nysg:nyng,nxlg:nxrg),                  &
                 advc_flags_2(nzb:nzt+1,nysg:nyng,nxlg:nxrg) )
       advc_flags_1 = 0
       advc_flags_2 = 0
       IF ( scalar_advec == 'ws-scheme' )  THEN
!
!--       Set flags to steer the degradation of the advection scheme in advec_ws
!--       near topography, inflow- and outflow boundaries as well as bottom and
!--       top of model domain. advc_flags_1 remains zero for all non-prognostic
!--       grid points.
          DO  i = nxl, nxr
             DO  j = nys, nyn
                DO  k = nzb+1, nzt
!
!--                scalar - x-direction
!--                WS1 (0), WS3 (1), WS5 (2)
                   IF ( ( .NOT. BTEST(wall_flags_0(k,j,i+1),0)                 &
                    .OR.  .NOT. BTEST(wall_flags_0(k,j,i+2),0)                 &   
                    .OR.  .NOT. BTEST(wall_flags_0(k,j,i-1),0) )               &
                      .OR.  ( ( bc_dirichlet_l .OR. bc_radiation_l )           &
                            .AND.  i == nxl   )                                &
                      .OR.  ( ( bc_dirichlet_r .OR. bc_radiation_r )           &
                            .AND.  i == nxr   ) )                              &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 0 )
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k,j,i+3),0)             &
                       .AND.        BTEST(wall_flags_0(k,j,i+1),0)             &
                       .AND.        BTEST(wall_flags_0(k,j,i+2),0)             &
                       .AND.        BTEST(wall_flags_0(k,j,i-1),0)             &
                            )                       .OR.                       &
                            ( .NOT. BTEST(wall_flags_0(k,j,i-2),0)             &
                       .AND.        BTEST(wall_flags_0(k,j,i+1),0)             &
                       .AND.        BTEST(wall_flags_0(k,j,i+2),0)             &
                       .AND.        BTEST(wall_flags_0(k,j,i-1),0)             &
                            )                                                  &
                                                    .OR.                       &
                            ( ( bc_dirichlet_r .OR. bc_radiation_r )           &
                              .AND. i == nxr-1 )    .OR.                       &
                            ( ( bc_dirichlet_l .OR. bc_radiation_l )           &
                              .AND. i == nxlu  ) )                             & ! why not nxl+1
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 1 )
                   ELSEIF ( BTEST(wall_flags_0(k,j,i+1),0)                     &
                      .AND. BTEST(wall_flags_0(k,j,i+2),0)                     &
                      .AND. BTEST(wall_flags_0(k,j,i+3),0)                     &
                      .AND. BTEST(wall_flags_0(k,j,i-1),0)                     &
                      .AND. BTEST(wall_flags_0(k,j,i-2),0) )                   &
                   THEN 
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 2 )
                   ENDIF
!
!--                scalar - y-direction
!--                WS1 (3), WS3 (4), WS5 (5)
                   IF ( ( .NOT. BTEST(wall_flags_0(k,j+1,i),0)                 &
                    .OR.  .NOT. BTEST(wall_flags_0(k,j+2,i),0)                 &   
                    .OR.  .NOT. BTEST(wall_flags_0(k,j-1,i),0))                &
                      .OR.  ( ( bc_dirichlet_s .OR. bc_radiation_s )           &
                            .AND.  j == nys   )                                &
                      .OR.  ( ( bc_dirichlet_n .OR. bc_radiation_n )           &
                            .AND.  j == nyn   ) )                              &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 3 )
!
!--                WS3
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k,j+3,i),0)             &
                       .AND.        BTEST(wall_flags_0(k,j+1,i),0)             &
                       .AND.        BTEST(wall_flags_0(k,j+2,i),0)             &
                       .AND.        BTEST(wall_flags_0(k,j-1,i),0)             &
                            )                       .OR.                       &
                            ( .NOT. BTEST(wall_flags_0(k,j-2,i),0)             &
                       .AND.        BTEST(wall_flags_0(k,j+1,i),0)             &
                       .AND.        BTEST(wall_flags_0(k,j+2,i),0)             &
                       .AND.        BTEST(wall_flags_0(k,j-1,i),0)             &
                            )                                                  &
                                                    .OR.                       &
                            ( ( bc_dirichlet_s .OR. bc_radiation_s )           &
                              .AND. j == nysv  )    .OR.                       & ! why not nys+1
                            ( ( bc_dirichlet_n .OR. bc_radiation_n )           &
                              .AND. j == nyn-1 ) )                             &          
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 4 )
!
!--                WS5
                   ELSEIF ( BTEST(wall_flags_0(k,j+1,i),0)                     &
                      .AND. BTEST(wall_flags_0(k,j+2,i),0)                     &
                      .AND. BTEST(wall_flags_0(k,j+3,i),0)                     &
                      .AND. BTEST(wall_flags_0(k,j-1,i),0)                     &
                      .AND. BTEST(wall_flags_0(k,j-2,i),0) )                   &
                   THEN 
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 5 )
                   ENDIF
! 
!--                scalar - z-direction
!--                WS1 (6), WS3 (7), WS5 (8)
                   IF ( k == nzb+1 )  THEN
                      k_mm = nzb
                   ELSE 
                      k_mm = k - 2
                   ENDIF
                   IF ( k > nzt-1 )  THEN
                      k_pp = nzt+1
                   ELSE 
                      k_pp = k + 2
                   ENDIF
                   IF ( k > nzt-2 )  THEN
                      k_ppp = nzt+1
                   ELSE 
                      k_ppp = k + 3
                   ENDIF

                   flag_set = .FALSE.
                   IF ( .NOT. BTEST(wall_flags_0(k-1,j,i),0)  .AND.            &
                              BTEST(wall_flags_0(k,j,i),0)    .OR.             &
                        .NOT. BTEST(wall_flags_0(k_pp,j,i),0) .AND.            &                              
                              BTEST(wall_flags_0(k,j,i),0)    .OR.             &
                        k == nzt )                                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 6 )
                      flag_set = .TRUE.
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k_mm,j,i),0)    .OR.    &
                              .NOT. BTEST(wall_flags_0(k_ppp,j,i),0) ) .AND.   & 
                                  BTEST(wall_flags_0(k-1,j,i),0)  .AND.        &
                                  BTEST(wall_flags_0(k,j,i),0)    .AND.        &
                                  BTEST(wall_flags_0(k+1,j,i),0)  .AND.        &
                                  BTEST(wall_flags_0(k_pp,j,i),0) .OR.         &    
                            k == nzt - 1 )                                     &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 7 )
                      flag_set = .TRUE.
                   ELSEIF ( BTEST(wall_flags_0(k_mm,j,i),0)                    &
                     .AND.  BTEST(wall_flags_0(k-1,j,i),0)                     &
                     .AND.  BTEST(wall_flags_0(k,j,i),0)                       &
                     .AND.  BTEST(wall_flags_0(k+1,j,i),0)                     &
                     .AND.  BTEST(wall_flags_0(k_pp,j,i),0)                    &
                     .AND.  BTEST(wall_flags_0(k_ppp,j,i),0)                   &
                     .AND. .NOT. flag_set )                                    &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 8 )
                   ENDIF

                ENDDO
             ENDDO
          ENDDO
       ENDIF

       IF ( momentum_advec == 'ws-scheme' )  THEN
!
!--       Set advc_flags_1 to steer the degradation of the advection scheme in advec_ws
!--       near topography, inflow- and outflow boundaries as well as bottom and
!--       top of model domain. advc_flags_1 remains zero for all non-prognostic
!--       grid points.
          DO  i = nxl, nxr
             DO  j = nys, nyn
                DO  k = nzb+1, nzt
!  
!--                At first, set flags to WS1. 
!--                Since fluxes are swapped in advec_ws.f90, this is necessary to 
!--                in order to handle the left/south flux.
!--                near vertical walls. 
                   advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 9 )
                   advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 12 )
!
!--                u component - x-direction
!--                WS1 (9), WS3 (10), WS5 (11)
                   IF ( .NOT. BTEST(wall_flags_0(k,j,i+1),1)  .OR.             &
                            ( ( bc_dirichlet_l .OR. bc_radiation_l )           &
                              .AND. i <= nxlu  )    .OR.                       &
                            ( ( bc_dirichlet_r .OR. bc_radiation_r )           &
                              .AND. i == nxr   ) )                             &
                   THEN
                       advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 9 )
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k,j,i+2),1)  .AND.      &
                                    BTEST(wall_flags_0(k,j,i+1),1)  .OR.       &
                              .NOT. BTEST(wall_flags_0(k,j,i-1),1) )           &
                                                        .OR.                   &
                            ( ( bc_dirichlet_r .OR. bc_radiation_r )           &
                              .AND. i == nxr-1 )    .OR.                       &
                            ( ( bc_dirichlet_l .OR. bc_radiation_l )           &
                              .AND. i == nxlu+1) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 10 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 9 )
                   ELSEIF ( BTEST(wall_flags_0(k,j,i+1),1)  .AND.              &
                            BTEST(wall_flags_0(k,j,i+2),1)  .AND.              &
                            BTEST(wall_flags_0(k,j,i-1),1) )                   &
                   THEN    
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 11 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 9 )
                   ENDIF
!
!--                u component - y-direction
!--                WS1 (12), WS3 (13), WS5 (14)
                   IF ( .NOT. BTEST(wall_flags_0(k,j+1,i),1)   .OR.            &
                            ( ( bc_dirichlet_s .OR. bc_radiation_s )           &
                              .AND. j == nys   )    .OR.                       &
                            ( ( bc_dirichlet_n .OR. bc_radiation_n )           &
                              .AND. j == nyn   ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 12 )
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k,j+2,i),1)  .AND.      &
                                    BTEST(wall_flags_0(k,j+1,i),1)  .OR.       &
                              .NOT. BTEST(wall_flags_0(k,j-1,i),1) )           &
                                                        .OR.                   &
                            ( ( bc_dirichlet_s .OR. bc_radiation_s )           &
                              .AND. j == nysv  )    .OR.                       &
                            ( ( bc_dirichlet_n .OR. bc_radiation_n )           &
                              .AND. j == nyn-1 ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 13 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 12 )
                   ELSEIF ( BTEST(wall_flags_0(k,j+1,i),1)  .AND.              &
                            BTEST(wall_flags_0(k,j+2,i),1)  .AND.              &
                            BTEST(wall_flags_0(k,j-1,i),1) )                   &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 14 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 12 )
                   ENDIF
!
!--                u component - z-direction
!--                WS1 (15), WS3 (16), WS5 (17)
                   IF ( k == nzb+1 )  THEN
                      k_mm = nzb
                   ELSE 
                      k_mm = k - 2
                   ENDIF
                   IF ( k > nzt-1 )  THEN
                      k_pp = nzt+1
                   ELSE 
                      k_pp = k + 2
                   ENDIF
                   IF ( k > nzt-2 )  THEN
                      k_ppp = nzt+1
                   ELSE 
                      k_ppp = k + 3
                   ENDIF                   

                   flag_set = .FALSE.
                   IF ( .NOT. BTEST(wall_flags_0(k-1,j,i),1)  .AND.            &
                              BTEST(wall_flags_0(k,j,i),1)    .OR.             &
                        .NOT. BTEST(wall_flags_0(k_pp,j,i),1) .AND.            &                              
                              BTEST(wall_flags_0(k,j,i),1)    .OR.             &
                        k == nzt )                                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 15 )
                      flag_set = .TRUE.                      
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k_mm,j,i),1)    .OR.    &
                              .NOT. BTEST(wall_flags_0(k_ppp,j,i),1) ) .AND.   & 
                                  BTEST(wall_flags_0(k-1,j,i),1)  .AND.        &
                                  BTEST(wall_flags_0(k,j,i),1)    .AND.        &
                                  BTEST(wall_flags_0(k+1,j,i),1)  .AND.        &
                                  BTEST(wall_flags_0(k_pp,j,i),1) .OR.         &
                                  k == nzt - 1 )                               &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 16 )
                      flag_set = .TRUE.
                   ELSEIF ( BTEST(wall_flags_0(k_mm,j,i),1)  .AND.             &
                            BTEST(wall_flags_0(k-1,j,i),1)   .AND.             &
                            BTEST(wall_flags_0(k,j,i),1)     .AND.             &
                            BTEST(wall_flags_0(k+1,j,i),1)   .AND.             &
                            BTEST(wall_flags_0(k_pp,j,i),1)  .AND.             &
                            BTEST(wall_flags_0(k_ppp,j,i),1) .AND.             &
                            .NOT. flag_set )                                   &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 17 )
                   ENDIF

                ENDDO
             ENDDO
          ENDDO

          DO  i = nxl, nxr
             DO  j = nys, nyn
                DO  k = nzb+1, nzt
!
!--                At first, set flags to WS1. 
!--                Since fluxes are swapped in advec_ws.f90, this is necessary to 
!--                in order to handle the left/south flux.
                   advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 18 )
                   advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 21 )
!
!--                v component - x-direction
!--                WS1 (18), WS3 (19), WS5 (20)
                   IF ( .NOT. BTEST(wall_flags_0(k,j,i+1),2)  .OR.             &
                            ( ( bc_dirichlet_l .OR. bc_radiation_l )           &
                              .AND. i == nxl   )    .OR.                       &
                            ( ( bc_dirichlet_r .OR. bc_radiation_r )           &
                              .AND. i == nxr   ) )                             &
                  THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 18 )
!
!--                WS3
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k,j,i+2),2)   .AND.     &
                                    BTEST(wall_flags_0(k,j,i+1),2) ) .OR.      &
                              .NOT. BTEST(wall_flags_0(k,j,i-1),2)             &
                                                    .OR.                       &
                            ( ( bc_dirichlet_r .OR. bc_radiation_r )           &
                              .AND. i == nxr-1 )    .OR.                       &
                            ( ( bc_dirichlet_l .OR. bc_radiation_l )           &
                              .AND. i == nxlu  ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 19 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 18 )
                   ELSEIF ( BTEST(wall_flags_0(k,j,i+1),2)  .AND.              &
                            BTEST(wall_flags_0(k,j,i+2),2)  .AND.              &
                            BTEST(wall_flags_0(k,j,i-1),2) )                   &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 20 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 18 )
                   ENDIF
!
!--                v component - y-direction
!--                WS1 (21), WS3 (22), WS5 (23)
                   IF ( .NOT. BTEST(wall_flags_0(k,j+1,i),2) .OR.              &
                            ( ( bc_dirichlet_s .OR. bc_radiation_s )           &
                              .AND. j <= nysv  )    .OR.                       &
                            ( ( bc_dirichlet_n .OR. bc_radiation_n )           &
                              .AND. j == nyn   ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 21 )
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k,j+2,i),2)  .AND.      &
                                    BTEST(wall_flags_0(k,j+1,i),2)  .OR.       &
                              .NOT. BTEST(wall_flags_0(k,j-1,i),2) )           &
                                                        .OR.                   &
                            ( (  bc_dirichlet_s .OR. bc_radiation_s )          &
                              .AND. j == nysv+1)    .OR.                       &
                            ( (  bc_dirichlet_n .OR. bc_radiation_n )          &
                              .AND. j == nyn-1 ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 22 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 21 )
                   ELSEIF ( BTEST(wall_flags_0(k,j+1,i),2)  .AND.              &
                            BTEST(wall_flags_0(k,j+2,i),2)  .AND.              &
                            BTEST(wall_flags_0(k,j-1,i),2) )                   & 
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 23 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 21 )
                   ENDIF
!
!--                v component - z-direction
!--                WS1 (24), WS3 (25), WS5 (26)
                   IF ( k == nzb+1 )  THEN
                      k_mm = nzb
                   ELSE 
                      k_mm = k - 2
                   ENDIF
                   IF ( k > nzt-1 )  THEN
                      k_pp = nzt+1
                   ELSE 
                      k_pp = k + 2
                   ENDIF
                   IF ( k > nzt-2 )  THEN
                      k_ppp = nzt+1
                   ELSE 
                      k_ppp = k + 3
                   ENDIF  
                   
                   flag_set = .FALSE.
                   IF ( .NOT. BTEST(wall_flags_0(k-1,j,i),2)  .AND.            &
                              BTEST(wall_flags_0(k,j,i),2)    .OR.             &
                        .NOT. BTEST(wall_flags_0(k_pp,j,i),2) .AND.            &                              
                              BTEST(wall_flags_0(k,j,i),2)    .OR.             &
                        k == nzt )                                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 24 )
                      flag_set = .TRUE.
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k_mm,j,i),2)    .OR.    &
                              .NOT. BTEST(wall_flags_0(k_ppp,j,i),2) ) .AND.   & 
                                  BTEST(wall_flags_0(k-1,j,i),2)  .AND.        &
                                  BTEST(wall_flags_0(k,j,i),2)    .AND.        &
                                  BTEST(wall_flags_0(k+1,j,i),2)  .AND.        &
                                  BTEST(wall_flags_0(k_pp,j,i),2)  .OR.        &
                                  k == nzt - 1 )                               &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 25 )
                      flag_set = .TRUE.
                   ELSEIF ( BTEST(wall_flags_0(k_mm,j,i),2)  .AND.             &
                            BTEST(wall_flags_0(k-1,j,i),2)   .AND.             &
                            BTEST(wall_flags_0(k,j,i),2)     .AND.             &
                            BTEST(wall_flags_0(k+1,j,i),2)   .AND.             &
                            BTEST(wall_flags_0(k_pp,j,i),2)  .AND.             &
                            BTEST(wall_flags_0(k_ppp,j,i),2) .AND.             &
                            .NOT. flag_set )                                   &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 26 )
                   ENDIF

                ENDDO
             ENDDO
          ENDDO
          DO  i = nxl, nxr
             DO  j = nys, nyn
                DO  k = nzb+1, nzt
!
!--                At first, set flags to WS1. 
!--                Since fluxes are swapped in advec_ws.f90, this is necessary to 
!--                in order to handle the left/south flux.
                   advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 27 )
                   advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 30 )
!
!--                w component - x-direction
!--                WS1 (27), WS3 (28), WS5 (29)
                   IF ( .NOT. BTEST(wall_flags_0(k,j,i+1),3) .OR.              &
                            ( (  bc_dirichlet_l .OR. bc_radiation_l )          &
                              .AND. i == nxl   )    .OR.                       &
                            ( (  bc_dirichlet_r .OR. bc_radiation_r )          &
                              .AND. i == nxr   ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 27 )
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k,j,i+2),3)  .AND.      &
                                    BTEST(wall_flags_0(k,j,i+1),3)  .OR.       &
                              .NOT. BTEST(wall_flags_0(k,j,i-1),3) )           &
                                                        .OR.                   &
                            ( ( bc_dirichlet_r .OR. bc_radiation_r )           &
                              .AND. i == nxr-1 )    .OR.                       &
                            ( ( bc_dirichlet_l .OR.  bc_radiation_l )          &
                              .AND. i == nxlu  ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 28 )
!   
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 27 )
                   ELSEIF ( BTEST(wall_flags_0(k,j,i+1),3)  .AND.              &
                            BTEST(wall_flags_0(k,j,i+2),3)  .AND.              &
                            BTEST(wall_flags_0(k,j,i-1),3) )                   &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i),29 )
!   
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 27 )
                   ENDIF
!
!--                w component - y-direction
!--                WS1 (30), WS3 (31), WS5 (32)
                   IF ( .NOT. BTEST(wall_flags_0(k,j+1,i),3) .OR.              &
                            ( ( bc_dirichlet_s .OR. bc_radiation_s )           &
                              .AND. j == nys   )    .OR.                       &
                            ( ( bc_dirichlet_n .OR. bc_radiation_n )           &
                              .AND. j == nyn   ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 30 )
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k,j+2,i),3)  .AND.      &
                                    BTEST(wall_flags_0(k,j+1,i),3)  .OR.       &
                              .NOT. BTEST(wall_flags_0(k,j-1,i),3) )           &
                                                        .OR.                   &
                            ( ( bc_dirichlet_s .OR. bc_radiation_s )           &
                              .AND. j == nysv  )    .OR.                       &
                            ( ( bc_dirichlet_n .OR. bc_radiation_n )           &
                              .AND. j == nyn-1 ) )                             &
                   THEN
                      advc_flags_1(k,j,i) = IBSET( advc_flags_1(k,j,i), 31 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 30 )
                   ELSEIF ( BTEST(wall_flags_0(k,j+1,i),3)  .AND.              &
                            BTEST(wall_flags_0(k,j+2,i),3)  .AND.              &
                            BTEST(wall_flags_0(k,j-1,i),3) )                   &
                   THEN
                      advc_flags_2(k,j,i) = IBSET( advc_flags_2(k,j,i), 0 )
!
!--                   Clear flag for WS1
                      advc_flags_1(k,j,i) = IBCLR( advc_flags_1(k,j,i), 30 )
                   ENDIF
!
!--                w component - z-direction
!--                WS1 (33), WS3 (34), WS5 (35)
                   flag_set = .FALSE.
                   IF ( k == nzb+1 )  THEN
                      k_mm = nzb
                   ELSE
                      k_mm = k - 2
                   ENDIF
                   IF ( k > nzt-1 )  THEN
                      k_pp = nzt+1
                   ELSE 
                      k_pp = k + 2
                   ENDIF
                   IF ( k > nzt-2 )  THEN
                      k_ppp = nzt+1
                   ELSE 
                      k_ppp = k + 3
                   ENDIF  
                   
                   IF ( ( .NOT. BTEST(wall_flags_0(k-1,j,i),3)  .AND.          &
                          .NOT. BTEST(wall_flags_0(k,j,i),3)    .AND.          &
                                BTEST(wall_flags_0(k+1,j,i),3) )  .OR.         &
                        ( .NOT. BTEST(wall_flags_0(k-1,j,i),3)  .AND.          &
                                BTEST(wall_flags_0(k,j,i),3) )  .OR.           &
                        ( .NOT. BTEST(wall_flags_0(k+1,j,i),3)  .AND.          &
                                BTEST(wall_flags_0(k,j,i),3) )  .OR.           &        
                        k == nzt )                                             &
                   THEN
!
!--                   Please note, at k == nzb_w_inner(j,i) a flag is explictely
!--                   set, although this is not a prognostic level. However, 
!--                   contrary to the advection of u,v and s this is necessary
!--                   because flux_t(nzb_w_inner(j,i)) is used for the tendency
!--                   at k == nzb_w_inner(j,i)+1.
                      advc_flags_2(k,j,i) = IBSET( advc_flags_2(k,j,i), 1 )
                      flag_set = .TRUE.
                   ELSEIF ( ( .NOT. BTEST(wall_flags_0(k_mm,j,i),3)     .OR.   &
                              .NOT. BTEST(wall_flags_0(k_ppp,j,i),3) ) .AND.   &
                                    BTEST(wall_flags_0(k-1,j,i),3)  .AND.      &
                                    BTEST(wall_flags_0(k,j,i),3)    .AND.      &
                                    BTEST(wall_flags_0(k+1,j,i),3)  .OR.       &
                            k == nzt - 1 )                                     &
                   THEN
                      advc_flags_2(k,j,i) = IBSET( advc_flags_2(k,j,i), 2 )
                      flag_set = .TRUE.
                   ELSEIF ( BTEST(wall_flags_0(k_mm,j,i),3)  .AND.             &
                            BTEST(wall_flags_0(k-1,j,i),3)   .AND.             &
                            BTEST(wall_flags_0(k,j,i),3)     .AND.             &
                            BTEST(wall_flags_0(k+1,j,i),3)   .AND.             &
                            BTEST(wall_flags_0(k_pp,j,i),3)  .AND.             &
                            BTEST(wall_flags_0(k_ppp,j,i),3) .AND.             &
                            .NOT. flag_set )                                   &
                   THEN
                      advc_flags_2(k,j,i) = IBSET( advc_flags_2(k,j,i), 3 )
                   ENDIF

                ENDDO
             ENDDO
          ENDDO

       ENDIF


!
!--    Exchange 3D integer wall_flags.
       IF ( momentum_advec == 'ws-scheme' .OR. scalar_advec == 'ws-scheme'     &
          )  THEN  
!
!--       Exchange ghost points for advection flags
          CALL exchange_horiz_int( advc_flags_1, nys, nyn, nxl, nxr, nzt, nbgp )
          CALL exchange_horiz_int( advc_flags_2, nys, nyn, nxl, nxr, nzt, nbgp )
!
!--       Set boundary flags at inflow and outflow boundary in case of 
!--       non-cyclic boundary conditions.
          IF ( bc_dirichlet_l  .OR.  bc_radiation_l )  THEN
             advc_flags_1(:,:,nxl-1) = advc_flags_1(:,:,nxl)
             advc_flags_2(:,:,nxl-1) = advc_flags_2(:,:,nxl)
          ENDIF

          IF ( bc_dirichlet_r  .OR.  bc_radiation_r )  THEN
            advc_flags_1(:,:,nxr+1) = advc_flags_1(:,:,nxr)
            advc_flags_2(:,:,nxr+1) = advc_flags_2(:,:,nxr)
          ENDIF

          IF ( bc_dirichlet_n  .OR.  bc_radiation_n )  THEN
             advc_flags_1(:,nyn+1,:) = advc_flags_1(:,nyn,:)
             advc_flags_2(:,nyn+1,:) = advc_flags_2(:,nyn,:)
          ENDIF

          IF ( bc_dirichlet_s  .OR.  bc_radiation_s )  THEN
             advc_flags_1(:,nys-1,:) = advc_flags_1(:,nys,:)
             advc_flags_2(:,nys-1,:) = advc_flags_2(:,nys,:)
          ENDIF
  
       ENDIF


    END SUBROUTINE ws_init_flags


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Initialize variables used for storing statistic quantities (fluxes, variances)
!------------------------------------------------------------------------------!
    SUBROUTINE ws_statistics
    
       USE control_parameters,                                                 &
           ONLY:  humidity, passive_scalar, ocean_mode, ws_scheme_mom,         &
                  ws_scheme_sca, salsa

       USE kinds

       USE salsa_util_mod,                                                     &
           ONLY:  sums_salsa_ws_l

       USE statistics,                                                         &
           ONLY:  sums_us2_ws_l, sums_vs2_ws_l, sums_ws2_ws_l, &
                  sums_wspts_ws_l,           &
                  sums_wsqs_ws_l, sums_wsss_ws_l,             &
                  sums_wssas_ws_l, sums_wsus_ws_l, sums_wsvs_ws_l
                   

       IMPLICIT NONE

!       
!--    The arrays needed for statistical evaluation are set to to 0 at the 
!--    beginning of prognostic_equations.
       IF ( ws_scheme_mom )  THEN
          !$ACC KERNELS PRESENT(sums_wsus_ws_l, sums_wsvs_ws_l) &
          !$ACC PRESENT(sums_us2_ws_l, sums_vs2_ws_l, sums_ws2_ws_l)
          sums_wsus_ws_l = 0.0_wp
          sums_wsvs_ws_l = 0.0_wp
          sums_us2_ws_l  = 0.0_wp
          sums_vs2_ws_l  = 0.0_wp
          sums_ws2_ws_l  = 0.0_wp
          !$ACC END KERNELS
       ENDIF

       IF ( ws_scheme_sca )  THEN
          !$ACC KERNELS PRESENT(sums_wspts_ws_l)
          sums_wspts_ws_l = 0.0_wp
          !$ACC END KERNELS
          IF ( humidity       )  sums_wsqs_ws_l = 0.0_wp
          IF ( passive_scalar )  sums_wsss_ws_l = 0.0_wp
          IF ( ocean_mode )  sums_wssas_ws_l = 0.0_wp
          IF ( salsa )       sums_salsa_ws_l = 0.0_wp

       ENDIF

    END SUBROUTINE ws_statistics


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Scalar advection - Call for grid point i,j
!------------------------------------------------------------------------------!
    SUBROUTINE advec_s_ws_ij( i, j, sk, sk_char, swap_flux_y_local,            &
                              swap_diss_y_local, swap_flux_x_local,            &
                              swap_diss_x_local, i_omp, tn )

       USE arrays_3d,                                                          &
           ONLY:  ddzw, drho_air, rho_air_zw, tend, u, u_stokes_zu, v,         &
                  v_stokes_zu, w

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s, intermediate_timestep_count, &
                  u_gtrans, v_gtrans 

       USE grid_variables,                                                     &
           ONLY:  ddx, ddy

       USE indices,                                                            &
           ONLY:  nxl, nxlg, nxr, nxrg, nyn, nyng, nys, nysg, nzb, nzb_max,    &
                  nzt, advc_flags_1

       USE kinds

       USE pegrid

       USE salsa_util_mod,                                                     &
           ONLY:  sums_salsa_ws_l

       USE statistics,                                                         &
           ONLY:  hom, sums_wsncs_ws_l, sums_wsnrs_ws_l, sums_wspts_ws_l,      &
                  sums_wsqcs_ws_l,  sums_wsqrs_ws_l, sums_wsqs_ws_l,           &
                  sums_wssas_ws_l, sums_wsss_ws_l, weight_substep
                  

       IMPLICIT NONE

       CHARACTER (LEN = *), INTENT(IN) ::  sk_char !< string identifier, used for assign fluxes to the correct dimension in the analysis array 
       
       INTEGER(iwp) ::  i         !< grid index along x-direction
       INTEGER(iwp) ::  i_omp     !< leftmost index on subdomain, or in case of OpenMP, on thread
       INTEGER(iwp) ::  j         !< grid index along y-direction
       INTEGER(iwp) ::  k         !< grid index along z-direction
       INTEGER(iwp) ::  k_mm      !< k-2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_pp      !< k+2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_ppp     !< k+3 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  nzb_max_l !< index indicating upper bound for order degradation of horizontal advection terms  
       INTEGER(iwp) ::  tn        !< number of OpenMP thread
       
       REAL(wp)     ::  ibit0  !< flag indicating 1st-order scheme along x-direction
       REAL(wp)     ::  ibit1  !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)     ::  ibit2  !< flag indicating 5th-order scheme along x-direction
       REAL(wp)     ::  ibit3  !< flag indicating 1st-order scheme along y-direction
       REAL(wp)     ::  ibit4  !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)     ::  ibit5  !< flag indicating 5th-order scheme along y-direction
       REAL(wp)     ::  ibit6  !< flag indicating 1st-order scheme along z-direction
       REAL(wp)     ::  ibit7  !< flag indicating 3rd-order scheme along z-direction
       REAL(wp)     ::  ibit8  !< flag indicating 5th-order scheme along z-direction
       REAL(wp)     ::  diss_d !< artificial dissipation term at grid box bottom
       REAL(wp)     ::  div    !< diverence on scalar grid
       REAL(wp)     ::  flux_d !< 6th-order flux at grid box bottom
       REAL(wp)     ::  u_comp !< advection velocity along x-direction 
       REAL(wp)     ::  v_comp !< advection velocity along y-direction 
!       
!--    sk is an array from parameter list. It should not be a pointer, because 
!--    in that case the compiler can not assume a stride 1 and cannot perform 
!--    a strided one vector load. Adding the CONTIGUOUS keyword makes things 
!--    even worse, because the compiler cannot assume strided one in the 
!--    caller side.
       REAL(wp), INTENT(IN),DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg) ::  sk !<  advected scalar

       REAL(wp), DIMENSION(nzb:nzt+1)         ::  diss_n !< discretized artificial dissipation at northward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)         ::  diss_r !< discretized artificial dissipation at rightward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)         ::  diss_t !< discretized artificial dissipation at top of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)         ::  flux_n !< discretized 6th-order flux at northward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)         ::  flux_r !< discretized 6th-order flux at rightward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)         ::  flux_t !< discretized 6th-order flux at top of the grid box
       
       REAL(wp), DIMENSION(nzb+1:nzt,0:threads_per_task-1) ::  swap_diss_y_local !< discretized artificial dissipation at southward-side of the grid box 
       REAL(wp), DIMENSION(nzb+1:nzt,0:threads_per_task-1) ::  swap_flux_y_local !< discretized 6th-order flux at northward-side of the grid box
       
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn,0:threads_per_task-1) ::  swap_diss_x_local !< discretized artificial dissipation at leftward-side of the grid box 
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn,0:threads_per_task-1) ::  swap_flux_x_local !< discretized 6th-order flux at leftward-side of the grid box
!
!--    Used local modified copy of nzb_max (used to degrade order of 
!--    discretization) at non-cyclic boundaries. Modify only at relevant points
!--    instead of the entire subdomain. This should lead to better 
!--    load balance between boundary and non-boundary PEs.
       IF( ( bc_dirichlet_l  .OR.  bc_radiation_l )  .AND.  i <= nxl + 2  .OR. &
           ( bc_dirichlet_r  .OR.  bc_radiation_r )  .AND.  i >= nxr - 2  .OR. &
           ( bc_dirichlet_s  .OR.  bc_radiation_s )  .AND.  j <= nys + 2  .OR. &
           ( bc_dirichlet_n  .OR.  bc_radiation_n )  .AND.  j >= nyn - 2 )  THEN
          nzb_max_l = nzt
       ELSE
          nzb_max_l = nzb_max
       END IF
!
!--    Compute southside fluxes of the respective PE bounds.
       IF ( j == nys )  THEN
!
!--       Up to the top of the highest topography.
          DO  k = nzb+1, nzb_max_l

             ibit5 = REAL( IBITS(advc_flags_1(k,j-1,i),5,1), KIND = wp )
             ibit4 = REAL( IBITS(advc_flags_1(k,j-1,i),4,1), KIND = wp )
             ibit3 = REAL( IBITS(advc_flags_1(k,j-1,i),3,1), KIND = wp )

             v_comp                  = v(k,j,i) - v_gtrans + v_stokes_zu(k)
             swap_flux_y_local(k,tn) = v_comp *         (                     &
                                               ( 37.0_wp * ibit5 * adv_sca_5  &
                                            +     7.0_wp * ibit4 * adv_sca_3  &
                                            +              ibit3 * adv_sca_1  &
                                               ) *                            &
                                           ( sk(k,j,i)  + sk(k,j-1,i)     )   &
                                         -     (  8.0_wp * ibit5 * adv_sca_5  &
                                            +              ibit4 * adv_sca_3  &
                                                ) *                           &
                                           ( sk(k,j+1,i) + sk(k,j-2,i)    )   &
                                         +     (           ibit5 * adv_sca_5  &
                                               ) *                            &
                                           ( sk(k,j+2,i) + sk(k,j-3,i)    )   &
                                                        )

             swap_diss_y_local(k,tn) = -ABS( v_comp ) * (                     &
                                               ( 10.0_wp * ibit5 * adv_sca_5  &
                                            +     3.0_wp * ibit4 * adv_sca_3  &
                                            +              ibit3 * adv_sca_1  &
                                               ) *                            &
                                            ( sk(k,j,i)   - sk(k,j-1,i)  )    &
                                        -      (  5.0_wp * ibit5 * adv_sca_5  &
                                            +              ibit4 * adv_sca_3  &
                                            ) *                               &
                                            ( sk(k,j+1,i) - sk(k,j-2,i)  )    &
                                        +      (           ibit5 * adv_sca_5  &
                                               ) *                            &
                                            ( sk(k,j+2,i) - sk(k,j-3,i)  )    &
                                                        )

          ENDDO
!
!--       Above to the top of the highest topography. No degradation necessary.
          DO  k = nzb_max_l+1, nzt

             v_comp                  = v(k,j,i) - v_gtrans + v_stokes_zu(k)
             swap_flux_y_local(k,tn) = v_comp * (                             &
                                    37.0_wp * ( sk(k,j,i)   + sk(k,j-1,i) )   &
                                  -  8.0_wp * ( sk(k,j+1,i) + sk(k,j-2,i) )   &
                                  +           ( sk(k,j+2,i) + sk(k,j-3,i) )   &
                                                ) * adv_sca_5
              swap_diss_y_local(k,tn) = -ABS( v_comp ) * (                    &
                                    10.0_wp * ( sk(k,j,i)   - sk(k,j-1,i) )   &
                                  -  5.0_wp * ( sk(k,j+1,i) - sk(k,j-2,i) )   &
                                  +             sk(k,j+2,i) - sk(k,j-3,i)     &
                                                         ) * adv_sca_5

          ENDDO

       ENDIF
!
!--    Compute leftside fluxes of the respective PE bounds.
       IF ( i == i_omp )  THEN
        
          DO  k = nzb+1, nzb_max_l

             ibit2 = REAL( IBITS(advc_flags_1(k,j,i-1),2,1), KIND = wp )
             ibit1 = REAL( IBITS(advc_flags_1(k,j,i-1),1,1), KIND = wp )
             ibit0 = REAL( IBITS(advc_flags_1(k,j,i-1),0,1), KIND = wp )

             u_comp                     = u(k,j,i) - u_gtrans + u_stokes_zu(k)
             swap_flux_x_local(k,j,tn) = u_comp * (                           &
                                               ( 37.0_wp * ibit2 * adv_sca_5  &
                                            +     7.0_wp * ibit1 * adv_sca_3  &
                                            +              ibit0 * adv_sca_1  &
                                               ) *                            &
                                            ( sk(k,j,i)   + sk(k,j,i-1)    )  &
                                        -      (  8.0_wp * ibit2 * adv_sca_5  &
                                            +              ibit1 * adv_sca_3  &
                                               ) *                            &
                                            ( sk(k,j,i+1) + sk(k,j,i-2)    )  &
                                        +      (           ibit2 * adv_sca_5  &
                                               ) *                            &
                                            ( sk(k,j,i+2) + sk(k,j,i-3)    )  &
                                                  )

              swap_diss_x_local(k,j,tn) = -ABS( u_comp ) * (                  &
                                               ( 10.0_wp * ibit2 * adv_sca_5  &
                                            +     3.0_wp * ibit1 * adv_sca_3  &
                                            +              ibit0 * adv_sca_1  &
                                               ) *                            &
                                            ( sk(k,j,i)   - sk(k,j,i-1)    )  &
                                        -      (  5.0_wp * ibit2 * adv_sca_5  &
                                            +              ibit1 * adv_sca_3  &
                                               ) *                            &
                                            ( sk(k,j,i+1) - sk(k,j,i-2)    )  &
                                        +      (           ibit2 * adv_sca_5  &
                                               ) *                            &
                                            ( sk(k,j,i+2) - sk(k,j,i-3)    )  &
                                                           )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             u_comp                 = u(k,j,i) - u_gtrans + u_stokes_zu(k)
             swap_flux_x_local(k,j,tn) = u_comp * (                           &
                                      37.0_wp * ( sk(k,j,i)   + sk(k,j,i-1) ) &
                                    -  8.0_wp * ( sk(k,j,i+1) + sk(k,j,i-2) ) &
                                    +           ( sk(k,j,i+2) + sk(k,j,i-3) ) &
                                                  ) * adv_sca_5

             swap_diss_x_local(k,j,tn) = -ABS( u_comp ) * (                   &
                                      10.0_wp * ( sk(k,j,i)   - sk(k,j,i-1) ) &
                                    -  5.0_wp * ( sk(k,j,i+1) - sk(k,j,i-2) ) &
                                    +           ( sk(k,j,i+2) - sk(k,j,i-3) ) &
                                                          ) * adv_sca_5

          ENDDO
           
       ENDIF
!        
!--    Now compute the fluxes and tendency terms for the horizontal and
!--    vertical parts up to the top of the highest topography.
       DO  k = nzb+1, nzb_max_l
!
!--       Note: It is faster to conduct all multiplications explicitly, e.g.
!--       * adv_sca_5 ... than to determine a factor and multiplicate the
!--       flux at the end. 

          ibit2 = REAL( IBITS(advc_flags_1(k,j,i),2,1), KIND = wp )
          ibit1 = REAL( IBITS(advc_flags_1(k,j,i),1,1), KIND = wp )
          ibit0 = REAL( IBITS(advc_flags_1(k,j,i),0,1), KIND = wp )

          u_comp    = u(k,j,i+1) - u_gtrans + u_stokes_zu(k)
          flux_r(k) = u_comp * (                                              &
                     ( 37.0_wp * ibit2 * adv_sca_5                            &
                  +     7.0_wp * ibit1 * adv_sca_3                            &
                  +              ibit0 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k,j,i+1) + sk(k,j,i)   )                    &
              -      (  8.0_wp * ibit2 * adv_sca_5                            &
                  +              ibit1 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k,j,i+2) + sk(k,j,i-1) )                    &
              +      (           ibit2 * adv_sca_5                            &
                     ) *                                                      &
                             ( sk(k,j,i+3) + sk(k,j,i-2) )                    &
                               )

          diss_r(k) = -ABS( u_comp ) * (                                      &
                     ( 10.0_wp * ibit2 * adv_sca_5                            &
                  +     3.0_wp * ibit1 * adv_sca_3                            &
                  +              ibit0 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k,j,i+1) - sk(k,j,i)  )                     &
              -      (  5.0_wp * ibit2 * adv_sca_5                            &
                  +              ibit1 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k,j,i+2) - sk(k,j,i-1) )                    &
              +      (           ibit2 * adv_sca_5                            &
                     ) *                                                      &
                             ( sk(k,j,i+3) - sk(k,j,i-2) )                    &
                                       )

          ibit5 = REAL( IBITS(advc_flags_1(k,j,i),5,1), KIND = wp )
          ibit4 = REAL( IBITS(advc_flags_1(k,j,i),4,1), KIND = wp )
          ibit3 = REAL( IBITS(advc_flags_1(k,j,i),3,1), KIND = wp )

          v_comp    = v(k,j+1,i) - v_gtrans + v_stokes_zu(k)
          flux_n(k) = v_comp * (                                              &
                     ( 37.0_wp * ibit5 * adv_sca_5                            &
                  +     7.0_wp * ibit4 * adv_sca_3                            &
                  +              ibit3 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k,j+1,i) + sk(k,j,i)   )                    &
              -      (  8.0_wp * ibit5 * adv_sca_5                            &
                  +              ibit4 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k,j+2,i) + sk(k,j-1,i) )                    &
              +      (           ibit5 * adv_sca_5                            &
                     ) *                                                      &
                             ( sk(k,j+3,i) + sk(k,j-2,i) )                    &
                               )

          diss_n(k) = -ABS( v_comp ) * (                                      &
                     ( 10.0_wp * ibit5 * adv_sca_5                            &
                  +     3.0_wp * ibit4 * adv_sca_3                            &
                  +              ibit3 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k,j+1,i) - sk(k,j,i)   )                    &
              -      (  5.0_wp * ibit5 * adv_sca_5                            &
                  +              ibit4 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k,j+2,i) - sk(k,j-1,i) )                    &
              +      (           ibit5 * adv_sca_5                            &
                     ) *                                                      &
                             ( sk(k,j+3,i) - sk(k,j-2,i) )                    &
                                       )
       ENDDO
!
!--    Now compute the fluxes and tendency terms for the horizontal and
!--    vertical parts above the top of the highest topography. No degradation
!--    for the horizontal parts, but for the vertical it is stell needed.
       DO  k = nzb_max_l+1, nzt

          u_comp    = u(k,j,i+1) - u_gtrans + u_stokes_zu(k)
          flux_r(k) = u_comp * (                                              &
                      37.0_wp * ( sk(k,j,i+1) + sk(k,j,i)   )                 &
                    -  8.0_wp * ( sk(k,j,i+2) + sk(k,j,i-1) )                 &
                    +           ( sk(k,j,i+3) + sk(k,j,i-2) ) ) * adv_sca_5
          diss_r(k) = -ABS( u_comp ) * (                                      &
                      10.0_wp * ( sk(k,j,i+1) - sk(k,j,i)   )                 &
                    -  5.0_wp * ( sk(k,j,i+2) - sk(k,j,i-1) )                 &
                    +           ( sk(k,j,i+3) - sk(k,j,i-2) ) ) * adv_sca_5

          v_comp    = v(k,j+1,i) - v_gtrans + v_stokes_zu(k)
          flux_n(k) = v_comp * (                                              &
                      37.0_wp * ( sk(k,j+1,i) + sk(k,j,i)   )                 &
                    -  8.0_wp * ( sk(k,j+2,i) + sk(k,j-1,i) )                 &
                    +           ( sk(k,j+3,i) + sk(k,j-2,i) ) ) * adv_sca_5
          diss_n(k) = -ABS( v_comp ) * (                                      &
                      10.0_wp * ( sk(k,j+1,i) - sk(k,j,i)   )                 &
                    -  5.0_wp * ( sk(k,j+2,i) - sk(k,j-1,i) )                 &
                    +           ( sk(k,j+3,i) - sk(k,j-2,i) ) ) * adv_sca_5

       ENDDO
!
!--    Now, compute vertical fluxes. Split loop into a part treating the 
!--    lowest 2 grid points with indirect indexing, a main loop without
!--    indirect indexing, and a loop for the uppermost 2 grip points with
!--    indirect indexing. This allows better vectorization for the main loop.
!--    First, compute the flux at model surface, which need has to be 
!--    calculated explicetely for the tendency at
!--    the first w-level. For topography wall this is done implicitely by
!--    advc_flags_1.
       flux_t(nzb) = 0.0_wp
       diss_t(nzb) = 0.0_wp
       DO  k = nzb+1, nzb+2
          ibit8 = REAL( IBITS(advc_flags_1(k,j,i),8,1), KIND = wp )
          ibit7 = REAL( IBITS(advc_flags_1(k,j,i),7,1), KIND = wp )
          ibit6 = REAL( IBITS(advc_flags_1(k,j,i),6,1), KIND = wp )
!
!--       k index has to be modified near bottom and top, else array
!--       subscripts will be exceeded.
          k_ppp = k + 3 * ibit8
          k_pp  = k + 2 * ( 1 - ibit6  )
          k_mm  = k - 2 * ibit8


          flux_t(k) = w(k,j,i) * rho_air_zw(k) * (                            &
                     ( 37.0_wp * ibit8 * adv_sca_5                            &
                  +     7.0_wp * ibit7 * adv_sca_3                            &
                  +              ibit6 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k+1,j,i)  + sk(k,j,i)    )                  &
              -      (  8.0_wp * ibit8 * adv_sca_5                            &
                  +              ibit7 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k_pp,j,i) + sk(k-1,j,i)  )                  &
              +      (           ibit8 * adv_sca_5                            &
                     ) *     ( sk(k_ppp,j,i)+ sk(k_mm,j,i) )                  &
                                 )

          diss_t(k) = -ABS( w(k,j,i) ) * rho_air_zw(k) * (                    &
                     ( 10.0_wp * ibit8 * adv_sca_5                            &
                  +     3.0_wp * ibit7 * adv_sca_3                            &
                  +              ibit6 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k+1,j,i)   - sk(k,j,i)    )                 &
              -      (  5.0_wp * ibit8 * adv_sca_5                            &
                  +              ibit7 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k_pp,j,i)  - sk(k-1,j,i)  )                 &
              +      (           ibit8 * adv_sca_5                            &
                     ) *                                                      &
                             ( sk(k_ppp,j,i) - sk(k_mm,j,i) )                 &
                                         )
       ENDDO 
       
       DO  k = nzb+3, nzt-2
          ibit8 = REAL( IBITS(advc_flags_1(k,j,i),8,1), KIND = wp )
          ibit7 = REAL( IBITS(advc_flags_1(k,j,i),7,1), KIND = wp )
          ibit6 = REAL( IBITS(advc_flags_1(k,j,i),6,1), KIND = wp )

          flux_t(k) = w(k,j,i) * rho_air_zw(k) * (                            &
                     ( 37.0_wp * ibit8 * adv_sca_5                            &
                  +     7.0_wp * ibit7 * adv_sca_3                            &
                  +              ibit6 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k+1,j,i)  + sk(k,j,i)    )                  &
              -      (  8.0_wp * ibit8 * adv_sca_5                            &
                  +              ibit7 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k+2,j,i) + sk(k-1,j,i)  )                   &
              +      (           ibit8 * adv_sca_5                            &
                     ) *     ( sk(k+3,j,i)+ sk(k-2,j,i) )                     &
                                                 )

          diss_t(k) = -ABS( w(k,j,i) ) * rho_air_zw(k) * (                    &
                     ( 10.0_wp * ibit8 * adv_sca_5                            &
                  +     3.0_wp * ibit7 * adv_sca_3                            &
                  +              ibit6 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k+1,j,i)   - sk(k,j,i)    )                 &
              -      (  5.0_wp * ibit8 * adv_sca_5                            &
                  +              ibit7 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k+2,j,i)  - sk(k-1,j,i)  )                  &
              +      (           ibit8 * adv_sca_5                            &
                     ) *                                                      &
                             ( sk(k+3,j,i) - sk(k-2,j,i) )                    &
                                                         )
       ENDDO       
       
       DO  k = nzt-1, nzt
          ibit8 = REAL( IBITS(advc_flags_1(k,j,i),8,1), KIND = wp )
          ibit7 = REAL( IBITS(advc_flags_1(k,j,i),7,1), KIND = wp )
          ibit6 = REAL( IBITS(advc_flags_1(k,j,i),6,1), KIND = wp )
!
!--       k index has to be modified near bottom and top, else array
!--       subscripts will be exceeded.
          k_ppp = k + 3 * ibit8
          k_pp  = k + 2 * ( 1 - ibit6  )
          k_mm  = k - 2 * ibit8


          flux_t(k) = w(k,j,i) * rho_air_zw(k) * (                            &
                     ( 37.0_wp * ibit8 * adv_sca_5                            &
                  +     7.0_wp * ibit7 * adv_sca_3                            &
                  +              ibit6 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k+1,j,i)  + sk(k,j,i)    )                  &
              -      (  8.0_wp * ibit8 * adv_sca_5                            &
                  +              ibit7 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k_pp,j,i) + sk(k-1,j,i)  )                  &
              +      (           ibit8 * adv_sca_5                            &
                     ) *     ( sk(k_ppp,j,i)+ sk(k_mm,j,i) )                  &
                                                 )

          diss_t(k) = -ABS( w(k,j,i) ) * rho_air_zw(k) * (                    &
                     ( 10.0_wp * ibit8 * adv_sca_5                            &
                  +     3.0_wp * ibit7 * adv_sca_3                            &
                  +              ibit6 * adv_sca_1                            &
                     ) *                                                      &
                             ( sk(k+1,j,i)   - sk(k,j,i)    )                 &
              -      (  5.0_wp * ibit8 * adv_sca_5                            &
                  +              ibit7 * adv_sca_3                            &
                     ) *                                                      &
                             ( sk(k_pp,j,i)  - sk(k-1,j,i)  )                 &
              +      (           ibit8 * adv_sca_5                            &
                     ) *                                                      &
                             ( sk(k_ppp,j,i) - sk(k_mm,j,i) )                 &
                                                         )
       ENDDO 

       DO  k = nzb+1, nzb_max_l

          flux_d    = flux_t(k-1)
          diss_d    = diss_t(k-1)
          
          ibit2 = REAL( IBITS(advc_flags_1(k,j,i),2,1), KIND = wp )
          ibit1 = REAL( IBITS(advc_flags_1(k,j,i),1,1), KIND = wp )
          ibit0 = REAL( IBITS(advc_flags_1(k,j,i),0,1), KIND = wp )
          
          ibit5 = REAL( IBITS(advc_flags_1(k,j,i),5,1), KIND = wp )
          ibit4 = REAL( IBITS(advc_flags_1(k,j,i),4,1), KIND = wp )
          ibit3 = REAL( IBITS(advc_flags_1(k,j,i),3,1), KIND = wp )
          
          ibit8 = REAL( IBITS(advc_flags_1(k,j,i),8,1), KIND = wp )
          ibit7 = REAL( IBITS(advc_flags_1(k,j,i),7,1), KIND = wp )
          ibit6 = REAL( IBITS(advc_flags_1(k,j,i),6,1), KIND = wp )
!
!--       Calculate the divergence of the velocity field. A respective
!--       correction is needed to overcome numerical instabilities introduced
!--       by a not sufficient reduction of divergences near topography.
          div         =   ( u(k,j,i+1) * ( ibit0 + ibit1 + ibit2 )            &
                          - u(k,j,i)   * (                                    &
                        REAL( IBITS(advc_flags_1(k,j,i-1),0,1), KIND = wp )   &
                      + REAL( IBITS(advc_flags_1(k,j,i-1),1,1), KIND = wp )   &
                      + REAL( IBITS(advc_flags_1(k,j,i-1),2,1), KIND = wp )   &
                                         )                                    &
                          ) * ddx                                             &
                        + ( v(k,j+1,i) * ( ibit3 + ibit4 + ibit5 )            &
                          - v(k,j,i)   * (                                    &
                        REAL( IBITS(advc_flags_1(k,j-1,i),3,1), KIND = wp )   &
                      + REAL( IBITS(advc_flags_1(k,j-1,i),4,1), KIND = wp )   &
                      + REAL( IBITS(advc_flags_1(k,j-1,i),5,1), KIND = wp )   &
                                         )                                    &
                          ) * ddy                                             &
                        + ( w(k,j,i) * rho_air_zw(k) *                        &
                                         ( ibit6 + ibit7 + ibit8 )            &
                          - w(k-1,j,i) * rho_air_zw(k-1) *                    &
                                         (                                    &
                        REAL( IBITS(advc_flags_1(k-1,j,i),6,1), KIND = wp )   &
                      + REAL( IBITS(advc_flags_1(k-1,j,i),7,1), KIND = wp )   &
                      + REAL( IBITS(advc_flags_1(k-1,j,i),8,1), KIND = wp )   &
                                         )                                    &      
                          ) * drho_air(k) * ddzw(k)

          tend(k,j,i) = tend(k,j,i) - (                                       &
                        ( flux_r(k) + diss_r(k) - swap_flux_x_local(k,j,tn) - &
                          swap_diss_x_local(k,j,tn)            ) * ddx        &
                      + ( flux_n(k) + diss_n(k) - swap_flux_y_local(k,tn)   - &
                          swap_diss_y_local(k,tn)              ) * ddy        &
                      + ( ( flux_t(k) + diss_t(k) ) -                         &
                          ( flux_d    + diss_d    )                           &
                                                    ) * drho_air(k) * ddzw(k) &
                                      ) + sk(k,j,i) * div


          swap_flux_y_local(k,tn)   = flux_n(k)
          swap_diss_y_local(k,tn)   = diss_n(k)
          swap_flux_x_local(k,j,tn) = flux_r(k)
          swap_diss_x_local(k,j,tn) = diss_r(k)

       ENDDO
       
       DO  k = nzb_max_l+1, nzt

          flux_d    = flux_t(k-1)
          diss_d    = diss_t(k-1)
!
!--       Calculate the divergence of the velocity field. A respective
!--       correction is needed to overcome numerical instabilities introduced
!--       by a not sufficient reduction of divergences near topography.
          div         =   ( u(k,j,i+1) - u(k,j,i) ) * ddx                     &
                        + ( v(k,j+1,i) - v(k,j,i) ) * ddy                     &
                        + ( w(k,j,i) * rho_air_zw(k)                          &
                          - w(k-1,j,i) * rho_air_zw(k-1)                      &
                                                  ) * drho_air(k) * ddzw(k)

          tend(k,j,i) = tend(k,j,i) - (                                       &
                        ( flux_r(k) + diss_r(k) - swap_flux_x_local(k,j,tn) - &
                          swap_diss_x_local(k,j,tn)            ) * ddx        &
                      + ( flux_n(k) + diss_n(k) - swap_flux_y_local(k,tn)   - &
                          swap_diss_y_local(k,tn)              ) * ddy        &
                      + ( ( flux_t(k) + diss_t(k) ) -                         &
                          ( flux_d    + diss_d    )                           &
                                                    ) * drho_air(k) * ddzw(k) &
                                      ) + sk(k,j,i) * div


          swap_flux_y_local(k,tn)   = flux_n(k)
          swap_diss_y_local(k,tn)   = diss_n(k)
          swap_flux_x_local(k,j,tn) = flux_r(k)
          swap_diss_x_local(k,j,tn) = diss_r(k)

       ENDDO

!
!--    Evaluation of statistics.
       SELECT CASE ( sk_char )

          CASE ( 'pt' )

             DO  k = nzb, nzt
                sums_wspts_ws_l(k,tn) = sums_wspts_ws_l(k,tn) +                &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO
            
          CASE ( 'sa' )

             DO  k = nzb, nzt
                sums_wssas_ws_l(k,tn) = sums_wssas_ws_l(k,tn) +                &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO
            
          CASE ( 'q' )

             DO  k = nzb, nzt
                sums_wsqs_ws_l(k,tn)  = sums_wsqs_ws_l(k,tn) +                 &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO

          CASE ( 'qc' )

             DO  k = nzb, nzt
                sums_wsqcs_ws_l(k,tn)  = sums_wsqcs_ws_l(k,tn) +               &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO


          CASE ( 'qr' )

             DO  k = nzb, nzt
                sums_wsqrs_ws_l(k,tn)  = sums_wsqrs_ws_l(k,tn) +               &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO

          CASE ( 'nc' )

             DO  k = nzb, nzt
                sums_wsncs_ws_l(k,tn)  = sums_wsncs_ws_l(k,tn) +               &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO

          CASE ( 'nr' )

             DO  k = nzb, nzt
                sums_wsnrs_ws_l(k,tn)  = sums_wsnrs_ws_l(k,tn) +               &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO

          CASE ( 's' )

             DO  k = nzb, nzt
                sums_wsss_ws_l(k,tn)  = sums_wsss_ws_l(k,tn) +                 &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO

         CASE ( 'aerosol_mass', 'aerosol_number', 'salsa_gas' )

             DO  k = nzb, nzt
                sums_salsa_ws_l(k,tn)  = sums_salsa_ws_l(k,tn) +               &
                    ( flux_t(k) / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                    + diss_t(k) / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS( w(k,j,i) - hom(k,1,3,0)            )  &
                    ) * weight_substep(intermediate_timestep_count)
             ENDDO

!          CASE ( 'kc' )
          !kk Has to be implemented for kpp chemistry


         END SELECT

    END SUBROUTINE advec_s_ws_ij




!------------------------------------------------------------------------------!
! Description:
! ------------
!> Advection of u-component - Call for grid point i,j
!------------------------------------------------------------------------------!
    SUBROUTINE advec_u_ws_ij( i, j, i_omp, tn )

       USE arrays_3d,                                                          &
           ONLY:  ddzw, diss_l_u, diss_s_u, flux_l_u, flux_s_u, tend, u, v, w, &
                  drho_air, rho_air_zw

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s, intermediate_timestep_count, &
                  u_gtrans, v_gtrans

       USE grid_variables,                                                     &
           ONLY:  ddx, ddy

       USE indices,                                                            &
           ONLY:  nyn, nys, nxl, nxlu, nxr, nzb, nzb_max, nzt, advc_flags_1

       USE kinds

       USE statistics,                                                         &
           ONLY:  hom, sums_us2_ws_l, sums_wsus_ws_l, weight_substep

       IMPLICIT NONE

       INTEGER(iwp) ::  i         !< grid index along x-direction
       INTEGER(iwp) ::  i_omp     !< leftmost index on subdomain, or in case of OpenMP, on thread
       INTEGER(iwp) ::  j         !< grid index along y-direction
       INTEGER(iwp) ::  k         !< grid index along z-direction
       INTEGER(iwp) ::  k_mm      !< k-2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_pp      !< k+2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_ppp     !< k+3 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  nzb_max_l !< index indicating upper bound for order degradation of horizontal advection terms  
       INTEGER(iwp) ::  tn        !< number of OpenMP thread
       
       REAL(wp)    ::  ibit9    !< flag indicating 1st-order scheme along x-direction
       REAL(wp)    ::  ibit10   !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)    ::  ibit11   !< flag indicating 5th-order scheme along x-direction
       REAL(wp)    ::  ibit12   !< flag indicating 1st-order scheme along y-direction
       REAL(wp)    ::  ibit13   !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)    ::  ibit14   !< flag indicating 5th-order scheme along y-direction
       REAL(wp)    ::  ibit15   !< flag indicating 1st-order scheme along z-direction
       REAL(wp)    ::  ibit16   !< flag indicating 3rd-order scheme along z-direction
       REAL(wp)    ::  ibit17   !< flag indicating 5th-order scheme along z-direction
       REAL(wp)    ::  diss_d   !< artificial dissipation term at grid box bottom
       REAL(wp)    ::  div      !< diverence on u-grid
       REAL(wp)    ::  flux_d   !< 6th-order flux at grid box bottom
       REAL(wp)    ::  gu       !< Galilei-transformation velocity along x
       REAL(wp)    ::  gv       !< Galilei-transformation velocity along y
       REAL(wp)    ::  u_comp_l !< advection velocity along x at leftmost grid point on subdomain
       
       REAL(wp), DIMENSION(nzb:nzt+1) ::  diss_n !< discretized artificial dissipation at northward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1) ::  diss_r !< discretized artificial dissipation at rightward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1) ::  diss_t !< discretized artificial dissipation at top of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1) ::  flux_n !< discretized 6th-order flux at northward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1) ::  flux_r !< discretized 6th-order flux at rightward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1) ::  flux_t !< discretized 6th-order flux at top of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1) ::  u_comp !< advection velocity along x
       REAL(wp), DIMENSION(nzb:nzt+1) ::  v_comp !< advection velocity along y
       REAL(wp), DIMENSION(nzb:nzt+1) ::  w_comp !< advection velocity along z
!
!--    Used local modified copy of nzb_max (used to degrade order of 
!--    discretization) at non-cyclic boundaries. Modify only at relevant points
!--    instead of the entire subdomain. This should lead to better 
!--    load balance between boundary and non-boundary PEs.
       IF( ( bc_dirichlet_l  .OR.  bc_radiation_l )  .AND.  i <= nxl + 2  .OR. &
           ( bc_dirichlet_r  .OR.  bc_radiation_r )  .AND.  i >= nxr - 2  .OR. &
           ( bc_dirichlet_s  .OR.  bc_radiation_s )  .AND.  j <= nys + 2  .OR. &
           ( bc_dirichlet_n  .OR.  bc_radiation_n )  .AND.  j >= nyn - 2 )  THEN
          nzb_max_l = nzt
       ELSE
          nzb_max_l = nzb_max
       END IF
       
       gu = 2.0_wp * u_gtrans
       gv = 2.0_wp * v_gtrans
!
!--    Compute southside fluxes for the respective boundary of PE
       IF ( j == nys  )  THEN
       
          DO  k = nzb+1, nzb_max_l

             ibit14 = REAL( IBITS(advc_flags_1(k,j-1,i),14,1), KIND = wp )
             ibit13 = REAL( IBITS(advc_flags_1(k,j-1,i),13,1), KIND = wp )
             ibit12 = REAL( IBITS(advc_flags_1(k,j-1,i),12,1), KIND = wp )

             v_comp(k)      = v(k,j,i) + v(k,j,i-1) - gv
             flux_s_u(k,tn) = v_comp(k) * (                                   &
                            ( 37.0_wp * ibit14 * adv_mom_5                    &
                         +     7.0_wp * ibit13 * adv_mom_3                    &
                         +              ibit12 * adv_mom_1                    &
                            ) *                                               &
                                        ( u(k,j,i)   + u(k,j-1,i) )           &
                     -      (  8.0_wp * ibit14 * adv_mom_5                    &
                         +              ibit13 * adv_mom_3                    &
                            ) *                                               &
                                        ( u(k,j+1,i) + u(k,j-2,i) )           &
                     +      (           ibit14 * adv_mom_5                    &
                            ) *                                               &
                                        ( u(k,j+2,i) + u(k,j-3,i) )           &
                                          )

             diss_s_u(k,tn) = - ABS ( v_comp(k) ) * (                         &
                            ( 10.0_wp * ibit14 * adv_mom_5                    &
                         +     3.0_wp * ibit13 * adv_mom_3                    &
                         +              ibit12 * adv_mom_1                    &
                            ) *                                               &
                                        ( u(k,j,i)   - u(k,j-1,i) )           &
                     -      (  5.0_wp * ibit14 * adv_mom_5                    &
                         +              ibit13 * adv_mom_3                    &
                            ) *                                               &
                                        ( u(k,j+1,i) - u(k,j-2,i) )           &
                     +      (           ibit14 * adv_mom_5                    &
                            ) *                                               &
                                        ( u(k,j+2,i) - u(k,j-3,i) )           &
                                                    )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             v_comp(k)      = v(k,j,i) + v(k,j,i-1) - gv
             flux_s_u(k,tn) = v_comp(k) * (                                   &
                           37.0_wp * ( u(k,j,i) + u(k,j-1,i)   )              &
                         -  8.0_wp * ( u(k,j+1,i) + u(k,j-2,i) )              &
                         +           ( u(k,j+2,i) + u(k,j-3,i) ) ) * adv_mom_5
             diss_s_u(k,tn) = - ABS(v_comp(k)) * (                            &
                           10.0_wp * ( u(k,j,i) - u(k,j-1,i)   )              &
                         -  5.0_wp * ( u(k,j+1,i) - u(k,j-2,i) )              &
                         +           ( u(k,j+2,i) - u(k,j-3,i) ) ) * adv_mom_5

          ENDDO
          
       ENDIF
!
!--    Compute leftside fluxes for the respective boundary of PE
       IF ( i == i_omp  .OR.  i == nxlu )  THEN
       
          DO  k = nzb+1, nzb_max_l

             ibit11 = REAL( IBITS(advc_flags_1(k,j,i-1),11,1), KIND = wp )
             ibit10 = REAL( IBITS(advc_flags_1(k,j,i-1),10,1), KIND = wp )
             ibit9  = REAL( IBITS(advc_flags_1(k,j,i-1),9,1),  KIND = wp )

             u_comp_l         = u(k,j,i) + u(k,j,i-1) - gu
             flux_l_u(k,j,tn) = u_comp_l * (                                  &
                              ( 37.0_wp * ibit11 * adv_mom_5                  &
                           +     7.0_wp * ibit10 * adv_mom_3                  &
                           +              ibit9  * adv_mom_1                  &
                              ) *                                             &
                                          ( u(k,j,i)   + u(k,j,i-1) )         &
                       -      (  8.0_wp * ibit11 * adv_mom_5                  &
                           +              ibit10 * adv_mom_3                  &
                              ) *                                             &
                                          ( u(k,j,i+1) + u(k,j,i-2) )         &
                       +      (           ibit11 * adv_mom_5                  &
                              ) *                                             &
                                          ( u(k,j,i+2) + u(k,j,i-3) )         &
                                           )

              diss_l_u(k,j,tn) = - ABS( u_comp_l ) * (                        &
                              ( 10.0_wp * ibit11 * adv_mom_5                  &
                           +     3.0_wp * ibit10 * adv_mom_3                  &
                           +              ibit9  * adv_mom_1                  &
                              ) *                                             &
                                        ( u(k,j,i)   - u(k,j,i-1) )           &
                       -      (  5.0_wp * ibit11 * adv_mom_5                  &
                           +              ibit10 * adv_mom_3                  &
                              ) *                                             &
                                        ( u(k,j,i+1) - u(k,j,i-2) )           &
                       +      (           ibit11 * adv_mom_5                  &
                              ) *                                             &
                                        ( u(k,j,i+2) - u(k,j,i-3) )           &
                                                     )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             u_comp_l         = u(k,j,i) + u(k,j,i-1) - gu
             flux_l_u(k,j,tn) = u_comp_l * (                                   &
                             37.0_wp * ( u(k,j,i) + u(k,j,i-1)   )             &
                           -  8.0_wp * ( u(k,j,i+1) + u(k,j,i-2) )             &
                           +           ( u(k,j,i+2) + u(k,j,i-3) ) ) * adv_mom_5
             diss_l_u(k,j,tn) = - ABS(u_comp_l) * (                            &
                             10.0_wp * ( u(k,j,i) - u(k,j,i-1)   )             &
                           -  5.0_wp * ( u(k,j,i+1) - u(k,j,i-2) )             &
                           +           ( u(k,j,i+2) - u(k,j,i-3) ) ) * adv_mom_5

          ENDDO
          
       ENDIF
!
!--    Now compute the fluxes tendency terms for the horizontal and
!--    vertical parts.
       DO  k = nzb+1, nzb_max_l

          ibit11 = REAL( IBITS(advc_flags_1(k,j,i),11,1), KIND = wp )
          ibit10 = REAL( IBITS(advc_flags_1(k,j,i),10,1), KIND = wp )
          ibit9  = REAL( IBITS(advc_flags_1(k,j,i),9,1),  KIND = wp )

          u_comp(k) = u(k,j,i+1) + u(k,j,i)
          flux_r(k) = ( u_comp(k) - gu ) * (                                  &
                     ( 37.0_wp * ibit11 * adv_mom_5                           &
                  +     7.0_wp * ibit10 * adv_mom_3                           &
                  +              ibit9  * adv_mom_1                           &
                     ) *                                                      &
                                    ( u(k,j,i+1) + u(k,j,i)   )               &
              -      (  8.0_wp * ibit11 * adv_mom_5                           &
                  +              ibit10 * adv_mom_3                           & 
                     ) *                                                      &
                                    ( u(k,j,i+2) + u(k,j,i-1) )               &
              +      (           ibit11 * adv_mom_5                           &
                     ) *                                                      &
                                    ( u(k,j,i+3) + u(k,j,i-2) )               &
                                           )

          diss_r(k) = - ABS( u_comp(k) - gu ) * (                             &
                     ( 10.0_wp * ibit11 * adv_mom_5                           &
                  +     3.0_wp * ibit10 * adv_mom_3                           &
                  +              ibit9  * adv_mom_1                           &
                     ) *                                                      &
                                    ( u(k,j,i+1) - u(k,j,i)   )               &
              -      (  5.0_wp * ibit11 * adv_mom_5                           &
                  +              ibit10 * adv_mom_3                           &
                     ) *                                                      &
                                    ( u(k,j,i+2) - u(k,j,i-1) )               &
              +      (           ibit11 * adv_mom_5                           &
                     ) *                                                      &
                                    ( u(k,j,i+3) - u(k,j,i-2) )               &
                                                )

          ibit14 = REAL( IBITS(advc_flags_1(k,j,i),14,1), KIND = wp )
          ibit13 = REAL( IBITS(advc_flags_1(k,j,i),13,1), KIND = wp )
          ibit12 = REAL( IBITS(advc_flags_1(k,j,i),12,1), KIND = wp )

          v_comp(k) = v(k,j+1,i) + v(k,j+1,i-1) - gv
          flux_n(k) = v_comp(k) * (                                           &
                     ( 37.0_wp * ibit14 * adv_mom_5                           &
                  +     7.0_wp * ibit13 * adv_mom_3                           &
                  +              ibit12 * adv_mom_1                           &
                     ) *                                                      &
                                    ( u(k,j+1,i) + u(k,j,i)   )               &
              -      (  8.0_wp * ibit14 * adv_mom_5                           &
                  +              ibit13 * adv_mom_3                           &
                     ) *                                                      &
                                    ( u(k,j+2,i) + u(k,j-1,i) )               &
              +      (           ibit14 * adv_mom_5                           &
                     ) *                                                      &
                                    ( u(k,j+3,i) + u(k,j-2,i) )               &
                                  )

          diss_n(k) = - ABS ( v_comp(k) ) * (                                 &
                     ( 10.0_wp * ibit14 * adv_mom_5                           &
                  +     3.0_wp * ibit13 * adv_mom_3                           &
                  +              ibit12 * adv_mom_1                           &
                     ) *                                                      &
                                    ( u(k,j+1,i) - u(k,j,i)   )               &
              -      (  5.0_wp * ibit14 * adv_mom_5                           &
                  +              ibit13 * adv_mom_3                           &
                     ) *                                                      &
                                    ( u(k,j+2,i) - u(k,j-1,i) )               &
              +      (           ibit14 * adv_mom_5                           &
                     ) *                                                      &
                                    ( u(k,j+3,i) - u(k,j-2,i) )               &
                                            )
       ENDDO

       DO  k = nzb_max_l+1, nzt

          u_comp(k) = u(k,j,i+1) + u(k,j,i)
          flux_r(k) = ( u_comp(k) - gu ) * (                                  &
                         37.0_wp * ( u(k,j,i+1) + u(k,j,i)   )                &
                       -  8.0_wp * ( u(k,j,i+2) + u(k,j,i-1) )                &
                       +           ( u(k,j,i+3) + u(k,j,i-2) ) ) * adv_mom_5
          diss_r(k) = - ABS( u_comp(k) - gu ) * (                             &
                         10.0_wp * ( u(k,j,i+1) - u(k,j,i)   )                &
                       -  5.0_wp * ( u(k,j,i+2) - u(k,j,i-1) )                &
                       +           ( u(k,j,i+3) - u(k,j,i-2) ) ) * adv_mom_5

          v_comp(k) = v(k,j+1,i) + v(k,j+1,i-1) - gv
          flux_n(k) = v_comp(k) * (                                           &
                         37.0_wp * ( u(k,j+1,i) + u(k,j,i)   )                &
                       -  8.0_wp * ( u(k,j+2,i) + u(k,j-1,i) )                &
                       +           ( u(k,j+3,i) + u(k,j-2,i) ) ) * adv_mom_5
          diss_n(k) = - ABS( v_comp(k) ) * (                                  &
                         10.0_wp * ( u(k,j+1,i) - u(k,j,i)   )                &
                       -  5.0_wp * ( u(k,j+2,i) - u(k,j-1,i) )                &
                       +           ( u(k,j+3,i) - u(k,j-2,i) ) ) * adv_mom_5

       ENDDO
!
!--    Now, compute vertical fluxes. Split loop into a part treating the 
!--    lowest 2 grid points with indirect indexing, a main loop without
!--    indirect indexing, and a loop for the uppermost 2 grip points with
!--    indirect indexing. This allows better vectorization for the main loop.
!--    First, compute the flux at model surface, which need has to be 
!--    calculated explicetely for the tendency at
!--    the first w-level. For topography wall this is done implicitely by
!--    advc_flags_1.
       flux_t(nzb) = 0.0_wp
       diss_t(nzb) = 0.0_wp
       w_comp(nzb) = 0.0_wp
       DO  k = nzb+1, nzb+2
!
!--       k index has to be modified near bottom and top, else array
!--       subscripts will be exceeded.
          ibit17 = REAL( IBITS(advc_flags_1(k,j,i),17,1), KIND = wp )
          ibit16 = REAL( IBITS(advc_flags_1(k,j,i),16,1), KIND = wp )
          ibit15 = REAL( IBITS(advc_flags_1(k,j,i),15,1), KIND = wp )

          k_ppp = k + 3 * ibit17
          k_pp  = k + 2 * ( 1 - ibit15 )
          k_mm  = k - 2 * ibit17

          w_comp(k) = w(k,j,i) + w(k,j,i-1)
          flux_t(k) = w_comp(k) * rho_air_zw(k) * (                           &
                     ( 37.0_wp * ibit17 * adv_mom_5                           &
                  +     7.0_wp * ibit16 * adv_mom_3                           &
                  +              ibit15 * adv_mom_1                           &
                     ) *                                                      &
                                ( u(k+1,j,i)  + u(k,j,i)     )                &
              -      (  8.0_wp * ibit17 * adv_mom_5                           &
                  +              ibit16 * adv_mom_3                           &
                     ) *                                                      &
                                ( u(k_pp,j,i) + u(k-1,j,i)   )                &
              +      (           ibit17 * adv_mom_5                           &
                     ) *                                                      &
                                ( u(k_ppp,j,i) + u(k_mm,j,i) )                &
                                                  )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air_zw(k) * (                  &
                     ( 10.0_wp * ibit17 * adv_mom_5                           &
                  +     3.0_wp * ibit16 * adv_mom_3                           &
                  +              ibit15 * adv_mom_1                           &
                     ) *                                                      &
                                ( u(k+1,j,i)   - u(k,j,i)    )                &
              -      (  5.0_wp * ibit17 * adv_mom_5                           &
                  +              ibit16 * adv_mom_3                           &
                     ) *                                                      &
                                ( u(k_pp,j,i)  - u(k-1,j,i)  )                &
              +      (           ibit17 * adv_mom_5                           &
                     ) *                                                      &
                                ( u(k_ppp,j,i) - u(k_mm,j,i) )                &
                                                           )
       ENDDO
       
       DO  k = nzb+3, nzt-2

          ibit17 = REAL( IBITS(advc_flags_1(k,j,i),17,1), KIND = wp )
          ibit16 = REAL( IBITS(advc_flags_1(k,j,i),16,1), KIND = wp )
          ibit15 = REAL( IBITS(advc_flags_1(k,j,i),15,1), KIND = wp )

          w_comp(k) = w(k,j,i) + w(k,j,i-1)
          flux_t(k) = w_comp(k) * rho_air_zw(k) * (                           &
                     ( 37.0_wp * ibit17 * adv_mom_5                           &
                  +     7.0_wp * ibit16 * adv_mom_3                           &
                  +              ibit15 * adv_mom_1                           &
                     ) *                                                      &
                                ( u(k+1,j,i)  + u(k,j,i)     )                &
              -      (  8.0_wp * ibit17 * adv_mom_5                           &
                  +              ibit16 * adv_mom_3                           &
                     ) *                                                      &
                                ( u(k+2,j,i) + u(k-1,j,i)   )                 &
              +      (           ibit17 * adv_mom_5                           &
                     ) *                                                      &
                                ( u(k+3,j,i) + u(k-2,j,i) )                   &
                                                  )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air_zw(k) * (                  &
                     ( 10.0_wp * ibit17 * adv_mom_5                           &
                  +     3.0_wp * ibit16 * adv_mom_3                           &
                  +              ibit15 * adv_mom_1                           &
                     ) *                                                      &
                                ( u(k+1,j,i)   - u(k,j,i)    )                &
              -      (  5.0_wp * ibit17 * adv_mom_5                           &
                  +              ibit16 * adv_mom_3                           &
                     ) *                                                      &
                                ( u(k+2,j,i)  - u(k-1,j,i)  )                 &
              +      (           ibit17 * adv_mom_5                           &
                     ) *                                                      &
                                ( u(k+3,j,i) - u(k-2,j,i) )                   &
                                                           )
       ENDDO
       
       DO  k = nzt-1, nzt
!
!--       k index has to be modified near bottom and top, else array
!--       subscripts will be exceeded.
          ibit17 = REAL( IBITS(advc_flags_1(k,j,i),17,1), KIND = wp )
          ibit16 = REAL( IBITS(advc_flags_1(k,j,i),16,1), KIND = wp )
          ibit15 = REAL( IBITS(advc_flags_1(k,j,i),15,1), KIND = wp )

          k_ppp = k + 3 * ibit17
          k_pp  = k + 2 * ( 1 - ibit15 )
          k_mm  = k - 2 * ibit17

          w_comp(k) = w(k,j,i) + w(k,j,i-1)
          flux_t(k) = w_comp(k) * rho_air_zw(k) * (                           &
                     ( 37.0_wp * ibit17 * adv_mom_5                           &
                  +     7.0_wp * ibit16 * adv_mom_3                           &
                  +              ibit15 * adv_mom_1                           &
                     ) *                                                      &
                                ( u(k+1,j,i)  + u(k,j,i)     )                &
              -      (  8.0_wp * ibit17 * adv_mom_5                           &
                  +              ibit16 * adv_mom_3                           &
                     ) *                                                      &
                                ( u(k_pp,j,i) + u(k-1,j,i)   )                &
              +      (           ibit17 * adv_mom_5                           &
                     ) *                                                      &
                                ( u(k_ppp,j,i) + u(k_mm,j,i) )                &
                                                  )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air_zw(k) * (                  &
                     ( 10.0_wp * ibit17 * adv_mom_5                           &
                  +     3.0_wp * ibit16 * adv_mom_3                           &
                  +              ibit15 * adv_mom_1                           &
                     ) *                                                      &
                                ( u(k+1,j,i)   - u(k,j,i)    )                &
              -      (  5.0_wp * ibit17 * adv_mom_5                           &
                  +              ibit16 * adv_mom_3                           &
                     ) *                                                      &
                                ( u(k_pp,j,i)  - u(k-1,j,i)  )                &
              +      (           ibit17 * adv_mom_5                           &
                     ) *                                                      &
                                ( u(k_ppp,j,i) - u(k_mm,j,i) )                &
                                                           )
       ENDDO
       
       DO  k = nzb+1, nzb_max_l

          flux_d    = flux_t(k-1)
          diss_d    = diss_t(k-1)
          
          ibit11 = REAL( IBITS(advc_flags_1(k,j,i),11,1), KIND = wp )
          ibit10 = REAL( IBITS(advc_flags_1(k,j,i),10,1), KIND = wp )
          ibit9  = REAL( IBITS(advc_flags_1(k,j,i),9,1),  KIND = wp )
          
          ibit14 = REAL( IBITS(advc_flags_1(k,j,i),14,1), KIND = wp )
          ibit13 = REAL( IBITS(advc_flags_1(k,j,i),13,1), KIND = wp )
          ibit12 = REAL( IBITS(advc_flags_1(k,j,i),12,1), KIND = wp )
          
          ibit17 = REAL( IBITS(advc_flags_1(k,j,i),17,1), KIND = wp )
          ibit16 = REAL( IBITS(advc_flags_1(k,j,i),16,1), KIND = wp )
          ibit15 = REAL( IBITS(advc_flags_1(k,j,i),15,1), KIND = wp )
!
!--       Calculate the divergence of the velocity field. A respective
!--       correction is needed to overcome numerical instabilities introduced
!--       by a not sufficient reduction of divergences near topography.
          div = ( ( u_comp(k)       * ( ibit9 + ibit10 + ibit11 )             &
                - ( u(k,j,i)   + u(k,j,i-1)   )                               &
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k,j,i-1),9,1),  KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j,i-1),10,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j,i-1),11,1), KIND = wp )     &
                                      )                                       &
                  ) * ddx                                                     &
               +  ( ( v_comp(k) + gv ) * ( ibit12 + ibit13 + ibit14 )         &
                  - ( v(k,j,i)   + v(k,j,i-1 )  )                             &
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k,j-1,i),12,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j-1,i),13,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j-1,i),14,1), KIND = wp )     &
                                      )                                       &
                  ) * ddy                                                     &
               +  ( w_comp(k)   * rho_air_zw(k) * ( ibit15 + ibit16 + ibit17 )&
                -   w_comp(k-1) * rho_air_zw(k-1)                             &
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k-1,j,i),15,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k-1,j,i),16,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k-1,j,i),17,1), KIND = wp )     &
                                      )                                       &  
                  ) * drho_air(k) * ddzw(k)                                   &
                ) * 0.5_wp

          tend(k,j,i) = tend(k,j,i) - (                                       &
                            ( flux_r(k) + diss_r(k)                           &
                          -   flux_l_u(k,j,tn) - diss_l_u(k,j,tn) ) * ddx     &
                          + ( flux_n(k) + diss_n(k)                           &
                          -   flux_s_u(k,tn) - diss_s_u(k,tn)     ) * ddy     &
                          + ( ( flux_t(k) + diss_t(k) )                       &
                          -   ( flux_d    + diss_d    )                       &
                                                    ) * drho_air(k) * ddzw(k) &
                                       ) + div * u(k,j,i)

          flux_l_u(k,j,tn) = flux_r(k)
          diss_l_u(k,j,tn) = diss_r(k)
          flux_s_u(k,tn)   = flux_n(k)
          diss_s_u(k,tn)   = diss_n(k)
!
!--       Statistical Evaluation of u'u'. The factor has to be applied for
!--       right evaluation when gallilei_trans = .T. .
          sums_us2_ws_l(k,tn) = sums_us2_ws_l(k,tn)                            &
                + ( flux_r(k)                                                  &
                    * ( u_comp(k) - 2.0_wp * hom(k,1,1,0)                   )  &
                    / ( u_comp(k) - gu + SIGN( 1.0E-20_wp, u_comp(k) - gu ) )  &
                  + diss_r(k)                                                  &
                    *   ABS( u_comp(k) - 2.0_wp * hom(k,1,1,0)              )  &
                    / ( ABS( u_comp(k) - gu ) + 1.0E-20_wp                  )  &
                  ) *   weight_substep(intermediate_timestep_count)
!
!--       Statistical Evaluation of w'u'.
          sums_wsus_ws_l(k,tn) = sums_wsus_ws_l(k,tn)                          &
                + ( flux_t(k)                                                  &
                    * ( w_comp(k) - 2.0_wp * hom(k,1,3,0)                   )  &
                    / ( w_comp(k) + SIGN( 1.0E-20_wp, w_comp(k) )           )  &
                  + diss_t(k)                                                  &
                    *   ABS( w_comp(k) - 2.0_wp * hom(k,1,3,0)              )  &
                    / ( ABS( w_comp(k) ) + 1.0E-20_wp                       )  &
                  ) *   weight_substep(intermediate_timestep_count)
       ENDDO
       
       DO  k = nzb_max_l+1, nzt

          flux_d    = flux_t(k-1)
          diss_d    = diss_t(k-1)
!
!--       Calculate the divergence of the velocity field. A respective
!--       correction is needed to overcome numerical instabilities introduced
!--       by a not sufficient reduction of divergences near topography.
          div = ( ( u_comp(k)       - ( u(k,j,i)   + u(k,j,i-1) ) ) * ddx      &
               +  ( v_comp(k) + gv  - ( v(k,j,i)   + v(k,j,i-1) ) ) * ddy      &
               +  ( w_comp(k)   * rho_air_zw(k)                                &
                 -  w_comp(k-1) * rho_air_zw(k-1)                              &  
                  ) * drho_air(k) * ddzw(k)                                    &
                ) * 0.5_wp

          tend(k,j,i) = tend(k,j,i) - (                                        &
                            ( flux_r(k) + diss_r(k)                            &
                          -   flux_l_u(k,j,tn) - diss_l_u(k,j,tn) ) * ddx      &
                          + ( flux_n(k) + diss_n(k)                            &
                          -   flux_s_u(k,tn) - diss_s_u(k,tn)     ) * ddy      &
                          + ( ( flux_t(k) + diss_t(k) )                        &
                          -   ( flux_d    + diss_d    )                        &
                                                    ) * drho_air(k) * ddzw(k)  &
                                       ) + div * u(k,j,i)

          flux_l_u(k,j,tn) = flux_r(k)
          diss_l_u(k,j,tn) = diss_r(k)
          flux_s_u(k,tn)   = flux_n(k)
          diss_s_u(k,tn)   = diss_n(k)
!
!--       Statistical Evaluation of u'u'. The factor has to be applied for
!--       right evaluation when gallilei_trans = .T. .
          sums_us2_ws_l(k,tn) = sums_us2_ws_l(k,tn)                            &
                + ( flux_r(k)                                                  &
                    * ( u_comp(k) - 2.0_wp * hom(k,1,1,0)                   )  &
                    / ( u_comp(k) - gu + SIGN( 1.0E-20_wp, u_comp(k) - gu ) )  &
                  + diss_r(k)                                                  &
                    *   ABS( u_comp(k) - 2.0_wp * hom(k,1,1,0)              )  &
                    / ( ABS( u_comp(k) - gu ) + 1.0E-20_wp                  )  &
                  ) *   weight_substep(intermediate_timestep_count)
!
!--       Statistical Evaluation of w'u'.
          sums_wsus_ws_l(k,tn) = sums_wsus_ws_l(k,tn)                          &
                + ( flux_t(k)                                                  &
                    * ( w_comp(k) - 2.0_wp * hom(k,1,3,0)                   )  &
                    / ( w_comp(k) + SIGN( 1.0E-20_wp, w_comp(k) )           )  &
                  + diss_t(k)                                                  &
                    *   ABS( w_comp(k) - 2.0_wp * hom(k,1,3,0)              )  &
                    / ( ABS( w_comp(k) ) + 1.0E-20_wp                       )  &
                  ) *   weight_substep(intermediate_timestep_count)
       ENDDO



    END SUBROUTINE advec_u_ws_ij



!-----------------------------------------------------------------------------!
! Description:
! ------------
!> Advection of v-component - Call for grid point i,j
!-----------------------------------------------------------------------------!
   SUBROUTINE advec_v_ws_ij( i, j, i_omp, tn )

       USE arrays_3d,                                                          &
           ONLY:  ddzw, diss_l_v, diss_s_v, flux_l_v, flux_s_v, tend, u, v, w, &
                  drho_air, rho_air_zw

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s, intermediate_timestep_count, &
                  u_gtrans, v_gtrans

       USE grid_variables,                                                     &
           ONLY:  ddx, ddy

       USE indices,                                                            &
           ONLY:  nyn, nys, nysv, nxl, nxr, nzb, nzb_max, nzt, advc_flags_1

       USE kinds

       USE statistics,                                                         &
           ONLY:  hom, sums_vs2_ws_l, sums_wsvs_ws_l, weight_substep

       IMPLICIT NONE

       INTEGER(iwp)  ::  i         !< grid index along x-direction
       INTEGER(iwp)  ::  i_omp     !< leftmost index on subdomain, or in case of OpenMP, on thread
       INTEGER(iwp)  ::  j         !< grid index along y-direction
       INTEGER(iwp)  ::  k         !< grid index along z-direction
       INTEGER(iwp)  ::  k_mm      !< k-2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp)  ::  k_pp      !< k+2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp)  ::  k_ppp     !< k+3 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp)  ::  nzb_max_l !< index indicating upper bound for order degradation of horizontal advection terms  
       INTEGER(iwp)  ::  tn        !< number of OpenMP thread
       
       REAL(wp)      ::  ibit18   !< flag indicating 1st-order scheme along x-direction
       REAL(wp)      ::  ibit19   !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)      ::  ibit20   !< flag indicating 5th-order scheme along x-direction
       REAL(wp)      ::  ibit21   !< flag indicating 1st-order scheme along y-direction 
       REAL(wp)      ::  ibit22   !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)      ::  ibit23   !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)      ::  ibit24   !< flag indicating 1st-order scheme along z-direction
       REAL(wp)      ::  ibit25   !< flag indicating 3rd-order scheme along z-direction
       REAL(wp)      ::  ibit26   !< flag indicating 3rd-order scheme along z-direction
       REAL(wp)      ::  diss_d   !< artificial dissipation term at grid box bottom
       REAL(wp)      ::  div      !< divergence on v-grid
       REAL(wp)      ::  flux_d   !< 6th-order flux at grid box bottom
       REAL(wp)      ::  gu       !< Galilei-transformation velocity along x
       REAL(wp)      ::  gv       !< Galilei-transformation velocity along y
       REAL(wp)      ::  v_comp_l !< advection velocity along y on leftmost grid point on subdomain
       
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  diss_n !< discretized artificial dissipation at northward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  diss_r !< discretized artificial dissipation at rightward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  diss_t !< discretized artificial dissipation at top of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  flux_n !< discretized 6th-order flux at northward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  flux_r !< discretized 6th-order flux at rightward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  flux_t !< discretized 6th-order flux at top of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  u_comp !< advection velocity along x
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  v_comp !< advection velocity along y
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  w_comp !< advection velocity along z
!
!--    Used local modified copy of nzb_max (used to degrade order of 
!--    discretization) at non-cyclic boundaries. Modify only at relevant points
!--    instead of the entire subdomain. This should lead to better 
!--    load balance between boundary and non-boundary PEs.
       IF( ( bc_dirichlet_l  .OR.  bc_radiation_l )  .AND.  i <= nxl + 2  .OR. &
           ( bc_dirichlet_r  .OR.  bc_radiation_r )  .AND.  i >= nxr - 2  .OR. &
           ( bc_dirichlet_s  .OR.  bc_radiation_s )  .AND.  j <= nys + 2  .OR. &
           ( bc_dirichlet_n  .OR.  bc_radiation_n )  .AND.  j >= nyn - 2 )  THEN
          nzb_max_l = nzt
       ELSE
          nzb_max_l = nzb_max
       END IF
       
       gu = 2.0_wp * u_gtrans
       gv = 2.0_wp * v_gtrans

!       
!--    Compute leftside fluxes for the respective boundary.
       IF ( i == i_omp )  THEN

          DO  k = nzb+1, nzb_max_l

             ibit20 = REAL( IBITS(advc_flags_1(k,j,i-1),20,1), KIND = wp )
             ibit19 = REAL( IBITS(advc_flags_1(k,j,i-1),19,1), KIND = wp )
             ibit18 = REAL( IBITS(advc_flags_1(k,j,i-1),18,1), KIND = wp )

             u_comp(k)        = u(k,j-1,i) + u(k,j,i) - gu
             flux_l_v(k,j,tn) = u_comp(k) * (                                 &
                              ( 37.0_wp * ibit20 * adv_mom_5                  &
                           +     7.0_wp * ibit19 * adv_mom_3                  &
                           +              ibit18 * adv_mom_1                  &
                              ) *                                             &
                                        ( v(k,j,i)   + v(k,j,i-1) )           &
                       -      (  8.0_wp * ibit20 * adv_mom_5                  &
                           +              ibit19 * adv_mom_3                  &
                              ) *                                             &
                                        ( v(k,j,i+1) + v(k,j,i-2) )           &
                       +      (           ibit20 * adv_mom_5                  &
                              ) *                                             &
                                        ( v(k,j,i+2) + v(k,j,i-3) )           &
                                            )

              diss_l_v(k,j,tn) = - ABS( u_comp(k) ) * (                       &
                              ( 10.0_wp * ibit20 * adv_mom_5                  &
                           +     3.0_wp * ibit19 * adv_mom_3                  &
                           +              ibit18 * adv_mom_1                  &
                              ) *                                             &
                                        ( v(k,j,i)   - v(k,j,i-1) )           &
                       -      (  5.0_wp * ibit20 * adv_mom_5                  &
                           +              ibit19 * adv_mom_3                  &
                              ) *                                             &
                                        ( v(k,j,i+1) - v(k,j,i-2) )           &
                       +      (           ibit20 * adv_mom_5                  &
                              ) *                                             &
                                        ( v(k,j,i+2) - v(k,j,i-3) )           &
                                                      )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             u_comp(k)           = u(k,j-1,i) + u(k,j,i) - gu
             flux_l_v(k,j,tn) = u_comp(k) * (                                 &
                             37.0_wp * ( v(k,j,i) + v(k,j,i-1)   )            &
                           -  8.0_wp * ( v(k,j,i+1) + v(k,j,i-2) )            &
                           +           ( v(k,j,i+2) + v(k,j,i-3) ) ) * adv_mom_5
             diss_l_v(k,j,tn) = - ABS( u_comp(k) ) * (                        &
                             10.0_wp * ( v(k,j,i) - v(k,j,i-1)   )            &
                           -  5.0_wp * ( v(k,j,i+1) - v(k,j,i-2) )            &
                           +           ( v(k,j,i+2) - v(k,j,i-3) ) ) * adv_mom_5

          ENDDO
          
       ENDIF
!
!--    Compute southside fluxes for the respective boundary.
       IF ( j == nysv )  THEN
       
          DO  k = nzb+1, nzb_max_l

             ibit23 = REAL( IBITS(advc_flags_1(k,j-1,i),23,1), KIND = wp )
             ibit22 = REAL( IBITS(advc_flags_1(k,j-1,i),22,1), KIND = wp )
             ibit21 = REAL( IBITS(advc_flags_1(k,j-1,i),21,1), KIND = wp )

             v_comp_l       = v(k,j,i) + v(k,j-1,i) - gv
             flux_s_v(k,tn) = v_comp_l * (                                    &
                            ( 37.0_wp * ibit23 * adv_mom_5                    &
                         +     7.0_wp * ibit22 * adv_mom_3                    &
                         +              ibit21 * adv_mom_1                    &
                            ) *                                               &
                                        ( v(k,j,i)   + v(k,j-1,i) )           &
                     -      (  8.0_wp * ibit23 * adv_mom_5                    &
                         +              ibit22 * adv_mom_3                    &
                            ) *                                               &
                                        ( v(k,j+1,i) + v(k,j-2,i) )           &
                     +      (           ibit23 * adv_mom_5                    &
                            ) *                                               &
                                        ( v(k,j+2,i) + v(k,j-3,i) )           &
                                         )

             diss_s_v(k,tn) = - ABS( v_comp_l ) * (                           &
                            ( 10.0_wp * ibit23 * adv_mom_5                    &
                         +     3.0_wp * ibit22 * adv_mom_3                    &
                         +              ibit21 * adv_mom_1                    &
                            ) *                                               &
                                        ( v(k,j,i)   - v(k,j-1,i) )           &
                     -      (  5.0_wp * ibit23 * adv_mom_5                    &
                         +              ibit22 * adv_mom_3                    &
                            ) *                                               &
                                        ( v(k,j+1,i) - v(k,j-2,i) )           &
                     +      (           ibit23 * adv_mom_5                    &
                            ) *                                               &
                                        ( v(k,j+2,i) - v(k,j-3,i) )           &
                                                  )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             v_comp_l       = v(k,j,i) + v(k,j-1,i) - gv
             flux_s_v(k,tn) = v_comp_l * (                                    &
                           37.0_wp * ( v(k,j,i) + v(k,j-1,i)   )              &
                         -  8.0_wp * ( v(k,j+1,i) + v(k,j-2,i) )              &
                         +           ( v(k,j+2,i) + v(k,j-3,i) ) ) * adv_mom_5
             diss_s_v(k,tn) = - ABS( v_comp_l ) * (                           &
                           10.0_wp * ( v(k,j,i) - v(k,j-1,i)   )              &
                         -  5.0_wp * ( v(k,j+1,i) - v(k,j-2,i) )              &
                         +           ( v(k,j+2,i) - v(k,j-3,i) ) ) * adv_mom_5

          ENDDO
          
       ENDIF
!
!--    Now compute the fluxes and tendency terms for the horizontal and
!--    verical parts.
       DO  k = nzb+1, nzb_max_l

          ibit20 = REAL( IBITS(advc_flags_1(k,j,i),20,1), KIND = wp )
          ibit19 = REAL( IBITS(advc_flags_1(k,j,i),19,1), KIND = wp )
          ibit18 = REAL( IBITS(advc_flags_1(k,j,i),18,1), KIND = wp )
 
          u_comp(k) = u(k,j-1,i+1) + u(k,j,i+1) - gu
          flux_r(k) = u_comp(k) * (                                           &
                     ( 37.0_wp * ibit20 * adv_mom_5                           &
                  +     7.0_wp * ibit19 * adv_mom_3                           &
                  +              ibit18 * adv_mom_1                           &
                     ) *                                                      &
                                    ( v(k,j,i+1) + v(k,j,i)   )               &
              -      (  8.0_wp * ibit20 * adv_mom_5                           &
                  +              ibit19 * adv_mom_3                           &
                     ) *                                                      &
                                    ( v(k,j,i+2) + v(k,j,i-1) )               &
              +      (           ibit20 * adv_mom_5                           &
                     ) *                                                      &
                                    ( v(k,j,i+3) + v(k,j,i-2) )               &
                                  )

          diss_r(k) = - ABS( u_comp(k) ) * (                                  &
                     ( 10.0_wp * ibit20 * adv_mom_5                           &
                  +     3.0_wp * ibit19 * adv_mom_3                           &
                  +              ibit18 * adv_mom_1                           &
                     ) *                                                      &
                                    ( v(k,j,i+1) - v(k,j,i)  )                &
              -      (  5.0_wp * ibit20 * adv_mom_5                           &
                  +              ibit19 * adv_mom_3                           &
                     ) *                                                      &
                                    ( v(k,j,i+2) - v(k,j,i-1) )               &
              +      (           ibit20 * adv_mom_5                           &
                     ) *                                                      &
                                    ( v(k,j,i+3) - v(k,j,i-2) )               &
                                           )

          ibit23 = REAL( IBITS(advc_flags_1(k,j,i),23,1), KIND = wp )
          ibit22 = REAL( IBITS(advc_flags_1(k,j,i),22,1), KIND = wp )
          ibit21 = REAL( IBITS(advc_flags_1(k,j,i),21,1), KIND = wp )


          v_comp(k) = v(k,j+1,i) + v(k,j,i)
          flux_n(k) = ( v_comp(k) - gv ) * (                                  &
                     ( 37.0_wp * ibit23 * adv_mom_5                           &
                  +     7.0_wp * ibit22 * adv_mom_3                           &
                  +              ibit21 * adv_mom_1                           &
                     ) *                                                      &
                                    ( v(k,j+1,i) + v(k,j,i)   )               &
              -      (  8.0_wp * ibit23 * adv_mom_5                           &
                  +              ibit22 * adv_mom_3                           &
                     ) *                                                      &
                                    ( v(k,j+2,i) + v(k,j-1,i) )               &
              +      (           ibit23 * adv_mom_5                           &
                     ) *                                                      &
                                    ( v(k,j+3,i) + v(k,j-2,i) )               &
                                           )

          diss_n(k) = - ABS( v_comp(k) - gv ) * (                             &
                     ( 10.0_wp * ibit23 * adv_mom_5                           &
                  +     3.0_wp * ibit22 * adv_mom_3                           &
                  +              ibit21 * adv_mom_1                           &
                     ) *                                                      &
                                    ( v(k,j+1,i) - v(k,j,i)   )               &
              -      (  5.0_wp * ibit23 * adv_mom_5                           &
                  +              ibit22 * adv_mom_3                           &
                     ) *                                                      &
                                    ( v(k,j+2,i) - v(k,j-1,i) )               &
              +      (           ibit23 * adv_mom_5                           &
                     ) *                                                      &
                                    ( v(k,j+3,i) - v(k,j-2,i) )               &
                                                )
       ENDDO

       DO  k = nzb_max_l+1, nzt

          u_comp(k) = u(k,j-1,i+1) + u(k,j,i+1) - gu
          flux_r(k) = u_comp(k) * (                                           &
                      37.0_wp * ( v(k,j,i+1) + v(k,j,i)   )                   &
                    -  8.0_wp * ( v(k,j,i+2) + v(k,j,i-1) )                   &
                    +           ( v(k,j,i+3) + v(k,j,i-2) ) ) * adv_mom_5

          diss_r(k) = - ABS( u_comp(k) ) * (                                  &
                      10.0_wp * ( v(k,j,i+1) - v(k,j,i) )                     &
                    -  5.0_wp * ( v(k,j,i+2) - v(k,j,i-1) )                   &
                    +           ( v(k,j,i+3) - v(k,j,i-2) ) ) * adv_mom_5


          v_comp(k) = v(k,j+1,i) + v(k,j,i)
          flux_n(k) = ( v_comp(k) - gv ) * (                                  &
                      37.0_wp * ( v(k,j+1,i) + v(k,j,i)   )                   &
                    -  8.0_wp * ( v(k,j+2,i) + v(k,j-1,i) )                   &
                      +         ( v(k,j+3,i) + v(k,j-2,i) ) ) * adv_mom_5

          diss_n(k) = - ABS( v_comp(k) - gv ) * (                             &
                      10.0_wp * ( v(k,j+1,i) - v(k,j,i)   )                   &
                    -  5.0_wp * ( v(k,j+2,i) - v(k,j-1,i) )                   &
                    +           ( v(k,j+3,i) - v(k,j-2,i) ) ) * adv_mom_5
       ENDDO
!
!--    Now, compute vertical fluxes. Split loop into a part treating the 
!--    lowest 2 grid points with indirect indexing, a main loop without
!--    indirect indexing, and a loop for the uppermost 2 grip points with
!--    indirect indexing. This allows better vectorization for the main loop.
!--    First, compute the flux at model surface, which need has to be 
!--    calculated explicetely for the tendency at
!--    the first w-level. For topography wall this is done implicitely by
!--    advc_flags_1.
       flux_t(nzb) = 0.0_wp
       diss_t(nzb) = 0.0_wp
       w_comp(nzb) = 0.0_wp
       
       DO  k = nzb+1, nzb+2
!
!--       k index has to be modified near bottom and top, else array
!--       subscripts will be exceeded.
          ibit26 = REAL( IBITS(advc_flags_1(k,j,i),26,1), KIND = wp )
          ibit25 = REAL( IBITS(advc_flags_1(k,j,i),25,1), KIND = wp )
          ibit24 = REAL( IBITS(advc_flags_1(k,j,i),24,1), KIND = wp )

          k_ppp = k + 3 * ibit26
          k_pp  = k + 2 * ( 1 - ibit24  )
          k_mm  = k - 2 * ibit26

          w_comp(k) = w(k,j-1,i) + w(k,j,i)
          flux_t(k) = w_comp(k) * rho_air_zw(k) * (                           &
                     ( 37.0_wp * ibit26 * adv_mom_5                           &
                  +     7.0_wp * ibit25 * adv_mom_3                           &
                  +              ibit24 * adv_mom_1                           &
                     ) *                                                      &
                                ( v(k+1,j,i)   + v(k,j,i)    )                &
              -      (  8.0_wp * ibit26 * adv_mom_5                           &
                  +              ibit25 * adv_mom_3                           &
                     ) *                                                      &
                                ( v(k_pp,j,i)  + v(k-1,j,i)  )                &
              +      (           ibit26 * adv_mom_5                           &
                     ) *                                                      &
                                ( v(k_ppp,j,i) + v(k_mm,j,i) )                &
                                                  )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air_zw(k) * (                  &
                     ( 10.0_wp * ibit26 * adv_mom_5                           &
                  +     3.0_wp * ibit25 * adv_mom_3                           &
                  +              ibit24 * adv_mom_1                           &
                     ) *                                                      &
                                ( v(k+1,j,i)   - v(k,j,i)    )                &
              -      (  5.0_wp * ibit26 * adv_mom_5                           &
                  +              ibit25 * adv_mom_3                           &
                     ) *                                                      &
                                ( v(k_pp,j,i)  - v(k-1,j,i)  )                &
              +      (           ibit26 * adv_mom_5                           &
                     ) *                                                      &
                                ( v(k_ppp,j,i) - v(k_mm,j,i) )                &
                                                           )
       ENDDO
       
       DO  k = nzb+3, nzt-2

          ibit26 = REAL( IBITS(advc_flags_1(k,j,i),26,1), KIND = wp )
          ibit25 = REAL( IBITS(advc_flags_1(k,j,i),25,1), KIND = wp )
          ibit24 = REAL( IBITS(advc_flags_1(k,j,i),24,1), KIND = wp )

          w_comp(k) = w(k,j-1,i) + w(k,j,i)
          flux_t(k) = w_comp(k) * rho_air_zw(k) * (                           &
                     ( 37.0_wp * ibit26 * adv_mom_5                           &
                  +     7.0_wp * ibit25 * adv_mom_3                           &
                  +              ibit24 * adv_mom_1                           &
                     ) *                                                      &
                                ( v(k+1,j,i)   + v(k,j,i)    )                &
              -      (  8.0_wp * ibit26 * adv_mom_5                           &
                  +              ibit25 * adv_mom_3                           &
                     ) *                                                      &
                                ( v(k+2,j,i)  + v(k-1,j,i)  )                 &
              +      (           ibit26 * adv_mom_5                           &
                     ) *                                                      &
                                ( v(k+3,j,i) + v(k-2,j,i) )                   &
                                                  )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air_zw(k) * (                  &
                     ( 10.0_wp * ibit26 * adv_mom_5                           &
                  +     3.0_wp * ibit25 * adv_mom_3                           &
                  +              ibit24 * adv_mom_1                           &
                     ) *                                                      &
                                ( v(k+1,j,i)   - v(k,j,i)    )                &
              -      (  5.0_wp * ibit26 * adv_mom_5                           &
                  +              ibit25 * adv_mom_3                           &
                     ) *                                                      &
                                ( v(k+2,j,i)  - v(k-1,j,i)  )                 &
              +      (           ibit26 * adv_mom_5                           &
                     ) *                                                      &
                                ( v(k+3,j,i) - v(k-2,j,i) )                   &
                                                           )
       ENDDO
       
       DO  k = nzt-1, nzt
!
!--       k index has to be modified near bottom and top, else array
!--       subscripts will be exceeded.
          ibit26 = REAL( IBITS(advc_flags_1(k,j,i),26,1), KIND = wp )
          ibit25 = REAL( IBITS(advc_flags_1(k,j,i),25,1), KIND = wp )
          ibit24 = REAL( IBITS(advc_flags_1(k,j,i),24,1), KIND = wp )

          k_ppp = k + 3 * ibit26
          k_pp  = k + 2 * ( 1 - ibit24  )
          k_mm  = k - 2 * ibit26

          w_comp(k) = w(k,j-1,i) + w(k,j,i)
          flux_t(k) = w_comp(k) * rho_air_zw(k) * (                           &
                     ( 37.0_wp * ibit26 * adv_mom_5                           &
                  +     7.0_wp * ibit25 * adv_mom_3                           &
                  +              ibit24 * adv_mom_1                           &
                     ) *                                                      &
                                ( v(k+1,j,i)   + v(k,j,i)    )                &
              -      (  8.0_wp * ibit26 * adv_mom_5                           &
                  +              ibit25 * adv_mom_3                           &
                     ) *                                                      &
                                ( v(k_pp,j,i)  + v(k-1,j,i)  )                &
              +      (           ibit26 * adv_mom_5                           &
                     ) *                                                      &
                                ( v(k_ppp,j,i) + v(k_mm,j,i) )                &
                                                  )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air_zw(k) * (                  &
                     ( 10.0_wp * ibit26 * adv_mom_5                           &
                  +     3.0_wp * ibit25 * adv_mom_3                           &
                  +              ibit24 * adv_mom_1                           &
                     ) *                                                      &
                                ( v(k+1,j,i)   - v(k,j,i)    )                &
              -      (  5.0_wp * ibit26 * adv_mom_5                           &
                  +              ibit25 * adv_mom_3                           &
                     ) *                                                      &
                                ( v(k_pp,j,i)  - v(k-1,j,i)  )                &
              +      (           ibit26 * adv_mom_5                           &
                     ) *                                                      &
                                ( v(k_ppp,j,i) - v(k_mm,j,i) )                &
                                                           )
       ENDDO
       
       DO  k = nzb+1, nzb_max_l

          flux_d    = flux_t(k-1)
          diss_d    = diss_t(k-1)
          
          ibit20 = REAL( IBITS(advc_flags_1(k,j,i),20,1), KIND = wp )
          ibit19 = REAL( IBITS(advc_flags_1(k,j,i),19,1), KIND = wp )
          ibit18 = REAL( IBITS(advc_flags_1(k,j,i),18,1), KIND = wp )
          
          ibit23 = REAL( IBITS(advc_flags_1(k,j,i),23,1), KIND = wp )
          ibit22 = REAL( IBITS(advc_flags_1(k,j,i),22,1), KIND = wp )
          ibit21 = REAL( IBITS(advc_flags_1(k,j,i),21,1), KIND = wp )
          
          ibit26 = REAL( IBITS(advc_flags_1(k,j,i),26,1), KIND = wp )
          ibit25 = REAL( IBITS(advc_flags_1(k,j,i),25,1), KIND = wp )
          ibit24 = REAL( IBITS(advc_flags_1(k,j,i),24,1), KIND = wp )
!
!--       Calculate the divergence of the velocity field. A respective
!--       correction is needed to overcome numerical instabilities introduced
!--       by a not sufficient reduction of divergences near topography.
          div = ( ( ( u_comp(k)     + gu )                                    &
                                       * ( ibit18 + ibit19 + ibit20 )         &
                  - ( u(k,j-1,i) + u(k,j,i) )                                 &
                                       * (                                    &
                        REAL( IBITS(advc_flags_1(k,j,i-1),18,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k,j,i-1),19,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k,j,i-1),20,1), KIND = wp )  &
                                         )                                    &
                  ) * ddx                                                     &
               +  ( v_comp(k)                                                 &
                                       * ( ibit21 + ibit22 + ibit23 )         &
                - ( v(k,j,i)     + v(k,j-1,i) )                               &
                                       * (                                    &
                        REAL( IBITS(advc_flags_1(k,j-1,i),21,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k,j-1,i),22,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k,j-1,i),23,1), KIND = wp )  &
                                         )                                    &
                  ) * ddy                                                     &
               +  ( w_comp(k)   * rho_air_zw(k) * ( ibit24 + ibit25 + ibit26 )&
                -   w_comp(k-1) * rho_air_zw(k-1)                             &
                                       * (                                    &
                        REAL( IBITS(advc_flags_1(k-1,j,i),24,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k-1,j,i),25,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k-1,j,i),26,1), KIND = wp )  &
                                         )                                    &
                   ) * drho_air(k) * ddzw(k)                                  &
                ) * 0.5_wp

          tend(k,j,i) = tend(k,j,i) - (                                       &
                         ( flux_r(k) + diss_r(k)                              &
                       -   flux_l_v(k,j,tn) - diss_l_v(k,j,tn)   ) * ddx      &
                       + ( flux_n(k) + diss_n(k)                              &
                       -   flux_s_v(k,tn) - diss_s_v(k,tn)       ) * ddy      &
                       + ( ( flux_t(k) + diss_t(k) )                          &
                       -   ( flux_d    + diss_d    )                          &
                                                   ) * drho_air(k) * ddzw(k)  &
                                      ) + v(k,j,i) * div

          flux_l_v(k,j,tn) = flux_r(k)
          diss_l_v(k,j,tn) = diss_r(k)
          flux_s_v(k,tn)   = flux_n(k)
          diss_s_v(k,tn)   = diss_n(k)
!
!--       Statistical Evaluation of v'v'. The factor has to be applied for
!--       right evaluation when gallilei_trans = .T. .
          sums_vs2_ws_l(k,tn) = sums_vs2_ws_l(k,tn)                            &
                + ( flux_n(k)                                                  &
                    * ( v_comp(k) - 2.0_wp * hom(k,1,2,0)                   )  &
                    / ( v_comp(k) - gv + SIGN( 1.0E-20_wp, v_comp(k) - gv ) )  &
                  + diss_n(k)                                                  &
                    *   ABS( v_comp(k) - 2.0_wp * hom(k,1,2,0)              )  &
                    / ( ABS( v_comp(k) - gv ) + 1.0E-20_wp                  )  &
                  ) *   weight_substep(intermediate_timestep_count)
!
!--       Statistical Evaluation of w'u'.
          sums_wsvs_ws_l(k,tn) = sums_wsvs_ws_l(k,tn)                          &
                + ( flux_t(k)                                                  &
                    * ( w_comp(k) - 2.0_wp * hom(k,1,3,0)                   )  &
                    / ( w_comp(k) + SIGN( 1.0E-20_wp, w_comp(k) )           )  &
                  + diss_t(k)                                                  &
                    *   ABS( w_comp(k) - 2.0_wp * hom(k,1,3,0)              )  &
                    / ( ABS( w_comp(k) ) + 1.0E-20_wp                       )  &
                  ) *   weight_substep(intermediate_timestep_count)

       ENDDO
       
       DO  k = nzb_max_l+1, nzt

          flux_d    = flux_t(k-1)
          diss_d    = diss_t(k-1)
!
!--       Calculate the divergence of the velocity field. A respective
!--       correction is needed to overcome numerical instabilities introduced
!--       by a not sufficient reduction of divergences near topography.
          div = ( ( u_comp(k) + gu - ( u(k,j-1,i) + u(k,j,i)   ) ) * ddx       &
               +  ( v_comp(k)      - ( v(k,j,i)   + v(k,j-1,i) ) ) * ddy       &
               +  ( w_comp(k)   * rho_air_zw(k)                                &
                 -  w_comp(k-1) * rho_air_zw(k-1)                              &
                  ) * drho_air(k) * ddzw(k)                                    &
                ) * 0.5_wp

          tend(k,j,i) = tend(k,j,i) - (                                        &
                         ( flux_r(k) + diss_r(k)                               &
                       -   flux_l_v(k,j,tn) - diss_l_v(k,j,tn)   ) * ddx       &
                       + ( flux_n(k) + diss_n(k)                               &
                       -   flux_s_v(k,tn) - diss_s_v(k,tn)       ) * ddy       &
                       + ( ( flux_t(k) + diss_t(k) )                           &
                       -   ( flux_d    + diss_d    )                           &
                                                   ) * drho_air(k) * ddzw(k)   &
                                      ) + v(k,j,i) * div

          flux_l_v(k,j,tn) = flux_r(k)
          diss_l_v(k,j,tn) = diss_r(k)
          flux_s_v(k,tn)   = flux_n(k)
          diss_s_v(k,tn)   = diss_n(k)
!
!--       Statistical Evaluation of v'v'. The factor has to be applied for
!--       right evaluation when gallilei_trans = .T. .
          sums_vs2_ws_l(k,tn) = sums_vs2_ws_l(k,tn)                            &
                + ( flux_n(k)                                                  &
                    * ( v_comp(k) - 2.0_wp * hom(k,1,2,0)                   )  &
                    / ( v_comp(k) - gv + SIGN( 1.0E-20_wp, v_comp(k) - gv ) )  &
                  + diss_n(k)                                                  &
                    *   ABS( v_comp(k) - 2.0_wp * hom(k,1,2,0)              )  &
                    / ( ABS( v_comp(k) - gv ) + 1.0E-20_wp                  )  &
                  ) *   weight_substep(intermediate_timestep_count)
!
!--       Statistical Evaluation of w'u'.
          sums_wsvs_ws_l(k,tn) = sums_wsvs_ws_l(k,tn)                          &
                + ( flux_t(k)                                                  &
                    * ( w_comp(k) - 2.0_wp * hom(k,1,3,0)                   )  &
                    / ( w_comp(k) + SIGN( 1.0E-20_wp, w_comp(k) )           )  &
                  + diss_t(k)                                                  &
                    *   ABS( w_comp(k) - 2.0_wp * hom(k,1,3,0)              )  &
                    / ( ABS( w_comp(k) ) + 1.0E-20_wp                       )  &
                  ) *   weight_substep(intermediate_timestep_count)

       ENDDO


    END SUBROUTINE advec_v_ws_ij



!------------------------------------------------------------------------------!
! Description:
! ------------
!> Advection of w-component - Call for grid point i,j
!------------------------------------------------------------------------------!
    SUBROUTINE advec_w_ws_ij( i, j, i_omp, tn )

       USE arrays_3d,                                                          &
           ONLY:  ddzu, diss_l_w, diss_s_w, flux_l_w, flux_s_w, tend, u, v, w, &
                  drho_air_zw, rho_air

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s,intermediate_timestep_count,  &
                  u_gtrans, v_gtrans

       USE grid_variables,                                                     &
           ONLY:  ddx, ddy

       USE indices,                                                            &
           ONLY:  nyn, nys, nxl, nxr, nzb, nzb_max, nzt, advc_flags_1,         &
                  advc_flags_2

       USE kinds
       
       USE statistics,                                                         &
           ONLY:  hom, sums_ws2_ws_l, weight_substep

       IMPLICIT NONE

       INTEGER(iwp) ::  i         !< grid index along x-direction
       INTEGER(iwp) ::  i_omp     !< leftmost index on subdomain, or in case of OpenMP, on thread
       INTEGER(iwp) ::  j         !< grid index along y-direction
       INTEGER(iwp) ::  k         !< grid index along z-direction
       INTEGER(iwp) ::  k_mm      !< k-2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_pp      !< k+2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_ppp     !< k+3 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  nzb_max_l !< index indicating upper bound for order degradation of horizontal advection terms  
       INTEGER(iwp) ::  tn        !< number of OpenMP thread
       
       REAL(wp)    ::  ibit27  !< flag indicating 1st-order scheme along x-direction
       REAL(wp)    ::  ibit28  !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)    ::  ibit29  !< flag indicating 5th-order scheme along x-direction
       REAL(wp)    ::  ibit30  !< flag indicating 1st-order scheme along y-direction
       REAL(wp)    ::  ibit31  !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)    ::  ibit32  !< flag indicating 5th-order scheme along y-direction
       REAL(wp)    ::  ibit33  !< flag indicating 1st-order scheme along z-direction
       REAL(wp)    ::  ibit34  !< flag indicating 3rd-order scheme along z-direction
       REAL(wp)    ::  ibit35  !< flag indicating 5th-order scheme along z-direction
       REAL(wp)    ::  diss_d  !< discretized artificial dissipation at top of the grid box
       REAL(wp)    ::  div     !< divergence on w-grid
       REAL(wp)    ::  flux_d  !< discretized 6th-order flux at top of the grid box
       REAL(wp)    ::  gu      !< Galilei-transformation velocity along x
       REAL(wp)    ::  gv      !< Galilei-transformation velocity along y
       
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  diss_n !< discretized artificial dissipation at northward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  diss_r !< discretized artificial dissipation at rightward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  diss_t !< discretized artificial dissipation at top of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  flux_n !< discretized 6th-order flux at northward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  flux_r !< discretized 6th-order flux at rightward-side of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  flux_t !< discretized 6th-order flux at top of the grid box
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  u_comp !< advection velocity along x
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  v_comp !< advection velocity along y
       REAL(wp), DIMENSION(nzb:nzt+1)  ::  w_comp !< advection velocity along z
!
!--    Used local modified copy of nzb_max (used to degrade order of 
!--    discretization) at non-cyclic boundaries. Modify only at relevant points
!--    instead of the entire subdomain. This should lead to better 
!--    load balance between boundary and non-boundary PEs.
       IF( ( bc_dirichlet_l  .OR.  bc_radiation_l )  .AND.  i <= nxl + 2  .OR. &
           ( bc_dirichlet_r  .OR.  bc_radiation_r )  .AND.  i >= nxr - 2  .OR. &
           ( bc_dirichlet_s  .OR.  bc_radiation_s )  .AND.  j <= nys + 2  .OR. &
           ( bc_dirichlet_n  .OR.  bc_radiation_n )  .AND.  j >= nyn - 2 )  THEN
          nzb_max_l = nzt
       ELSE
          nzb_max_l = nzb_max
       END IF
       
       gu = 2.0_wp * u_gtrans
       gv = 2.0_wp * v_gtrans
!
!--    Compute southside fluxes for the respective boundary.
       IF ( j == nys )  THEN

          DO  k = nzb+1, nzb_max_l
             ibit32 = REAL( IBITS(advc_flags_2(k,j-1,i),0,1),  KIND = wp )
             ibit31 = REAL( IBITS(advc_flags_1(k,j-1,i),31,1), KIND = wp )
             ibit30 = REAL( IBITS(advc_flags_1(k,j-1,i),30,1), KIND = wp )

             v_comp(k)      = v(k+1,j,i) + v(k,j,i) - gv
             flux_s_w(k,tn) = v_comp(k) * (                                   &
                            ( 37.0_wp * ibit32 * adv_mom_5                    &
                         +     7.0_wp * ibit31 * adv_mom_3                    &
                         +              ibit30 * adv_mom_1                    &
                            ) *                                               &
                                        ( w(k,j,i)   + w(k,j-1,i) )           &
                     -      (  8.0_wp * ibit32 * adv_mom_5                    &
                         +              ibit31 * adv_mom_3                    &
                            ) *                                               &
                                        ( w(k,j+1,i) + w(k,j-2,i) )           &
                     +      (           ibit32 * adv_mom_5                    &
                            ) *                                               &
                                        ( w(k,j+2,i) + w(k,j-3,i) )           &
                                          )

             diss_s_w(k,tn) = - ABS( v_comp(k) ) * (                          &
                            ( 10.0_wp * ibit32 * adv_mom_5                    &
                         +     3.0_wp * ibit31 * adv_mom_3                    &
                         +              ibit30 * adv_mom_1                    &
                            ) *                                               &
                                        ( w(k,j,i)   - w(k,j-1,i) )           &
                     -      (  5.0_wp * ibit32 * adv_mom_5                    &
                         +              ibit31 * adv_mom_3                    &
                            ) *                                               &
                                        ( w(k,j+1,i) - w(k,j-2,i) )           &
                     +      (           ibit32 * adv_mom_5                    &
                            ) *                                               &
                                        ( w(k,j+2,i) - w(k,j-3,i) )           &
                                                   )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             v_comp(k)      = v(k+1,j,i) + v(k,j,i) - gv
             flux_s_w(k,tn) = v_comp(k) * (                                   &
                           37.0_wp * ( w(k,j,i) + w(k,j-1,i)   )              &
                         -  8.0_wp * ( w(k,j+1,i) +w(k,j-2,i)  )              &
                         +           ( w(k,j+2,i) + w(k,j-3,i) ) ) * adv_mom_5
             diss_s_w(k,tn) = - ABS( v_comp(k) ) * (                          &
                           10.0_wp * ( w(k,j,i) - w(k,j-1,i)   )              &
                         -  5.0_wp * ( w(k,j+1,i) - w(k,j-2,i) )              &
                         +           ( w(k,j+2,i) - w(k,j-3,i) ) ) * adv_mom_5

          ENDDO

       ENDIF
!
!--    Compute leftside fluxes for the respective boundary.
       IF ( i == i_omp ) THEN

          DO  k = nzb+1, nzb_max_l

             ibit29 = REAL( IBITS(advc_flags_1(k,j,i-1),29,1), KIND = wp )
             ibit28 = REAL( IBITS(advc_flags_1(k,j,i-1),28,1), KIND = wp )
             ibit27 = REAL( IBITS(advc_flags_1(k,j,i-1),27,1), KIND = wp )

             u_comp(k)        = u(k+1,j,i) + u(k,j,i) - gu
             flux_l_w(k,j,tn) = u_comp(k) * (                                 &
                             ( 37.0_wp * ibit29 * adv_mom_5                   &
                          +     7.0_wp * ibit28 * adv_mom_3                   &
                          +              ibit27 * adv_mom_1                   &
                             ) *                                              &
                                        ( w(k,j,i)   + w(k,j,i-1) )           &
                      -      (  8.0_wp * ibit29 * adv_mom_5                   &
                          +              ibit28 * adv_mom_3                   &
                             ) *                                              &
                                        ( w(k,j,i+1) + w(k,j,i-2) )           &
                      +      (           ibit29 * adv_mom_5                   &
                             ) *                                              &
                                        ( w(k,j,i+2) + w(k,j,i-3) )           &
                                            )

               diss_l_w(k,j,tn) = - ABS( u_comp(k) ) * (                      &
                             ( 10.0_wp * ibit29 * adv_mom_5                   &
                          +     3.0_wp * ibit28 * adv_mom_3                   &
                          +              ibit27 * adv_mom_1                   &
                             ) *                                              &
                                        ( w(k,j,i)   - w(k,j,i-1) )           &
                      -      (  5.0_wp * ibit29 * adv_mom_5                   &
                          +              ibit28 * adv_mom_3                   &
                             ) *                                              &
                                        ( w(k,j,i+1) - w(k,j,i-2) )           &
                      +      (           ibit29 * adv_mom_5                   &
                             ) *                                              &
                                        ( w(k,j,i+2) - w(k,j,i-3) )           &
                                                       )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             u_comp(k)        = u(k+1,j,i) + u(k,j,i) - gu
             flux_l_w(k,j,tn) = u_comp(k) * (                                 &
                            37.0_wp * ( w(k,j,i) + w(k,j,i-1)   )             &
                          -  8.0_wp * ( w(k,j,i+1) + w(k,j,i-2) )             &
                          +           ( w(k,j,i+2) + w(k,j,i-3) ) ) * adv_mom_5
             diss_l_w(k,j,tn) = - ABS( u_comp(k) ) * (                        &
                            10.0_wp * ( w(k,j,i) - w(k,j,i-1)   )             &
                          -  5.0_wp * ( w(k,j,i+1) - w(k,j,i-2) )             &
                          +           ( w(k,j,i+2) - w(k,j,i-3) ) ) * adv_mom_5 

          ENDDO

       ENDIF
!
!--    Now compute the fluxes and tendency terms for the horizontal
!--    and vertical parts.
       DO  k = nzb+1, nzb_max_l

          ibit29 = REAL( IBITS(advc_flags_1(k,j,i),29,1), KIND = wp )
          ibit28 = REAL( IBITS(advc_flags_1(k,j,i),28,1), KIND = wp )
          ibit27 = REAL( IBITS(advc_flags_1(k,j,i),27,1), KIND = wp )

          u_comp(k) = u(k+1,j,i+1) + u(k,j,i+1) - gu
          flux_r(k) = u_comp(k) * (                                           &
                     ( 37.0_wp * ibit29 * adv_mom_5                           &
                  +     7.0_wp * ibit28 * adv_mom_3                           &
                  +              ibit27 * adv_mom_1                           &
                     ) *                                                      &
                                    ( w(k,j,i+1) + w(k,j,i)   )               &
              -      (  8.0_wp * ibit29 * adv_mom_5                           &
                  +              ibit28 * adv_mom_3                           &
                     ) *                                                      &
                                    ( w(k,j,i+2) + w(k,j,i-1) )               &
              +      (           ibit29 * adv_mom_5                           &
                     ) *                                                      &
                                    ( w(k,j,i+3) + w(k,j,i-2) )               &
                                  )

          diss_r(k) = - ABS( u_comp(k) ) * (                                  &
                     ( 10.0_wp * ibit29 * adv_mom_5                           &
                  +     3.0_wp * ibit28 * adv_mom_3                           &
                  +              ibit27 * adv_mom_1                           &
                     ) *                                                      &
                                    ( w(k,j,i+1) - w(k,j,i)   )               &
              -      (  5.0_wp * ibit29 * adv_mom_5                           &
                  +              ibit28 * adv_mom_3                           &
                     ) *                                                      &
                                    ( w(k,j,i+2) - w(k,j,i-1) )               &
              +      (           ibit29 * adv_mom_5                           &
                     ) *                                                      &
                                    ( w(k,j,i+3) - w(k,j,i-2) )               &
                                           )

          ibit32 = REAL( IBITS(advc_flags_2(k,j,i),0,1),  KIND = wp )
          ibit31 = REAL( IBITS(advc_flags_1(k,j,i),31,1), KIND = wp )
          ibit30 = REAL( IBITS(advc_flags_1(k,j,i),30,1), KIND = wp )

          v_comp(k) = v(k+1,j+1,i) + v(k,j+1,i) - gv
          flux_n(k) = v_comp(k) * (                                           &
                     ( 37.0_wp * ibit32 * adv_mom_5                           &
                  +     7.0_wp * ibit31 * adv_mom_3                           &
                  +              ibit30 * adv_mom_1                           &
                     ) *                                                      &
                                    ( w(k,j+1,i) + w(k,j,i)   )               &
              -      (  8.0_wp * ibit32 * adv_mom_5                           &
                  +              ibit31 * adv_mom_3                           &
                     ) *                                                      &
                                    ( w(k,j+2,i) + w(k,j-1,i) )               &
              +      (           ibit32 * adv_mom_5                           &
                     ) *                                                      &
                                    ( w(k,j+3,i) + w(k,j-2,i) )               &
                                  )

          diss_n(k) = - ABS( v_comp(k) ) * (                                  &
                     ( 10.0_wp * ibit32 * adv_mom_5                           &
                  +     3.0_wp * ibit31 * adv_mom_3                           &
                  +              ibit30 * adv_mom_1                           &
                     ) *                                                      &
                                    ( w(k,j+1,i) - w(k,j,i)  )                &
              -      (  5.0_wp * ibit32 * adv_mom_5                           &
                  +              ibit31 * adv_mom_3                           &
                     ) *                                                      &
                                   ( w(k,j+2,i) - w(k,j-1,i) )                &
              +      (           ibit32 * adv_mom_5                           &
                     ) *                                                      &
                                   ( w(k,j+3,i) - w(k,j-2,i) )                &
                                           )
       ENDDO

       DO  k = nzb_max_l+1, nzt

          u_comp(k) = u(k+1,j,i+1) + u(k,j,i+1) - gu
          flux_r(k) = u_comp(k) * (                                           &
                      37.0_wp * ( w(k,j,i+1) + w(k,j,i)   )                   &
                    -  8.0_wp * ( w(k,j,i+2) + w(k,j,i-1) )                   &
                    +           ( w(k,j,i+3) + w(k,j,i-2) ) ) * adv_mom_5

          diss_r(k) = - ABS( u_comp(k) ) * (                                  &
                      10.0_wp * ( w(k,j,i+1) - w(k,j,i)   )                   &
                    -  5.0_wp * ( w(k,j,i+2) - w(k,j,i-1) )                   &
                    +           ( w(k,j,i+3) - w(k,j,i-2) ) ) * adv_mom_5

          v_comp(k) = v(k+1,j+1,i) + v(k,j+1,i) - gv
          flux_n(k) = v_comp(k) * (                                           &
                      37.0_wp * ( w(k,j+1,i) + w(k,j,i)   )                   &
                    -  8.0_wp * ( w(k,j+2,i) + w(k,j-1,i) )                   &
                    +           ( w(k,j+3,i) + w(k,j-2,i) ) ) * adv_mom_5

          diss_n(k) = - ABS( v_comp(k) ) * (                                  &
                      10.0_wp * ( w(k,j+1,i) - w(k,j,i)   )                   &
                    -  5.0_wp * ( w(k,j+2,i) - w(k,j-1,i) )                   &
                    +           ( w(k,j+3,i) - w(k,j-2,i) ) ) * adv_mom_5
       ENDDO

!
!--    Now, compute vertical fluxes. Split loop into a part treating the 
!--    lowest 2 grid points with indirect indexing, a main loop without
!--    indirect indexing, and a loop for the uppermost 2 grip points with
!--    indirect indexing. This allows better vectorization for the main loop.
!--    First, compute the flux at model surface, which need has to be 
!--    calculated explicetely for the tendency at
!--    the first w-level. For topography wall this is done implicitely by
!--    advc_flags_1.
       k         = nzb + 1
       w_comp(k) = w(k,j,i) + w(k-1,j,i)
       flux_t(0) = w_comp(k)       * ( w(k,j,i) + w(k-1,j,i) ) * adv_mom_1
       diss_t(0) = -ABS(w_comp(k)) * ( w(k,j,i) - w(k-1,j,i) ) * adv_mom_1
       
       DO  k = nzb+1, nzb+2
!
!--       k index has to be modified near bottom and top, else array
!--       subscripts will be exceeded.
          ibit35 = REAL( IBITS(advc_flags_2(k,j,i),3,1), KIND = wp )
          ibit34 = REAL( IBITS(advc_flags_2(k,j,i),2,1), KIND = wp )
          ibit33 = REAL( IBITS(advc_flags_2(k,j,i),1,1), KIND = wp )

          k_ppp = k + 3 * ibit35
          k_pp  = k + 2 * ( 1 - ibit33  )
          k_mm  = k - 2 * ibit35

          w_comp(k) = w(k+1,j,i) + w(k,j,i)
          flux_t(k) = w_comp(k) * rho_air(k+1) * (                            &
                     ( 37.0_wp * ibit35 * adv_mom_5                           &
                  +     7.0_wp * ibit34 * adv_mom_3                           &
                  +              ibit33 * adv_mom_1                           &
                     ) *                                                      &
                                ( w(k+1,j,i)  + w(k,j,i)     )                &
              -      (  8.0_wp * ibit35 * adv_mom_5                           &
                  +              ibit34 * adv_mom_3                           &
                     ) *                                                      &
                                ( w(k_pp,j,i)  + w(k-1,j,i)  )                &
              +      (           ibit35 * adv_mom_5                           &
                     ) *                                                      &
                                ( w(k_ppp,j,i) + w(k_mm,j,i) )                &
                                                 )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air(k+1) * (                   &
                     ( 10.0_wp * ibit35 * adv_mom_5                           &
                  +     3.0_wp * ibit34 * adv_mom_3                           &
                  +              ibit33 * adv_mom_1                           &
                     ) *                                                      &
                                ( w(k+1,j,i)   - w(k,j,i)    )                &
              -      (  5.0_wp * ibit35 * adv_mom_5                           &
                  +              ibit34 * adv_mom_3                           &
                     ) *                                                      &
                                ( w(k_pp,j,i)  - w(k-1,j,i)  )                &
              +      (           ibit35 * adv_mom_5                           &
                     ) *                                                      &
                                ( w(k_ppp,j,i) - w(k_mm,j,i) )                &
                                                          )
       ENDDO
       
       DO  k = nzb+3, nzt-2
       
          ibit35 = REAL( IBITS(advc_flags_2(k,j,i),3,1), KIND = wp )
          ibit34 = REAL( IBITS(advc_flags_2(k,j,i),2,1), KIND = wp )
          ibit33 = REAL( IBITS(advc_flags_2(k,j,i),1,1), KIND = wp )

          w_comp(k) = w(k+1,j,i) + w(k,j,i)
          flux_t(k) = w_comp(k) * rho_air(k+1) * (                            &
                     ( 37.0_wp * ibit35 * adv_mom_5                           &
                  +     7.0_wp * ibit34 * adv_mom_3                           &
                  +              ibit33 * adv_mom_1                           &
                     ) *                                                      &
                                ( w(k+1,j,i)  + w(k,j,i)     )                &
              -      (  8.0_wp * ibit35 * adv_mom_5                           &
                  +              ibit34 * adv_mom_3                           &
                     ) *                                                      &
                                ( w(k+2,j,i)  + w(k-1,j,i)  )                 &
              +      (           ibit35 * adv_mom_5                           &
                     ) *                                                      &
                                ( w(k+3,j,i) + w(k-2,j,i) )                   &
                                                 )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air(k+1) * (                   &
                     ( 10.0_wp * ibit35 * adv_mom_5                           &
                  +     3.0_wp * ibit34 * adv_mom_3                           &
                  +              ibit33 * adv_mom_1                           &
                     ) *                                                      &
                                ( w(k+1,j,i)   - w(k,j,i)    )                &
              -      (  5.0_wp * ibit35 * adv_mom_5                           &
                  +              ibit34 * adv_mom_3                           &
                     ) *                                                      &
                                ( w(k+2,j,i)  - w(k-1,j,i)  )                 &
              +      (           ibit35 * adv_mom_5                           &
                     ) *                                                      &
                                ( w(k+3,j,i) - w(k-2,j,i) )                   &
                                                          )
       ENDDO
       
       DO  k = nzt-1, nzt
!
!--       k index has to be modified near bottom and top, else array
!--       subscripts will be exceeded.
          ibit35 = REAL( IBITS(advc_flags_2(k,j,i),3,1), KIND = wp )
          ibit34 = REAL( IBITS(advc_flags_2(k,j,i),2,1), KIND = wp )
          ibit33 = REAL( IBITS(advc_flags_2(k,j,i),1,1), KIND = wp )

          k_ppp = k + 3 * ibit35
          k_pp  = k + 2 * ( 1 - ibit33  )
          k_mm  = k - 2 * ibit35

          w_comp(k) = w(k+1,j,i) + w(k,j,i)
          flux_t(k) = w_comp(k) * rho_air(k+1) * (                            &
                     ( 37.0_wp * ibit35 * adv_mom_5                           &
                  +     7.0_wp * ibit34 * adv_mom_3                           &
                  +              ibit33 * adv_mom_1                           &
                     ) *                                                      &
                                ( w(k+1,j,i)  + w(k,j,i)     )                &
              -      (  8.0_wp * ibit35 * adv_mom_5                           &
                  +              ibit34 * adv_mom_3                           &
                     ) *                                                      &
                                ( w(k_pp,j,i)  + w(k-1,j,i)  )                &
              +      (           ibit35 * adv_mom_5                           &
                     ) *                                                      &
                                ( w(k_ppp,j,i) + w(k_mm,j,i) )                &
                                                 )

          diss_t(k) = - ABS( w_comp(k) ) * rho_air(k+1) * (                   &
                     ( 10.0_wp * ibit35 * adv_mom_5                           &
                  +     3.0_wp * ibit34 * adv_mom_3                           &
                  +              ibit33 * adv_mom_1                           &
                     ) *                                                      &
                                ( w(k+1,j,i)   - w(k,j,i)    )                &
              -      (  5.0_wp * ibit35 * adv_mom_5                           &
                  +              ibit34 * adv_mom_3                           &
                     ) *                                                      &
                                ( w(k_pp,j,i)  - w(k-1,j,i)  )                &
              +      (           ibit35 * adv_mom_5                           &
                     ) *                                                      &
                                ( w(k_ppp,j,i) - w(k_mm,j,i) )                &
                                                          )
       ENDDO
       
       DO  k = nzb+1, nzb_max_l

          flux_d    = flux_t(k-1)
          diss_d    = diss_t(k-1)
          
          ibit29 = REAL( IBITS(advc_flags_1(k,j,i),29,1), KIND = wp )
          ibit28 = REAL( IBITS(advc_flags_1(k,j,i),28,1), KIND = wp )
          ibit27 = REAL( IBITS(advc_flags_1(k,j,i),27,1), KIND = wp )
          
          ibit32 = REAL( IBITS(advc_flags_2(k,j,i),0,1),  KIND = wp )
          ibit31 = REAL( IBITS(advc_flags_1(k,j,i),31,1), KIND = wp )
          ibit30 = REAL( IBITS(advc_flags_1(k,j,i),30,1), KIND = wp )
          
          ibit35 = REAL( IBITS(advc_flags_2(k,j,i),3,1), KIND = wp )
          ibit34 = REAL( IBITS(advc_flags_2(k,j,i),2,1), KIND = wp )
          ibit33 = REAL( IBITS(advc_flags_2(k,j,i),1,1), KIND = wp )
!
!--       Calculate the divergence of the velocity field. A respective
!--       correction is needed to overcome numerical instabilities introduced
!--       by a not sufficient reduction of divergences near topography.
          div = ( ( ( u_comp(k) + gu ) * ( ibit27 + ibit28 + ibit29 )         &
                  - ( u(k+1,j,i) + u(k,j,i)   )                               & 
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k,j,i-1),27,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j,i-1),28,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j,i-1),29,1), KIND = wp )     &
                                      )                                       &
                  ) * ddx                                                     &
              +   ( ( v_comp(k) + gv ) * ( ibit30 + ibit31 + ibit32 )         & 
                  - ( v(k+1,j,i) + v(k,j,i)   )                               &
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k,j-1,i),30,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j-1,i),31,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_2(k,j-1,i),0,1),  KIND = wp )     &
                                      )                                       &
                  ) * ddy                                                     &
              +   ( w_comp(k)               * rho_air(k+1)                    &
                                            * ( ibit33 + ibit34 + ibit35 )    &
                - ( w(k,j,i) + w(k-1,j,i) ) * rho_air(k)                      &
                                    * (                                       &
                     REAL( IBITS(advc_flags_2(k-1,j,i),1,1), KIND = wp )      &
                   + REAL( IBITS(advc_flags_2(k-1,j,i),2,1), KIND = wp )      &
                   + REAL( IBITS(advc_flags_2(k-1,j,i),3,1), KIND = wp )      &
                                      )                                       & 
                  ) * drho_air_zw(k) * ddzu(k+1)                              &
                ) * 0.5_wp

          tend(k,j,i) = tend(k,j,i) - (                                       &
                      ( flux_r(k) + diss_r(k)                                 &
                    -   flux_l_w(k,j,tn) - diss_l_w(k,j,tn)   ) * ddx         &
                    + ( flux_n(k) + diss_n(k)                                 &
                    -   flux_s_w(k,tn) - diss_s_w(k,tn)       ) * ddy         &
                    + ( ( flux_t(k) + diss_t(k) )                             &
                    -   ( flux_d    + diss_d    )                             &
                                              ) * drho_air_zw(k) * ddzu(k+1)  &
                                      ) + div * w(k,j,i)

          flux_l_w(k,j,tn) = flux_r(k)
          diss_l_w(k,j,tn) = diss_r(k)
          flux_s_w(k,tn)   = flux_n(k)
          diss_s_w(k,tn)   = diss_n(k)
!
!--       Statistical Evaluation of w'w'.
          sums_ws2_ws_l(k,tn)  = sums_ws2_ws_l(k,tn)                          &
                      + ( flux_t(k)                                           &
                       * ( w_comp(k) - 2.0_wp * hom(k,1,3,0)                ) &
                       / ( w_comp(k) + SIGN( 1.0E-20_wp, w_comp(k) )        ) &
                        + diss_t(k)                                           &
                       *   ABS( w_comp(k) - 2.0_wp * hom(k,1,3,0)           ) &
                       / ( ABS( w_comp(k) ) + 1.0E-20_wp                    ) &
                        ) *   weight_substep(intermediate_timestep_count)

       ENDDO
       
       DO  k = nzb_max_l+1, nzt

          flux_d    = flux_t(k-1)
          diss_d    = diss_t(k-1)
!
!--       Calculate the divergence of the velocity field. A respective
!--       correction is needed to overcome numerical instabilities introduced
!--       by a not sufficient reduction of divergences near topography.
          div = ( ( u_comp(k) + gu - ( u(k+1,j,i) + u(k,j,i)   ) ) * ddx       &
              +   ( v_comp(k) + gv - ( v(k+1,j,i) + v(k,j,i)   ) ) * ddy       &
              +   ( w_comp(k)               * rho_air(k+1)                     &
                - ( w(k,j,i) + w(k-1,j,i) ) * rho_air(k)                       & 
                  ) * drho_air_zw(k) * ddzu(k+1)                               &
                ) * 0.5_wp

          tend(k,j,i) = tend(k,j,i) - (                                        &
                      ( flux_r(k) + diss_r(k)                                  &
                    -   flux_l_w(k,j,tn) - diss_l_w(k,j,tn)   ) * ddx          &
                    + ( flux_n(k) + diss_n(k)                                  &
                    -   flux_s_w(k,tn) - diss_s_w(k,tn)       ) * ddy          &
                    + ( ( flux_t(k) + diss_t(k) )                              &
                    -   ( flux_d    + diss_d    )                              &
                                              ) * drho_air_zw(k) * ddzu(k+1)   &
                                      ) + div * w(k,j,i)

          flux_l_w(k,j,tn) = flux_r(k)
          diss_l_w(k,j,tn) = diss_r(k)
          flux_s_w(k,tn)   = flux_n(k)
          diss_s_w(k,tn)   = diss_n(k)
!
!--       Statistical Evaluation of w'w'.
          sums_ws2_ws_l(k,tn)  = sums_ws2_ws_l(k,tn)                           &
                      + ( flux_t(k)                                            &
                       * ( w_comp(k) - 2.0_wp * hom(k,1,3,0)                )  &
                       / ( w_comp(k) + SIGN( 1.0E-20_wp, w_comp(k) )        )  &
                        + diss_t(k)                                            &
                       *   ABS( w_comp(k) - 2.0_wp * hom(k,1,3,0)           )  &
                       / ( ABS( w_comp(k) ) + 1.0E-20_wp                    )  &
                        ) *   weight_substep(intermediate_timestep_count)

       ENDDO

    END SUBROUTINE advec_w_ws_ij
    

!------------------------------------------------------------------------------!
! Description:
! ------------
!> Scalar advection - Call for all grid points
!------------------------------------------------------------------------------!
    SUBROUTINE advec_s_ws( sk, sk_char )

       USE arrays_3d,                                                          &
           ONLY:  ddzw, drho_air, rho_air_zw, tend, u, u_stokes_zu, v,         &
                  v_stokes_zu, w

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s, intermediate_timestep_count, &
                  u_gtrans, v_gtrans 

       USE grid_variables,                                                     &
           ONLY:  ddx, ddy

       USE indices,                                                            &
           ONLY:  nxl, nxlg, nxr, nxrg, nyn, nyng, nys, nysg, nzb, nzb_max,    &
                  nzt, advc_flags_1

       USE kinds

       USE salsa_util_mod,                                                     &
           ONLY:  sums_salsa_ws_l

       USE statistics,                                                         &
           ONLY:  hom, sums_wspts_ws_l, sums_wsqs_ws_l, sums_wssas_ws_l,       &
                  sums_wsqcs_ws_l, sums_wsqrs_ws_l, sums_wsncs_ws_l,           &
                  sums_wsnrs_ws_l, sums_wsss_ws_l, weight_substep  



       IMPLICIT NONE

       CHARACTER (LEN = *), INTENT(IN)    ::  sk_char !< string identifier, used for assign fluxes to the correct dimension in the analysis array 
       INTEGER(iwp) ::  sk_num !< integer identifier, used for assign fluxes to the correct dimension in the analysis array
       
       INTEGER(iwp) ::  i         !< grid index along x-direction
       INTEGER(iwp) ::  j         !< grid index along y-direction
       INTEGER(iwp) ::  k         !< grid index along z-direction
       INTEGER(iwp) ::  k_mm      !< k-2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_pp      !< k+2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_ppp     !< k+3 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  nzb_max_l !< index indicating upper bound for order degradation of horizontal advection terms 
       INTEGER(iwp) ::  tn = 0    !< number of OpenMP thread
       
!       
!--    sk is an array from parameter list. It should not be a pointer, because 
!--    in that case the compiler can not assume a stride 1 and cannot perform 
!--    a strided one vector load. Adding the CONTIGUOUS keyword makes things 
!--    even worse, because the compiler cannot assume strided one in the 
!--    caller side.
       REAL(wp), INTENT(IN),DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg) ::  sk !<  advected scalar

       REAL(wp) ::  ibit0  !< flag indicating 1st-order scheme along x-direction
       REAL(wp) ::  ibit1  !< flag indicating 3rd-order scheme along x-direction
       REAL(wp) ::  ibit2  !< flag indicating 5th-order scheme along x-direction
#ifdef _OPENACC
       REAL(wp) ::  ibit0_l  !< flag indicating 1st-order scheme along x-direction
       REAL(wp) ::  ibit1_l  !< flag indicating 3rd-order scheme along x-direction
       REAL(wp) ::  ibit2_l  !< flag indicating 5th-order scheme along x-direction
#endif
       REAL(wp) ::  ibit3  !< flag indicating 1st-order scheme along y-direction
       REAL(wp) ::  ibit4  !< flag indicating 3rd-order scheme along y-direction
       REAL(wp) ::  ibit5  !< flag indicating 5th-order scheme along y-direction
#ifdef _OPENACC
       REAL(wp) ::  ibit3_s  !< flag indicating 1st-order scheme along y-direction
       REAL(wp) ::  ibit4_s  !< flag indicating 3rd-order scheme along y-direction
       REAL(wp) ::  ibit5_s  !< flag indicating 5th-order scheme along y-direction
#endif
       REAL(wp) ::  ibit6  !< flag indicating 1st-order scheme along z-direction
       REAL(wp) ::  ibit7  !< flag indicating 3rd-order scheme along z-direction
       REAL(wp) ::  ibit8  !< flag indicating 5th-order scheme along z-direction
       REAL(wp) ::  diss_d !< artificial dissipation term at grid box bottom 
       REAL(wp) ::  div    !< diverence on scalar grid
       REAL(wp) ::  flux_d !< 6th-order flux at grid box bottom
       REAL(wp) ::  u_comp !< advection velocity along x-direction
#ifdef _OPENACC
       REAL(wp) ::  u_comp_l !< advection velocity along x-direction
#endif
       REAL(wp) ::  v_comp !< advection velocity along y-direction
#ifdef _OPENACC
       REAL(wp) ::  v_comp_s !< advection velocity along y-direction
#endif
       
       REAL(wp) ::  diss_n !< discretized artificial dissipation at northward-side of the grid box
       REAL(wp) ::  diss_r !< discretized artificial dissipation at rightward-side of the grid box
       REAL(wp) ::  diss_t !< discretized artificial dissipation at rightward-side of the grid box
       REAL(wp) ::  flux_n !< discretized 6th-order flux at northward-side of the grid box
       REAL(wp) ::  flux_r !< discretized 6th-order flux at rightward-side of the grid box
       REAL(wp) ::  flux_t !< discretized 6th-order flux at rightward-side of the grid box
       
       REAL(wp) ::  diss_s !< discretized artificial dissipation term at southward-side of the grid box
       REAL(wp) ::  flux_s !< discretized 6th-order flux at northward-side of the grid box
#ifndef _OPENACC
       REAL(wp), DIMENSION(nzb+1:nzt) ::  swap_diss_y_local !< discretized artificial dissipation term at southward-side of the grid box
       REAL(wp), DIMENSION(nzb+1:nzt) ::  swap_flux_y_local !< discretized 6th-order flux at northward-side of the grid box
#endif
       
       REAL(wp) ::  diss_l !< discretized artificial dissipation term at leftward-side of the grid box
       REAL(wp) ::  flux_l !< discretized 6th-order flux at leftward-side of the grid box
#ifndef _OPENACC
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn) ::  swap_diss_x_local !< discretized artificial dissipation term at leftward-side of the grid box
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn) ::  swap_flux_x_local !< discretized 6th-order flux at leftward-side of the grid box
#endif
        
!
!--    Set local version of nzb_max. At non-cyclic boundaries the order of the 
!--    advection need to be degraded near the boundary. Please note, in contrast
!--    to the cache-optimized routines, nzb_max_l is set constantly for the 
!--    entire subdomain, in order to avoid unsymmetric loops which might be 
!--    an issue for GPUs.
       IF( bc_dirichlet_l  .OR.  bc_radiation_l  .OR.                          &
           bc_dirichlet_r  .OR.  bc_radiation_r  .OR.                          &
           bc_dirichlet_s  .OR.  bc_radiation_s  .OR.                          &
           bc_dirichlet_n  .OR.  bc_radiation_n )  THEN
          nzb_max_l = nzt
       ELSE
          nzb_max_l = nzb_max
       END IF 
        
       SELECT CASE ( sk_char )

           CASE ( 'pt' )
              sk_num = 1
           CASE ( 'sa' )
              sk_num = 2
           CASE ( 'q' )
              sk_num = 3
           CASE ( 'qc' )
              sk_num = 4
           CASE ( 'qr' )
              sk_num = 5
           CASE ( 'nc' )
              sk_num = 6
           CASE ( 'nr' )
              sk_num = 7
           CASE ( 's' )
              sk_num = 8
           CASE ( 'aerosol_mass', 'aerosol_number', 'salsa_gas' )
              sk_num = 9

       END SELECT

#ifndef _OPENACC
!
!--    Compute the fluxes for the whole left boundary of the processor domain.
       i = nxl
       DO  j = nys, nyn

          DO  k = nzb+1, nzb_max_l

             ibit2 = REAL( IBITS(advc_flags_1(k,j,i-1),2,1), KIND = wp )
             ibit1 = REAL( IBITS(advc_flags_1(k,j,i-1),1,1), KIND = wp )
             ibit0 = REAL( IBITS(advc_flags_1(k,j,i-1),0,1), KIND = wp )

             u_comp                 = u(k,j,i) - u_gtrans + u_stokes_zu(k)
             swap_flux_x_local(k,j) = u_comp * (                              &
                                             ( 37.0_wp * ibit2 * adv_sca_5    &
                                          +     7.0_wp * ibit1 * adv_sca_3    &
                                          +              ibit0 * adv_sca_1    &
                                             ) *                              &
                                          ( sk(k,j,i)   + sk(k,j,i-1)    )    &
                                      -      (  8.0_wp * ibit2 * adv_sca_5    &
                                          +              ibit1 * adv_sca_3    &
                                             ) *                              &
                                          ( sk(k,j,i+1) + sk(k,j,i-2)    )    &
                                      +      (           ibit2 * adv_sca_5    & 
                                             ) *                              &
                                          ( sk(k,j,i+2) + sk(k,j,i-3)    )    &
                                               )

              swap_diss_x_local(k,j) = -ABS( u_comp ) * (                     &
                                             ( 10.0_wp * ibit2 * adv_sca_5    &
                                          +     3.0_wp * ibit1 * adv_sca_3    &
                                          +              ibit0 * adv_sca_1    &
                                             ) *                              &
                                          ( sk(k,j,i)   - sk(k,j,i-1) )       &
                                      -      (  5.0_wp * ibit2 * adv_sca_5    &
                                          +              ibit1 * adv_sca_3    &
                                             ) *                              &
                                         ( sk(k,j,i+1) - sk(k,j,i-2)  )       &
                                      +      (           ibit2 * adv_sca_5    &
                                             ) *                              &
                                          ( sk(k,j,i+2) - sk(k,j,i-3) )       &
                                                        )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             u_comp                 = u(k,j,i) - u_gtrans + u_stokes_zu(k)
             swap_flux_x_local(k,j) = u_comp * (                              &
                                      37.0_wp * ( sk(k,j,i)   + sk(k,j,i-1) ) &
                                    -  8.0_wp * ( sk(k,j,i+1) + sk(k,j,i-2) ) &
                                    +           ( sk(k,j,i+2) + sk(k,j,i-3) ) &
                                               ) * adv_sca_5

             swap_diss_x_local(k,j) = -ABS( u_comp ) * (                      &
                                      10.0_wp * ( sk(k,j,i)   - sk(k,j,i-1) ) &
                                    -  5.0_wp * ( sk(k,j,i+1) - sk(k,j,i-2) ) &
                                    +           ( sk(k,j,i+2) - sk(k,j,i-3) ) &
                                                       ) * adv_sca_5

          ENDDO

       ENDDO
#endif

       !$ACC PARALLEL LOOP COLLAPSE(2) FIRSTPRIVATE(tn, sk_num) &
       !$ACC PRIVATE(i, j, k, k_mm, k_pp, k_ppp) &
       !$ACC PRIVATE(ibit0, ibit1, ibit2, ibit3, ibit4, ibit5) &
       !$ACC PRIVATE(ibit0_l, ibit1_l, ibit2_l) &
       !$ACC PRIVATE(ibit3_s, ibit4_s, ibit5_s) &
       !$ACC PRIVATE(ibit6, ibit7, ibit8) &
       !$ACC PRIVATE(flux_r, diss_r, flux_l, diss_l) &
       !$ACC PRIVATE(flux_n, diss_n, flux_s, diss_s) &
       !$ACC PRIVATE(flux_t, diss_t, flux_d, diss_d) &
       !$ACC PRIVATE(div, u_comp, u_comp_l, v_comp, v_comp_s) &
       !$ACC PRESENT(advc_flags_1) &
       !$ACC PRESENT(sk, u, v, w, u_stokes_zu, v_stokes_zu) &
       !$ACC PRESENT(drho_air, rho_air_zw, ddzw) &
       !$ACC PRESENT(tend) &
       !$ACC PRESENT(hom(nzb+1:nzb_max_l,1,1:3,0)) &
       !$ACC PRESENT(weight_substep(intermediate_timestep_count)) &
       !$ACC PRESENT(sums_wspts_ws_l, sums_wssas_ws_l) &
       !$ACC PRESENT(sums_wsqs_ws_l, sums_wsqcs_ws_l) &
       !$ACC PRESENT(sums_wsqrs_ws_l, sums_wsncs_ws_l) &
       !$ACC PRESENT(sums_wsnrs_ws_l, sums_wsss_ws_l) &
       !$ACC PRESENT(sums_salsa_ws_l)
       DO  i = nxl, nxr

#ifndef _OPENACC
          j = nys
          DO  k = nzb+1, nzb_max_l

             ibit5 = REAL( IBITS(advc_flags_1(k,j-1,i),5,1), KIND = wp )
             ibit4 = REAL( IBITS(advc_flags_1(k,j-1,i),4,1), KIND = wp )
             ibit3 = REAL( IBITS(advc_flags_1(k,j-1,i),3,1), KIND = wp )

             v_comp               = v(k,j,i) - v_gtrans + v_stokes_zu(k)
             swap_flux_y_local(k) = v_comp * (                                &
                                             ( 37.0_wp * ibit5 * adv_sca_5    &
                                          +     7.0_wp * ibit4 * adv_sca_3    &
                                          +              ibit3 * adv_sca_1    &
                                             ) *                              &
                                         ( sk(k,j,i)  + sk(k,j-1,i)     )     &
                                       -     (  8.0_wp * ibit5 * adv_sca_5    &
                                          +              ibit4 * adv_sca_3    &
                                              ) *                             &
                                         ( sk(k,j+1,i) + sk(k,j-2,i)    )     &
                                      +      (           ibit5 * adv_sca_5    &
                                             ) *                              &
                                        ( sk(k,j+2,i) + sk(k,j-3,i)     )     &
                                             )

             swap_diss_y_local(k) = -ABS( v_comp ) * (                        &
                                             ( 10.0_wp * ibit5 * adv_sca_5    &
                                          +     3.0_wp * ibit4 * adv_sca_3    &
                                          +              ibit3 * adv_sca_1    &
                                             ) *                              &
                                          ( sk(k,j,i)   - sk(k,j-1,i)    )    &
                                      -      (  5.0_wp * ibit5 * adv_sca_5    &
                                          +              ibit4 * adv_sca_3    &
                                             ) *                              &
                                          ( sk(k,j+1,i) - sk(k,j-2,i)    )    &
                                      +      (           ibit5 * adv_sca_5    &
                                             ) *                              &
                                          ( sk(k,j+2,i) - sk(k,j-3,i)    )    &
                                                     )

          ENDDO
!
!--       Above to the top of the highest topography. No degradation necessary.
          DO  k = nzb_max_l+1, nzt

             v_comp               = v(k,j,i) - v_gtrans + v_stokes_zu(k)
             swap_flux_y_local(k) = v_comp * (                               &
                                    37.0_wp * ( sk(k,j,i)   + sk(k,j-1,i) )  &
                                  -  8.0_wp * ( sk(k,j+1,i) + sk(k,j-2,i) )  &
                                  +           ( sk(k,j+2,i) + sk(k,j-3,i) )  &
                                             ) * adv_sca_5
             swap_diss_y_local(k) = -ABS( v_comp ) * (                       &
                                    10.0_wp * ( sk(k,j,i)   - sk(k,j-1,i) )  &
                                  -  5.0_wp * ( sk(k,j+1,i) - sk(k,j-2,i) )  &
                                  +             sk(k,j+2,i) - sk(k,j-3,i)    &
                                                      ) * adv_sca_5

          ENDDO
#endif

          DO  j = nys, nyn

             flux_d    = 0.0_wp
             diss_d    = 0.0_wp

             DO  k = nzb+1, nzb_max_l

                ibit2 = REAL( IBITS(advc_flags_1(k,j,i),2,1), KIND = wp )
                ibit1 = REAL( IBITS(advc_flags_1(k,j,i),1,1), KIND = wp )
                ibit0 = REAL( IBITS(advc_flags_1(k,j,i),0,1), KIND = wp )

                u_comp = u(k,j,i+1) - u_gtrans + u_stokes_zu(k)
                flux_r = u_comp * (                                           &
                          ( 37.0_wp * ibit2 * adv_sca_5                       &
                      +      7.0_wp * ibit1 * adv_sca_3                       &
                      +               ibit0 * adv_sca_1                       &
                          ) *                                                 &
                             ( sk(k,j,i+1) + sk(k,j,i)   )                    &
                   -      (  8.0_wp * ibit2 * adv_sca_5                       &
                       +              ibit1 * adv_sca_3                       &
                          ) *                                                 &
                             ( sk(k,j,i+2) + sk(k,j,i-1) )                    &
                   +      (           ibit2 * adv_sca_5                       &
                          ) *                                                 &
                             ( sk(k,j,i+3) + sk(k,j,i-2) )                    &
                                     )

                diss_r = -ABS( u_comp ) * (                                   &
                          ( 10.0_wp * ibit2 * adv_sca_5                       &
                       +     3.0_wp * ibit1 * adv_sca_3                       &
                       +              ibit0 * adv_sca_1                       &
                          ) *                                                 &
                             ( sk(k,j,i+1) - sk(k,j,i)   )                    &
                   -      (  5.0_wp * ibit2 * adv_sca_5                       &
                       +              ibit1 * adv_sca_3                       &
                          ) *                                                 &
                             ( sk(k,j,i+2) - sk(k,j,i-1) )                    &
                   +      (           ibit2 * adv_sca_5                       &
                          ) *                                                 &
                             ( sk(k,j,i+3) - sk(k,j,i-2) )                    &
                                             )

#ifdef _OPENACC
!
!--             Recompute the left fluxes.
                ibit2_l = REAL( IBITS(advc_flags_1(k,j,i-1),2,1), KIND = wp )
                ibit1_l = REAL( IBITS(advc_flags_1(k,j,i-1),1,1), KIND = wp )
                ibit0_l = REAL( IBITS(advc_flags_1(k,j,i-1),0,1), KIND = wp )

                u_comp_l = u(k,j,i) - u_gtrans + u_stokes_zu(k)
                flux_l = u_comp_l * (                                         &
                                             ( 37.0_wp * ibit2_l * adv_sca_5  &
                                          +     7.0_wp * ibit1_l * adv_sca_3  &
                                          +              ibit0_l * adv_sca_1  &
                                             ) *                              &
                                          ( sk(k,j,i)   + sk(k,j,i-1)    )    &
                                      -      (  8.0_wp * ibit2_l * adv_sca_5  &
                                          +              ibit1_l * adv_sca_3  &
                                             ) *                              &
                                          ( sk(k,j,i+1) + sk(k,j,i-2)    )    &
                                      +      (           ibit2_l * adv_sca_5  & 
                                             ) *                              &
                                          ( sk(k,j,i+2) + sk(k,j,i-3)    )    &
                                               )

                 diss_l = -ABS( u_comp_l ) * (                                &
                                             ( 10.0_wp * ibit2_l * adv_sca_5  &
                                          +     3.0_wp * ibit1_l * adv_sca_3  &
                                          +              ibit0_l * adv_sca_1  &
                                             ) *                              &
                                          ( sk(k,j,i)   - sk(k,j,i-1) )       &
                                      -      (  5.0_wp * ibit2_l * adv_sca_5  &
                                          +              ibit1_l * adv_sca_3  &
                                             ) *                              &
                                         ( sk(k,j,i+1) - sk(k,j,i-2)  )       &
                                      +      (           ibit2_l * adv_sca_5  &
                                             ) *                              &
                                          ( sk(k,j,i+2) - sk(k,j,i-3) )       &
                                                        )
#else
                flux_l = swap_flux_x_local(k,j)
                diss_l = swap_diss_x_local(k,j)
#endif

                ibit5 = REAL( IBITS(advc_flags_1(k,j,i),5,1), KIND = wp )
                ibit4 = REAL( IBITS(advc_flags_1(k,j,i),4,1), KIND = wp )
                ibit3 = REAL( IBITS(advc_flags_1(k,j,i),3,1), KIND = wp )

                v_comp = v(k,j+1,i) - v_gtrans + v_stokes_zu(k)
                flux_n = v_comp * (                                           &
                          ( 37.0_wp * ibit5 * adv_sca_5                       &
                       +     7.0_wp * ibit4 * adv_sca_3                       &
                       +              ibit3 * adv_sca_1                       &
                          ) *                                                 &
                             ( sk(k,j+1,i) + sk(k,j,i)   )                    &
                   -      (  8.0_wp * ibit5 * adv_sca_5                       &
                       +              ibit4 * adv_sca_3                       &
                          ) *                                                 &
                             ( sk(k,j+2,i) + sk(k,j-1,i) )                    &
                   +      (           ibit5 * adv_sca_5                       &
                          ) *                                                 &
                             ( sk(k,j+3,i) + sk(k,j-2,i) )                    &
                                     )

                diss_n = -ABS( v_comp ) * (                                   &
                          ( 10.0_wp * ibit5 * adv_sca_5                       &
                       +     3.0_wp * ibit4 * adv_sca_3                       &
                       +              ibit3 * adv_sca_1                       &
                          ) *                                                 &
                             ( sk(k,j+1,i) - sk(k,j,i)    )                   &
                   -      (  5.0_wp * ibit5 * adv_sca_5                       &
                       +              ibit4 * adv_sca_3                       &
                          ) *                                                 &
                             ( sk(k,j+2,i) - sk(k,j-1,i)  )                   &
                   +      (           ibit5 * adv_sca_5                       &
                          ) *                                                 &
                             ( sk(k,j+3,i) - sk(k,j-2,i) )                    &
                                             )

#ifdef _OPENACC
!
!--             Recompute the south fluxes.
                ibit5_s = REAL( IBITS(advc_flags_1(k,j-1,i),5,1), KIND = wp )
                ibit4_s = REAL( IBITS(advc_flags_1(k,j-1,i),4,1), KIND = wp )
                ibit3_s = REAL( IBITS(advc_flags_1(k,j-1,i),3,1), KIND = wp )

                v_comp_s = v(k,j,i) - v_gtrans + v_stokes_zu(k)
                flux_s = v_comp_s * (                                         &
                                             ( 37.0_wp * ibit5_s * adv_sca_5  &
                                          +     7.0_wp * ibit4_s * adv_sca_3  &
                                          +              ibit3_s * adv_sca_1  &
                                             ) *                              &
                                         ( sk(k,j,i)  + sk(k,j-1,i)     )     &
                                       -     (  8.0_wp * ibit5_s * adv_sca_5  &
                                          +              ibit4_s * adv_sca_3  &
                                              ) *                             &
                                         ( sk(k,j+1,i) + sk(k,j-2,i)    )     &
                                      +      (           ibit5_s * adv_sca_5  &
                                             ) *                              &
                                        ( sk(k,j+2,i) + sk(k,j-3,i)     )     &
                                             )

                diss_s = -ABS( v_comp_s ) * (                                 &
                                             ( 10.0_wp * ibit5_s * adv_sca_5  &
                                          +     3.0_wp * ibit4_s * adv_sca_3  &
                                          +              ibit3_s * adv_sca_1  &
                                             ) *                              &
                                          ( sk(k,j,i)   - sk(k,j-1,i)    )    &
                                      -      (  5.0_wp * ibit5_s * adv_sca_5  &
                                          +              ibit4_s * adv_sca_3  &
                                             ) *                              &
                                          ( sk(k,j+1,i) - sk(k,j-2,i)    )    &
                                      +      (           ibit5_s * adv_sca_5  &
                                             ) *                              &
                                          ( sk(k,j+2,i) - sk(k,j-3,i)    )    &
                                                     )
#else
                flux_s = swap_flux_y_local(k)
                diss_s = swap_diss_y_local(k)
#endif

!
!--             k index has to be modified near bottom and top, else array
!--             subscripts will be exceeded.
                ibit8 = REAL( IBITS(advc_flags_1(k,j,i),8,1), KIND = wp )
                ibit7 = REAL( IBITS(advc_flags_1(k,j,i),7,1), KIND = wp )
                ibit6 = REAL( IBITS(advc_flags_1(k,j,i),6,1), KIND = wp )

                k_ppp = k + 3 * ibit8
                k_pp  = k + 2 * ( 1 - ibit6  )
                k_mm  = k - 2 * ibit8


                flux_t = w(k,j,i) * rho_air_zw(k) * (                         &
                           ( 37.0_wp * ibit8 * adv_sca_5                      &
                        +     7.0_wp * ibit7 * adv_sca_3                      &
                        +           ibit6 * adv_sca_1                         &
                           ) *                                                &
                                   ( sk(k+1,j,i)  + sk(k,j,i)    )            &
                    -      (  8.0_wp * ibit8 * adv_sca_5                      &
                        +              ibit7 * adv_sca_3                      &
                           ) *                                                &
                                   ( sk(k_pp,j,i) + sk(k-1,j,i)  )            &
                    +      (           ibit8 * adv_sca_5                      &
                           ) *     ( sk(k_ppp,j,i)+ sk(k_mm,j,i) )            &
                                       )

                diss_t = -ABS( w(k,j,i) ) * rho_air_zw(k) * (                 &
                           ( 10.0_wp * ibit8 * adv_sca_5                      &
                        +     3.0_wp * ibit7 * adv_sca_3                      &
                        +              ibit6 * adv_sca_1                      &
                           ) *                                                &
                                   ( sk(k+1,j,i)   - sk(k,j,i)    )           &
                    -      (  5.0_wp * ibit8 * adv_sca_5                      &
                        +              ibit7 * adv_sca_3                      &
                           ) *                                                &
                                   ( sk(k_pp,j,i)  - sk(k-1,j,i)  )           &
                    +      (           ibit8 * adv_sca_5                      &
                           ) *                                                &
                                   ( sk(k_ppp,j,i) - sk(k_mm,j,i) )           &
                                               )
!
!--             Calculate the divergence of the velocity field. A respective
!--             correction is needed to overcome numerical instabilities caused
!--             by a not sufficient reduction of divergences near topography.
                div   =   ( u(k,j,i+1) * ( ibit0 + ibit1 + ibit2 )             &
                          - u(k,j,i)   * (                                     &
                        REAL( IBITS(advc_flags_1(k,j,i-1),0,1), KIND = wp )    &
                      + REAL( IBITS(advc_flags_1(k,j,i-1),1,1), KIND = wp )    &
                      + REAL( IBITS(advc_flags_1(k,j,i-1),2,1), KIND = wp )    &
                                         )                                     &
                          ) * ddx                                              &
                        + ( v(k,j+1,i) * ( ibit3 + ibit4 + ibit5 )             &
                          - v(k,j,i)   * (                                     &
                        REAL( IBITS(advc_flags_1(k,j-1,i),3,1), KIND = wp )    &
                      + REAL( IBITS(advc_flags_1(k,j-1,i),4,1), KIND = wp )    &
                      + REAL( IBITS(advc_flags_1(k,j-1,i),5,1), KIND = wp )    &
                                         )                                     &
                          ) * ddy                                              &
                        + ( w(k,j,i) * rho_air_zw(k) *                         &
                                         ( ibit6 + ibit7 + ibit8 )             &
                          - w(k-1,j,i) * rho_air_zw(k-1) *                     &
                                         (                                     &
                        REAL( IBITS(advc_flags_1(k-1,j,i),6,1), KIND = wp )    &
                      + REAL( IBITS(advc_flags_1(k-1,j,i),7,1), KIND = wp )    &
                      + REAL( IBITS(advc_flags_1(k-1,j,i),8,1), KIND = wp )    &
                                         )                                     &      
                          ) * drho_air(k) * ddzw(k)


                tend(k,j,i) = tend(k,j,i) - (                                 &
                        ( ( flux_r + diss_r )                                 &
                      -   ( flux_l + diss_l ) ) * ddx                         &
                      + ( ( flux_n + diss_n )                                 &
                      -   ( flux_s + diss_s ) ) * ddy                         &
                      + ( ( flux_t + diss_t )                                 &
                      -   ( flux_d + diss_d ) ) * drho_air(k) * ddzw(k)       &
                                            ) + sk(k,j,i) * div

#ifndef _OPENACC
                swap_flux_y_local(k)   = flux_n
                swap_diss_y_local(k)   = diss_n
                swap_flux_x_local(k,j) = flux_r
                swap_diss_x_local(k,j) = diss_r
#endif
                flux_d                 = flux_t
                diss_d                 = diss_t

!
!--             Evaluation of statistics. 
                SELECT CASE ( sk_num )

                    CASE ( 1 )
                       !$ACC ATOMIC
                       sums_wspts_ws_l(k,tn) = sums_wspts_ws_l(k,tn)           &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 2 )
                       !$ACC ATOMIC
                       sums_wssas_ws_l(k,tn) = sums_wssas_ws_l(k,tn)           &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 3 )
                       !$ACC ATOMIC
                       sums_wsqs_ws_l(k,tn)  = sums_wsqs_ws_l(k,tn)            &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 4 )
                       !$ACC ATOMIC
                       sums_wsqcs_ws_l(k,tn)  = sums_wsqcs_ws_l(k,tn)          &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 5 )
                       !$ACC ATOMIC
                       sums_wsqrs_ws_l(k,tn)  = sums_wsqrs_ws_l(k,tn)          &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 6 )
                       !$ACC ATOMIC
                       sums_wsncs_ws_l(k,tn)  = sums_wsncs_ws_l(k,tn)          &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 7 )
                       !$ACC ATOMIC
                       sums_wsnrs_ws_l(k,tn)  = sums_wsnrs_ws_l(k,tn)          &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 8 )
                       !$ACC ATOMIC
                       sums_wsss_ws_l(k,tn)  = sums_wsss_ws_l(k,tn)            &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 9 )
                        !$ACC ATOMIC
                        sums_salsa_ws_l(k,tn)  = sums_salsa_ws_l(k,tn)         &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)

                END SELECT

             ENDDO

             DO  k = nzb_max_l+1, nzt

                u_comp = u(k,j,i+1) - u_gtrans + u_stokes_zu(k)
                flux_r = u_comp * (                                           &
                      37.0_wp * ( sk(k,j,i+1) + sk(k,j,i)   )                 &
                    -  8.0_wp * ( sk(k,j,i+2) + sk(k,j,i-1) )                 &
                    +           ( sk(k,j,i+3) + sk(k,j,i-2) ) ) * adv_sca_5
                diss_r = -ABS( u_comp ) * (                                   &
                      10.0_wp * ( sk(k,j,i+1) - sk(k,j,i)   )                 &
                    -  5.0_wp * ( sk(k,j,i+2) - sk(k,j,i-1) )                 &
                    +           ( sk(k,j,i+3) - sk(k,j,i-2) ) ) * adv_sca_5

#ifdef _OPENACC
!
!--             Recompute the left fluxes.
                u_comp_l = u(k,j,i) - u_gtrans + u_stokes_zu(k)
                flux_l = u_comp_l * (                                         &
                                      37.0_wp * ( sk(k,j,i)   + sk(k,j,i-1) ) &
                                    -  8.0_wp * ( sk(k,j,i+1) + sk(k,j,i-2) ) &
                                    +           ( sk(k,j,i+2) + sk(k,j,i-3) ) &
                                               ) * adv_sca_5

                diss_l = -ABS( u_comp_l ) * (                                 &
                                      10.0_wp * ( sk(k,j,i)   - sk(k,j,i-1) ) &
                                    -  5.0_wp * ( sk(k,j,i+1) - sk(k,j,i-2) ) &
                                    +           ( sk(k,j,i+2) - sk(k,j,i-3) ) &
                                                       ) * adv_sca_5
#else
                flux_l = swap_flux_x_local(k,j)
                diss_l = swap_diss_x_local(k,j)
#endif

                v_comp = v(k,j+1,i) - v_gtrans + v_stokes_zu(k)
                flux_n = v_comp * (                                           &
                      37.0_wp * ( sk(k,j+1,i) + sk(k,j,i)   )                 &
                    -  8.0_wp * ( sk(k,j+2,i) + sk(k,j-1,i) )                 &
                    +           ( sk(k,j+3,i) + sk(k,j-2,i) ) ) * adv_sca_5
                diss_n = -ABS( v_comp ) * (                                   &
                      10.0_wp * ( sk(k,j+1,i) - sk(k,j,i)   )                 &
                    -  5.0_wp * ( sk(k,j+2,i) - sk(k,j-1,i) )                 &
                    +           ( sk(k,j+3,i) - sk(k,j-2,i) ) ) * adv_sca_5

#ifdef _OPENACC
!
!--             Recompute the south fluxes.
                v_comp_s = v(k,j,i) - v_gtrans + v_stokes_zu(k)
                flux_s = v_comp_s * (                                        &
                                    37.0_wp * ( sk(k,j,i)   + sk(k,j-1,i) )  &
                                  -  8.0_wp * ( sk(k,j+1,i) + sk(k,j-2,i) )  &
                                  +           ( sk(k,j+2,i) + sk(k,j-3,i) )  &
                                             ) * adv_sca_5
                diss_s = -ABS( v_comp_s ) * (                                &
                                    10.0_wp * ( sk(k,j,i)   - sk(k,j-1,i) )  &
                                  -  5.0_wp * ( sk(k,j+1,i) - sk(k,j-2,i) )  &
                                  +             sk(k,j+2,i) - sk(k,j-3,i)    &
                                                      ) * adv_sca_5
#else
                flux_s = swap_flux_y_local(k)
                diss_s = swap_diss_y_local(k)
#endif

!
!--             k index has to be modified near bottom and top, else array
!--             subscripts will be exceeded.
                ibit8 = REAL( IBITS(advc_flags_1(k,j,i),8,1), KIND = wp )
                ibit7 = REAL( IBITS(advc_flags_1(k,j,i),7,1), KIND = wp )
                ibit6 = REAL( IBITS(advc_flags_1(k,j,i),6,1), KIND = wp )

                k_ppp = k + 3 * ibit8
                k_pp  = k + 2 * ( 1 - ibit6  )
                k_mm  = k - 2 * ibit8


                flux_t = w(k,j,i) * rho_air_zw(k) * (                         &
                           ( 37.0_wp * ibit8 * adv_sca_5                      &
                        +     7.0_wp * ibit7 * adv_sca_3                      &
                        +              ibit6 * adv_sca_1                      &
                           ) *                                                &
                                   ( sk(k+1,j,i)  + sk(k,j,i)     )           &
                    -      (  8.0_wp * ibit8 * adv_sca_5                      &
                        +              ibit7 * adv_sca_3                      &
                           ) *                                                &
                                   ( sk(k_pp,j,i) + sk(k-1,j,i)   )           &
                    +      (           ibit8 * adv_sca_5                      &
                           ) *     ( sk(k_ppp,j,i)+ sk(k_mm,j,i)  )           &
                                       )

                diss_t = -ABS( w(k,j,i) ) * rho_air_zw(k) * (                 &
                           ( 10.0_wp * ibit8 * adv_sca_5                      &
                        +     3.0_wp * ibit7 * adv_sca_3                      &
                        +              ibit6 * adv_sca_1                      &
                           ) *                                                &
                                   ( sk(k+1,j,i)   - sk(k,j,i)    )           &
                    -      (  5.0_wp * ibit8 * adv_sca_5                      &
                        +              ibit7 * adv_sca_3                      &
                           ) *                                                &
                                   ( sk(k_pp,j,i)  - sk(k-1,j,i)  )           &
                    +      (           ibit8 * adv_sca_5                      &
                           ) *                                                &
                                   ( sk(k_ppp,j,i) - sk(k_mm,j,i) )           &
                                               )
!
!--             Calculate the divergence of the velocity field. A respective
!--             correction is needed to overcome numerical instabilities introduced
!--             by a not sufficient reduction of divergences near topography.
                div         =   ( u(k,j,i+1) - u(k,j,i)   ) * ddx              &
                              + ( v(k,j+1,i) - v(k,j,i)   ) * ddy              &
                              + ( w(k,j,i)   * rho_air_zw(k) -                 &
                                  w(k-1,j,i) * rho_air_zw(k-1)                 &
                                ) * drho_air(k) * ddzw(k)

                tend(k,j,i) = tend(k,j,i) - (                                 &
                        ( ( flux_r + diss_r )                                 &
                      -   ( flux_l + diss_l ) ) * ddx                         &
                      + ( ( flux_n + diss_n )                                 &
                      -   ( flux_s + diss_s ) ) * ddy                         &
                      + ( ( flux_t + diss_t )                                 &
                      -   ( flux_d + diss_d ) ) * drho_air(k) * ddzw(k)       &
                                            ) + sk(k,j,i) * div

#ifndef _OPENACC
                swap_flux_y_local(k)   = flux_n
                swap_diss_y_local(k)   = diss_n
                swap_flux_x_local(k,j) = flux_r
                swap_diss_x_local(k,j) = diss_r
#endif
                flux_d                 = flux_t
                diss_d                 = diss_t

!
!--             Evaluation of statistics. 
                SELECT CASE ( sk_num )

                    CASE ( 1 )
                       !$ACC ATOMIC
                       sums_wspts_ws_l(k,tn) = sums_wspts_ws_l(k,tn)           &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 2 )
                       !$ACC ATOMIC
                       sums_wssas_ws_l(k,tn) = sums_wssas_ws_l(k,tn)           &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 3 )
                       !$ACC ATOMIC
                       sums_wsqs_ws_l(k,tn)  = sums_wsqs_ws_l(k,tn)            &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 4 )
                       !$ACC ATOMIC
                       sums_wsqcs_ws_l(k,tn)  = sums_wsqcs_ws_l(k,tn)          &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 5 )
                       !$ACC ATOMIC
                       sums_wsqrs_ws_l(k,tn)  = sums_wsqrs_ws_l(k,tn)          &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 6 )
                       !$ACC ATOMIC
                       sums_wsncs_ws_l(k,tn)  = sums_wsncs_ws_l(k,tn)          &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 7 )
                       !$ACC ATOMIC
                       sums_wsnrs_ws_l(k,tn)  = sums_wsnrs_ws_l(k,tn)          &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 8 )
                       !$ACC ATOMIC
                       sums_wsss_ws_l(k,tn)  = sums_wsss_ws_l(k,tn)            &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)
                    CASE ( 9 )
                        !$ACC ATOMIC
                        sums_salsa_ws_l(k,tn)  = sums_salsa_ws_l(k,tn)         &
                          + ( flux_t                                           &
                                / ( w(k,j,i) + SIGN( 1.0E-20_wp, w(k,j,i) ) )  &
                                * ( w(k,j,i) - hom(k,1,3,0)                 )  &
                            + diss_t                                           &
                                / ( ABS(w(k,j,i)) + 1.0E-20_wp              )  &
                                *   ABS(w(k,j,i) - hom(k,1,3,0)             )  &
                            ) * weight_substep(intermediate_timestep_count)

                END SELECT

             ENDDO

         ENDDO
      ENDDO

    END SUBROUTINE advec_s_ws


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Advection of u - Call for all grid points
!------------------------------------------------------------------------------!
    SUBROUTINE advec_u_ws

       USE arrays_3d,                                                          &
           ONLY:  ddzw, drho_air, tend, u, v, w, rho_air_zw

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s, intermediate_timestep_count, &
                  u_gtrans, v_gtrans

       USE grid_variables,                                                     &
           ONLY:  ddx, ddy

       USE indices,                                                            &
           ONLY:  nxlu, nxr, nyn, nys, nzb, nzb_max, nzt, advc_flags_1
           
       USE kinds
       
       USE statistics,                                                         &
           ONLY:  hom, sums_us2_ws_l, sums_wsus_ws_l, weight_substep

       IMPLICIT NONE

       INTEGER(iwp) ::  i         !< grid index along x-direction
       INTEGER(iwp) ::  j         !< grid index along y-direction
       INTEGER(iwp) ::  k         !< grid index along z-direction
       INTEGER(iwp) ::  k_mm      !< k-2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_pp      !< k+2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_ppp     !< k+3 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  nzb_max_l !< index indicating upper bound for order degradation of horizontal advection terms 
       INTEGER(iwp) ::  tn = 0    !< number of OpenMP thread
       
       REAL(wp)    ::  ibit9  !< flag indicating 1st-order scheme along x-direction
       REAL(wp)    ::  ibit10 !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)    ::  ibit11 !< flag indicating 5th-order scheme along x-direction
#ifdef _OPENACC
       REAL(wp)    ::  ibit9_l  !< flag indicating 1st-order scheme along x-direction
       REAL(wp)    ::  ibit10_l !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)    ::  ibit11_l !< flag indicating 5th-order scheme along x-direction
#endif
       REAL(wp)    ::  ibit12 !< flag indicating 1st-order scheme along y-direction
       REAL(wp)    ::  ibit13 !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)    ::  ibit14 !< flag indicating 5th-order scheme along y-direction
#ifdef _OPENACC
       REAL(wp)    ::  ibit12_s !< flag indicating 1st-order scheme along y-direction
       REAL(wp)    ::  ibit13_s !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)    ::  ibit14_s !< flag indicating 5th-order scheme along y-direction
#endif
       REAL(wp)    ::  ibit15 !< flag indicating 1st-order scheme along z-direction
       REAL(wp)    ::  ibit16 !< flag indicating 3rd-order scheme along z-direction
       REAL(wp)    ::  ibit17 !< flag indicating 5th-order scheme along z-direction
       REAL(wp)    ::  diss_d !< artificial dissipation term at grid box bottom
       REAL(wp)    ::  div    !< diverence on u-grid
       REAL(wp)    ::  flux_d !< 6th-order flux at grid box bottom
       REAL(wp)    ::  gu     !< Galilei-transformation velocity along x
       REAL(wp)    ::  gv     !< Galilei-transformation velocity along y 
       REAL(wp)    ::  v_comp !< advection velocity along y
#ifdef _OPENACC
       REAL(wp)    ::  v_comp_s !< advection velocity along y
#endif
       REAL(wp)    ::  w_comp !< advection velocity along z
       
       REAL(wp)    :: diss_s  !< discretized artificial dissipation at southward-side of the grid box
       REAL(wp)    :: flux_s  !< discretized 6th-order flux at southward-side of the grid box
#ifndef _OPENACC
       REAL(wp), DIMENSION(nzb+1:nzt) ::  swap_diss_y_local_u !< discretized artificial dissipation at southward-side of the grid box
       REAL(wp), DIMENSION(nzb+1:nzt) ::  swap_flux_y_local_u !< discretized 6th-order flux at southward-side of the grid box
#endif
       
       REAL(wp)    :: diss_l  !< discretized artificial dissipation at leftward-side of the grid box
       REAL(wp)    :: flux_l  !< discretized 6th-order flux at leftward-side of the grid box
#ifndef _OPENACC
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn) ::  swap_diss_x_local_u !< discretized artificial dissipation at leftward-side of the grid box
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn) ::  swap_flux_x_local_u !< discretized 6th-order flux at leftward-side of the grid box
#endif
       
       REAL(wp) ::  diss_n !< discretized artificial dissipation at northward-side of the grid box
       REAL(wp) ::  diss_r !< discretized artificial dissipation at leftward-side of the grid box
       REAL(wp) ::  diss_t !< discretized artificial dissipation at top of the grid box
       REAL(wp) ::  flux_n !< discretized 6th-order flux at northward-side of the grid box
       REAL(wp) ::  flux_r !< discretized 6th-order flux at rightward-side of the grid box
       REAL(wp) ::  flux_t !< discretized 6th-order flux at top of the grid box
       REAL(wp) ::  u_comp !< advection velocity along x
#ifdef _OPENACC
       REAL(wp)    ::  u_comp_l !<
#endif
!
!--    Set local version of nzb_max. At non-cyclic boundaries the order of the 
!--    advection need to be degraded near the boundary. Please note, in contrast
!--    to the cache-optimized routines, nzb_max_l is set constantly for the 
!--    entire subdomain, in order to avoid unsymmetric loops which might be 
!--    an issue for GPUs.
       IF( bc_dirichlet_l  .OR.  bc_radiation_l  .OR.                          &
           bc_dirichlet_r  .OR.  bc_radiation_r  .OR.                          &
           bc_dirichlet_s  .OR.  bc_radiation_s  .OR.                          &
           bc_dirichlet_n  .OR.  bc_radiation_n )  THEN
          nzb_max_l = nzt
       ELSE
          nzb_max_l = nzb_max
       END IF 
 
       gu = 2.0_wp * u_gtrans
       gv = 2.0_wp * v_gtrans

#ifndef _OPENACC
!
!--    Compute the fluxes for the whole left boundary of the processor domain.
       i = nxlu
       DO  j = nys, nyn
          DO  k = nzb+1, nzb_max_l

             ibit11 = REAL( IBITS(advc_flags_1(k,j,i-1),11,1), KIND = wp )
             ibit10 = REAL( IBITS(advc_flags_1(k,j,i-1),10,1), KIND = wp )
             ibit9  = REAL( IBITS(advc_flags_1(k,j,i-1),9,1),  KIND = wp )

             u_comp                   = u(k,j,i) + u(k,j,i-1) - gu
             swap_flux_x_local_u(k,j) = u_comp * (                             &
                                       ( 37.0_wp * ibit11 * adv_mom_5          &
                                    +     7.0_wp * ibit10 * adv_mom_3          &
                                    +              ibit9  * adv_mom_1          &
                                       ) *                                     &
                                     ( u(k,j,i)   + u(k,j,i-1) )               &
                                -      (  8.0_wp * ibit11 * adv_mom_5          &
                                    +              ibit10 * adv_mom_3          &
                                       ) *                                     &
                                     ( u(k,j,i+1) + u(k,j,i-2) )               &
                                +      (           ibit11 * adv_mom_5          &
                                       ) *                                     &
                                     ( u(k,j,i+2) + u(k,j,i-3) )               &
                                                   )

              swap_diss_x_local_u(k,j) = - ABS( u_comp ) * (                   &
                                       ( 10.0_wp * ibit11 * adv_mom_5          &
                                    +     3.0_wp * ibit10 * adv_mom_3          &
                                    +              ibit9  * adv_mom_1          &
                                       ) *                                     &
                                     ( u(k,j,i)   - u(k,j,i-1) )               &
                                -      (  5.0_wp * ibit11 * adv_mom_5          &
                                    +              ibit10 * adv_mom_3          &
                                       ) *                                     &
                                     ( u(k,j,i+1) - u(k,j,i-2) )               &
                                +      (           ibit11 * adv_mom_5          &
                                       ) *                                     &
                                     ( u(k,j,i+2) - u(k,j,i-3) )               &
                                                             )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             u_comp            = u(k,j,i) + u(k,j,i-1) - gu
             swap_flux_x_local_u(k,j) = u_comp * (                             &
                             37.0_wp * ( u(k,j,i) + u(k,j,i-1)   )             &
                           -  8.0_wp * ( u(k,j,i+1) + u(k,j,i-2) )             &
                           +           ( u(k,j,i+2) + u(k,j,i-3) ) ) * adv_mom_5
             swap_diss_x_local_u(k,j) = - ABS(u_comp) * (                      &
                             10.0_wp * ( u(k,j,i) - u(k,j,i-1)   )             &
                           -  5.0_wp * ( u(k,j,i+1) - u(k,j,i-2) )             &
                           +           ( u(k,j,i+2) - u(k,j,i-3) ) ) * adv_mom_5

          ENDDO
       ENDDO
#endif

       !$ACC PARALLEL LOOP COLLAPSE(2) FIRSTPRIVATE(tn, gu, gv) &
       !$ACC PRIVATE(i, j, k, k_mm, k_pp, k_ppp) &
       !$ACC PRIVATE(ibit9, ibit10, ibit11, ibit12, ibit13, ibit14) &
       !$ACC PRIVATE(ibit9_l, ibit10_l, ibit11_l) &
       !$ACC PRIVATE(ibit12_s, ibit13_s, ibit14_s) &
       !$ACC PRIVATE(ibit15, ibit16, ibit17) &
       !$ACC PRIVATE(flux_r, diss_r, flux_l, diss_l) &
       !$ACC PRIVATE(flux_n, diss_n, flux_s, diss_s) &
       !$ACC PRIVATE(flux_t, diss_t, flux_d, diss_d) &
       !$ACC PRIVATE(div, u_comp, u_comp_l, v_comp, v_comp_s, w_comp) &
       !$ACC PRESENT(advc_flags_1) &
       !$ACC PRESENT(u, v, w) &
       !$ACC PRESENT(drho_air, rho_air_zw, ddzw) &
       !$ACC PRESENT(tend) &
       !$ACC PRESENT(hom(nzb+1:nzb_max_l,1,1:3,0)) &
       !$ACC PRESENT(weight_substep(intermediate_timestep_count)) &
       !$ACC PRESENT(sums_us2_ws_l, sums_wsus_ws_l)
       DO i = nxlu, nxr
#ifndef _OPENACC
!       
!--       The following loop computes the fluxes for the south boundary points
          j = nys
          DO  k = nzb+1, nzb_max_l

             ibit14 = REAL( IBITS(advc_flags_1(k,j-1,i),14,1), KIND = wp )
             ibit13 = REAL( IBITS(advc_flags_1(k,j-1,i),13,1), KIND = wp )
             ibit12 = REAL( IBITS(advc_flags_1(k,j-1,i),12,1), KIND = wp )

             v_comp                 = v(k,j,i) + v(k,j,i-1) - gv
             swap_flux_y_local_u(k) = v_comp * (                              &
                                   ( 37.0_wp * ibit14 * adv_mom_5             &
                                +     7.0_wp * ibit13 * adv_mom_3             &
                                +              ibit12 * adv_mom_1             &
                                   ) *                                        &
                                     ( u(k,j,i)   + u(k,j-1,i) )              &
                            -      (  8.0_wp * ibit14 * adv_mom_5             &
                            +                  ibit13 * adv_mom_3             &
                                   ) *                                        &
                                     ( u(k,j+1,i) + u(k,j-2,i) )              &
                        +      (               ibit14 * adv_mom_5             &
                               ) *                                            &
                                     ( u(k,j+2,i) + u(k,j-3,i) )              &
                                               )

             swap_diss_y_local_u(k) = - ABS ( v_comp ) * (                    &
                                   ( 10.0_wp * ibit14 * adv_mom_5             &
                                +     3.0_wp * ibit13 * adv_mom_3             &
                                +              ibit12 * adv_mom_1             &
                                   ) *                                        &
                                     ( u(k,j,i)   - u(k,j-1,i) )              &
                            -      (  5.0_wp * ibit14 * adv_mom_5             &
                                +              ibit13 * adv_mom_3             &
                                   ) *                                        &
                                     ( u(k,j+1,i) - u(k,j-2,i) )              &
                            +      (           ibit14 * adv_mom_5             &
                                   ) *                                        &
                                     ( u(k,j+2,i) - u(k,j-3,i) )              &
                                                         )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             v_comp                 = v(k,j,i) + v(k,j,i-1) - gv
             swap_flux_y_local_u(k) = v_comp * (                              &
                           37.0_wp * ( u(k,j,i) + u(k,j-1,i)   )                 &
                         -  8.0_wp * ( u(k,j+1,i) + u(k,j-2,i) )                 &
                         +           ( u(k,j+2,i) + u(k,j-3,i) ) ) * adv_mom_5
             swap_diss_y_local_u(k) = - ABS(v_comp) * (                       &
                           10.0_wp * ( u(k,j,i) - u(k,j-1,i)   )                 &
                         -  5.0_wp * ( u(k,j+1,i) - u(k,j-2,i) )                 &
                         +           ( u(k,j+2,i) - u(k,j-3,i) ) ) * adv_mom_5

          ENDDO
#endif

!
!--       Computation of interior fluxes and tendency terms
          DO  j = nys, nyn

             flux_d    = 0.0_wp
             diss_d    = 0.0_wp

             DO  k = nzb+1, nzb_max_l

                ibit11 = REAL( IBITS(advc_flags_1(k,j,i),11,1), KIND = wp )
                ibit10 = REAL( IBITS(advc_flags_1(k,j,i),10,1), KIND = wp )
                ibit9  = REAL( IBITS(advc_flags_1(k,j,i),9,1),  KIND = wp )

                u_comp = u(k,j,i+1) + u(k,j,i)
                flux_r = ( u_comp - gu ) * (                                 &
                          ( 37.0_wp * ibit11 * adv_mom_5                     &
                       +     7.0_wp * ibit10 * adv_mom_3                     &
                       +              ibit9  * adv_mom_1                     &
                          ) *                                                &
                                 ( u(k,j,i+1) + u(k,j,i)   )                 &
                   -      (  8.0_wp * ibit11 * adv_mom_5                     &
                       +              ibit10 * adv_mom_3                     &
                          ) *                                                &
                                 ( u(k,j,i+2) + u(k,j,i-1) )                 &
                   +      (           ibit11 * adv_mom_5                     &
                          ) *                                                &
                                 ( u(k,j,i+3) + u(k,j,i-2) )                 &
                                                 )

                diss_r = - ABS( u_comp - gu ) * (                            &
                          ( 10.0_wp * ibit11 * adv_mom_5                     &
                       +     3.0_wp * ibit10 * adv_mom_3                     & 
                       +              ibit9  * adv_mom_1                     &
                          ) *                                                &
                                 ( u(k,j,i+1) - u(k,j,i)  )                  &
                   -      (  5.0_wp * ibit11 * adv_mom_5                     &
                       +              ibit10 * adv_mom_3                     &
                          ) *                                                &
                                 ( u(k,j,i+2) - u(k,j,i-1) )                 &
                   +      (           ibit11 * adv_mom_5                     &
                          ) *                                                &
                                 ( u(k,j,i+3) - u(k,j,i-2) )                 &
                                                     )

#ifdef _OPENACC
!
!--             Recompute the left fluxes.
                ibit11_l = REAL( IBITS(advc_flags_1(k,j,i-1),11,1), KIND = wp )
                ibit10_l = REAL( IBITS(advc_flags_1(k,j,i-1),10,1), KIND = wp )
                ibit9_l  = REAL( IBITS(advc_flags_1(k,j,i-1),9,1),  KIND = wp )

                u_comp_l = u(k,j,i) + u(k,j,i-1) - gu
                flux_l   = u_comp_l * (                                        &
                                       ( 37.0_wp * ibit11_l * adv_mom_5           &
                                    +     7.0_wp * ibit10_l * adv_mom_3           &
                                    +              ibit9_l  * adv_mom_1           &
                                       ) *                                     &
                                     ( u(k,j,i)   + u(k,j,i-1) )               &
                                -      (  8.0_wp * ibit11_l * adv_mom_5           &
                                    +              ibit10_l * adv_mom_3           &
                                       ) *                                     &
                                     ( u(k,j,i+1) + u(k,j,i-2) )               &
                                +      (           ibit11_l * adv_mom_5           &
                                       ) *                                     &
                                     ( u(k,j,i+2) + u(k,j,i-3) )               &
                                                   )

                diss_l   = - ABS( u_comp_l ) * (                               &
                                       ( 10.0_wp * ibit11_l * adv_mom_5           &
                                    +     3.0_wp * ibit10_l * adv_mom_3           &
                                    +              ibit9_l  * adv_mom_1           &
                                       ) *                                     &
                                     ( u(k,j,i)   - u(k,j,i-1) )               &
                                -      (  5.0_wp * ibit11_l * adv_mom_5           &
                                    +              ibit10_l * adv_mom_3           &
                                       ) *                                     &
                                     ( u(k,j,i+1) - u(k,j,i-2) )               &
                                +      (           ibit11_l * adv_mom_5           &
                                       ) *                                     &
                                     ( u(k,j,i+2) - u(k,j,i-3) )               &
                                                             )
#else
                flux_l = swap_flux_x_local_u(k,j)
                diss_l = swap_diss_x_local_u(k,j)
#endif

                ibit14 = REAL( IBITS(advc_flags_1(k,j,i),14,1), KIND = wp )
                ibit13 = REAL( IBITS(advc_flags_1(k,j,i),13,1), KIND = wp )
                ibit12 = REAL( IBITS(advc_flags_1(k,j,i),12,1), KIND = wp )

                v_comp = v(k,j+1,i) + v(k,j+1,i-1) - gv
                flux_n = v_comp * (                                          &
                          ( 37.0_wp * ibit14 * adv_mom_5                     &
                       +     7.0_wp * ibit13 * adv_mom_3                     &
                       +              ibit12 * adv_mom_1                     &
                          ) *                                                &
                                 ( u(k,j+1,i) + u(k,j,i)   )                 &
                   -      (  8.0_wp * ibit14 * adv_mom_5                     &
                       +              ibit13 * adv_mom_3                     &
                          ) *                                                &
                                 ( u(k,j+2,i) + u(k,j-1,i) )                 &
                   +      (           ibit14 * adv_mom_5                     & 
                          ) *                                                &
                                 ( u(k,j+3,i) + u(k,j-2,i) )                 &
                                                 )

                diss_n = - ABS ( v_comp ) * (                                &
                          ( 10.0_wp * ibit14 * adv_mom_5                     &
                       +     3.0_wp * ibit13 * adv_mom_3                     &
                       +              ibit12 * adv_mom_1                     &
                          ) *                                                &
                                 ( u(k,j+1,i) - u(k,j,i)  )                  &
                   -      (  5.0_wp * ibit14 * adv_mom_5                     &
                       +              ibit13 * adv_mom_3                     &
                          ) *                                                &
                                 ( u(k,j+2,i) - u(k,j-1,i) )                 &
                   +      (           ibit14 * adv_mom_5                     &
                          ) *                                                &
                                 ( u(k,j+3,i) - u(k,j-2,i) )                 &
                                                      )

#ifdef _OPENACC
!
!--             Recompute the south fluxes.
                ibit14_s = REAL( IBITS(advc_flags_1(k,j-1,i),14,1), KIND = wp )
                ibit13_s = REAL( IBITS(advc_flags_1(k,j-1,i),13,1), KIND = wp )
                ibit12_s = REAL( IBITS(advc_flags_1(k,j-1,i),12,1), KIND = wp )

                v_comp_s = v(k,j,i) + v(k,j,i-1) - gv
                flux_s   = v_comp_s * (                                      &
                                   ( 37.0_wp * ibit14_s * adv_mom_5             &
                                +     7.0_wp * ibit13_s * adv_mom_3             &
                                +              ibit12_s * adv_mom_1             &
                                   ) *                                       &
                                     ( u(k,j,i)   + u(k,j-1,i) )             &
                            -      (  8.0_wp * ibit14_s * adv_mom_5             &
                            +                  ibit13_s * adv_mom_3             &
                                   ) *                                       &
                                     ( u(k,j+1,i) + u(k,j-2,i) )             &
                        +      (               ibit14_s * adv_mom_5             &
                               ) *                                           &
                                     ( u(k,j+2,i) + u(k,j-3,i) )             &
                                               )

                diss_s   = - ABS ( v_comp_s ) * (                            &
                                   ( 10.0_wp * ibit14_s * adv_mom_5              &
                                +     3.0_wp * ibit13_s * adv_mom_3              &
                                +              ibit12_s * adv_mom_1              &
                                   ) *                                        &
                                     ( u(k,j,i)   - u(k,j-1,i) )              &
                            -      (  5.0_wp * ibit14_s * adv_mom_5              &
                                +              ibit13_s * adv_mom_3              &
                                   ) *                                        &
                                     ( u(k,j+1,i) - u(k,j-2,i) )              &
                            +      (           ibit14_s * adv_mom_5              &
                                   ) *                                        &
                                     ( u(k,j+2,i) - u(k,j-3,i) )              &
                                                         )
#else
                flux_s = swap_flux_y_local_u(k)
                diss_s = swap_diss_y_local_u(k)
#endif

!
!--             k index has to be modified near bottom and top, else array
!--             subscripts will be exceeded.
                ibit17 = REAL( IBITS(advc_flags_1(k,j,i),17,1), KIND = wp )
                ibit16 = REAL( IBITS(advc_flags_1(k,j,i),16,1), KIND = wp )
                ibit15 = REAL( IBITS(advc_flags_1(k,j,i),15,1), KIND = wp )

                k_ppp = k + 3 * ibit17
                k_pp  = k + 2 * ( 1 - ibit15  )
                k_mm  = k - 2 * ibit17

                w_comp = w(k,j,i) + w(k,j,i-1)
                flux_t = w_comp * rho_air_zw(k) * (                          &
                          ( 37.0_wp * ibit17 * adv_mom_5                     &
                       +     7.0_wp * ibit16 * adv_mom_3                     &
                       +              ibit15 * adv_mom_1                     & 
                          ) *                                                &
                             ( u(k+1,j,i)  + u(k,j,i)     )                  &
                   -      (  8.0_wp * ibit17 * adv_mom_5                     &
                       +              ibit16 * adv_mom_3                     &
                          ) *                                                &
                             ( u(k_pp,j,i) + u(k-1,j,i)   )                  &
                   +      (           ibit17 * adv_mom_5                     &
                          ) *                                                &
                             ( u(k_ppp,j,i) + u(k_mm,j,i) )                  &
                                      )

                diss_t = - ABS( w_comp ) * rho_air_zw(k) * (                 &
                          ( 10.0_wp * ibit17 * adv_mom_5                     &
                       +     3.0_wp * ibit16 * adv_mom_3                     &
                       +              ibit15 * adv_mom_1                     &
                          ) *                                                &
                             ( u(k+1,j,i)   - u(k,j,i)    )                  &
                   -      (  5.0_wp * ibit17 * adv_mom_5                     &
                       +              ibit16 * adv_mom_3                     &
                          ) *                                                &
                             ( u(k_pp,j,i)  - u(k-1,j,i)  )                  &
                   +      (           ibit17 * adv_mom_5                     &
                           ) *                                               &
                             ( u(k_ppp,j,i) - u(k_mm,j,i) )                  &
                                              )
!
!--             Calculate the divergence of the velocity field. A respective
!--             correction is needed to overcome numerical instabilities caused
!--             by a not sufficient reduction of divergences near topography.
                div = ( ( u_comp * ( ibit9 + ibit10 + ibit11 )                &
                - ( u(k,j,i)   + u(k,j,i-1)   )                               &
                                    * (                                       &
                    REAL( IBITS(advc_flags_1(k,j,i-1),9,1),  KIND = wp )      &
                  + REAL( IBITS(advc_flags_1(k,j,i-1),10,1), KIND = wp )      &
                  + REAL( IBITS(advc_flags_1(k,j,i-1),11,1), KIND = wp )      &
                                      )                                       &
                  ) * ddx                                                     &
               +  ( ( v_comp + gv ) * ( ibit12 + ibit13 + ibit14 )            &
                  - ( v(k,j,i)   + v(k,j,i-1 )  )                             &
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k,j-1,i),12,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j-1,i),13,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j-1,i),14,1), KIND = wp )     &
                                      )                                       &
                  ) * ddy                                                     &
               +  ( w_comp * rho_air_zw(k) * ( ibit15 + ibit16 + ibit17 )     &
                - ( w(k-1,j,i) + w(k-1,j,i-1) ) * rho_air_zw(k-1)             &
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k-1,j,i),15,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k-1,j,i),16,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k-1,j,i),17,1), KIND = wp )     &
                                      )                                       &  
                  ) * drho_air(k) * ddzw(k)                                   &
                ) * 0.5_wp



                tend(k,j,i) = tend(k,j,i) - (                                  &
                 ( ( flux_r + diss_r )                                         &
               -   ( flux_l + diss_l ) ) * ddx                                 &
               + ( ( flux_n + diss_n )                                         &
               -   ( flux_s + diss_s ) ) * ddy                                 &
               + ( ( flux_t + diss_t )                                         &
               -   ( flux_d + diss_d ) ) * drho_air(k) * ddzw(k)               &
                                           ) + div * u(k,j,i)

#ifndef _OPENACC
                swap_flux_x_local_u(k,j) = flux_r
                swap_diss_x_local_u(k,j) = diss_r
                swap_flux_y_local_u(k)   = flux_n
                swap_diss_y_local_u(k)   = diss_n
#endif
                flux_d                   = flux_t
                diss_d                   = diss_t
!
!--             Statistical Evaluation of u'u'. The factor has to be applied
!--             for right evaluation when gallilei_trans = .T. .
                !$ACC ATOMIC
                sums_us2_ws_l(k,tn) = sums_us2_ws_l(k,tn)                      &
                + ( flux_r                                                     &
                    * ( u_comp - 2.0_wp * hom(k,1,1,0)                   )     &
                    / ( u_comp - gu + SIGN( 1.0E-20_wp, u_comp - gu )    )     &
                  + diss_r                                                     &
                    *   ABS( u_comp - 2.0_wp * hom(k,1,1,0)              )     &
                    / ( ABS( u_comp - gu ) + 1.0E-20_wp                  )     &
                  ) *   weight_substep(intermediate_timestep_count)
!
!--             Statistical Evaluation of w'u'.
                !$ACC ATOMIC
                sums_wsus_ws_l(k,tn) = sums_wsus_ws_l(k,tn)                    &
                + ( flux_t                                                     &
                    * ( w_comp - 2.0_wp * hom(k,1,3,0)                   )     &
                    / ( w_comp + SIGN( 1.0E-20_wp, w_comp )              )     &
                  + diss_t                                                     &
                    *   ABS( w_comp - 2.0_wp * hom(k,1,3,0)              )     &
                    / ( ABS( w_comp ) + 1.0E-20_wp                       )     &
                  ) *   weight_substep(intermediate_timestep_count)

             ENDDO

             DO  k = nzb_max_l+1, nzt

                u_comp = u(k,j,i+1) + u(k,j,i)
                flux_r = ( u_comp - gu ) * (                                  &
                         37.0_wp * ( u(k,j,i+1) + u(k,j,i)   )                   &
                       -  8.0_wp * ( u(k,j,i+2) + u(k,j,i-1) )                   &
                       +           ( u(k,j,i+3) + u(k,j,i-2) ) ) * adv_mom_5
                diss_r = - ABS( u_comp - gu ) * (                             &
                         10.0_wp * ( u(k,j,i+1) - u(k,j,i)   )                   &
                       -  5.0_wp * ( u(k,j,i+2) - u(k,j,i-1) )                   &
                       +           ( u(k,j,i+3) - u(k,j,i-2) ) ) * adv_mom_5

#ifdef _OPENACC
!
!--             Recompute the left fluxes.
                u_comp_l = u(k,j,i) + u(k,j,i-1) - gu
                flux_l   = u_comp_l * (                                       &
                             37.0_wp * ( u(k,j,i) + u(k,j,i-1)   )                &
                           -  8.0_wp * ( u(k,j,i+1) + u(k,j,i-2) )                &
                           +           ( u(k,j,i+2) + u(k,j,i-3) ) ) * adv_mom_5
                diss_l   = - ABS(u_comp_l) * (                                &
                             10.0_wp * ( u(k,j,i) - u(k,j,i-1)   )                &
                           -  5.0_wp * ( u(k,j,i+1) - u(k,j,i-2) )                &
                           +           ( u(k,j,i+2) - u(k,j,i-3) ) ) * adv_mom_5
#else
                flux_l = swap_flux_x_local_u(k,j)
                diss_l = swap_diss_x_local_u(k,j)
#endif

                v_comp = v(k,j+1,i) + v(k,j+1,i-1) - gv
                flux_n = v_comp * (                                           &
                         37.0_wp * ( u(k,j+1,i) + u(k,j,i)   )                   &
                       -  8.0_wp * ( u(k,j+2,i) + u(k,j-1,i) )                   &
                       +           ( u(k,j+3,i) + u(k,j-2,i) ) ) * adv_mom_5
                diss_n = - ABS( v_comp ) * (                                  &
                         10.0_wp * ( u(k,j+1,i) - u(k,j,i)   )                   &
                       -  5.0_wp * ( u(k,j+2,i) - u(k,j-1,i) )                   &
                       +           ( u(k,j+3,i) - u(k,j-2,i) ) ) * adv_mom_5

#ifdef _OPENACC
!
!--             Recompute the south fluxes.
                v_comp_s = v(k,j,i) + v(k,j,i-1) - gv
                flux_s   = v_comp_s * (                                       &
                           37.0_wp * ( u(k,j,i) + u(k,j-1,i)   )                 &
                         -  8.0_wp * ( u(k,j+1,i) + u(k,j-2,i) )                 &
                         +           ( u(k,j+2,i) + u(k,j-3,i) ) ) * adv_mom_5
                diss_s   = - ABS( v_comp_s ) * (                              &
                           10.0_wp * ( u(k,j,i) - u(k,j-1,i)   )                 &
                         -  5.0_wp * ( u(k,j+1,i) - u(k,j-2,i) )                 &
                         +           ( u(k,j+2,i) - u(k,j-3,i) ) ) * adv_mom_5
#else
                flux_s = swap_flux_y_local_u(k)
                diss_s = swap_diss_y_local_u(k)
#endif

!
!--             k index has to be modified near bottom and top, else array
!--             subscripts will be exceeded.
                ibit17 = REAL( IBITS(advc_flags_1(k,j,i),17,1), KIND = wp )
                ibit16 = REAL( IBITS(advc_flags_1(k,j,i),16,1), KIND = wp )
                ibit15 = REAL( IBITS(advc_flags_1(k,j,i),15,1), KIND = wp )

                k_ppp = k + 3 * ibit17
                k_pp  = k + 2 * ( 1 - ibit15  )
                k_mm  = k - 2 * ibit17

                w_comp = w(k,j,i) + w(k,j,i-1)
                flux_t = w_comp * rho_air_zw(k) * (                          &
                          ( 37.0_wp * ibit17 * adv_mom_5                        &
                       +     7.0_wp * ibit16 * adv_mom_3                        &
                       +              ibit15 * adv_mom_1                        &
                          ) *                                                &
                             ( u(k+1,j,i)  + u(k,j,i)     )                  &
                   -      (  8.0_wp * ibit17 * adv_mom_5                        &
                       +              ibit16 * adv_mom_3                        &
                          ) *                                                &
                             ( u(k_pp,j,i) + u(k-1,j,i)   )                  &
                   +      (           ibit17 * adv_mom_5                        &
                          ) *                                                &
                             ( u(k_ppp,j,i) + u(k_mm,j,i) )                  &
                                      )

                diss_t = - ABS( w_comp ) * rho_air_zw(k) * (                 &
                          ( 10.0_wp * ibit17 * adv_mom_5                        &
                       +     3.0_wp * ibit16 * adv_mom_3                        &
                       +              ibit15 * adv_mom_1                        &
                          ) *                                                &
                             ( u(k+1,j,i)   - u(k,j,i)    )                  &
                   -      (  5.0_wp * ibit17 * adv_mom_5                        &
                       +              ibit16 * adv_mom_3                        &
                          ) *                                                &
                             ( u(k_pp,j,i)  - u(k-1,j,i)  )                  &
                   +      (           ibit17 * adv_mom_5                        &
                           ) *                                               &
                             ( u(k_ppp,j,i) - u(k_mm,j,i) )                  &
                                              )
!
!--             Calculate the divergence of the velocity field. A respective
!--             correction is needed to overcome numerical instabilities caused
!--             by a not sufficient reduction of divergences near topography.
                div = ( ( u_comp      - ( u(k,j,i)   + u(k,j,i-1)   ) ) * ddx &
                     +  ( v_comp + gv - ( v(k,j,i)   + v(k,j,i-1 )  ) ) * ddy &
                     +  (   w_comp                      * rho_air_zw(k) -     &
                          ( w(k-1,j,i) + w(k-1,j,i-1) ) * rho_air_zw(k-1)     &
                        ) * drho_air(k) * ddzw(k)                             &
                      ) * 0.5_wp

                tend(k,j,i) = tend(k,j,i) - (                                  &
                 ( ( flux_r + diss_r )                                         &
               -   ( flux_l + diss_l ) ) * ddx                                 &
               + ( ( flux_n + diss_n )                                         &
               -   ( flux_s + diss_s ) ) * ddy                                 &
               + ( ( flux_t + diss_t )                                         &
               -   ( flux_d + diss_d ) ) * drho_air(k) * ddzw(k)               &
                                           ) + div * u(k,j,i)

#ifndef _OPENACC
                swap_flux_x_local_u(k,j) = flux_r
                swap_diss_x_local_u(k,j) = diss_r
                swap_flux_y_local_u(k)   = flux_n
                swap_diss_y_local_u(k)   = diss_n
#endif
                flux_d                   = flux_t
                diss_d                   = diss_t
!
!--             Statistical Evaluation of u'u'. The factor has to be applied
!--             for right evaluation when gallilei_trans = .T. .
                !$ACC ATOMIC
                sums_us2_ws_l(k,tn) = sums_us2_ws_l(k,tn)                      &
                + ( flux_r                                                     &
                    * ( u_comp - 2.0_wp * hom(k,1,1,0)                   )     &
                    / ( u_comp - gu + SIGN( 1.0E-20_wp, u_comp - gu )    )     &
                  + diss_r                                                     &
                    *   ABS( u_comp - 2.0_wp * hom(k,1,1,0)              )     &
                    / ( ABS( u_comp - gu ) + 1.0E-20_wp                  )     &
                  ) *   weight_substep(intermediate_timestep_count)
!
!--             Statistical Evaluation of w'u'.
                !$ACC ATOMIC
                sums_wsus_ws_l(k,tn) = sums_wsus_ws_l(k,tn)                    &
                + ( flux_t                                                     &
                    * ( w_comp - 2.0_wp * hom(k,1,3,0)                   )     &
                    / ( w_comp + SIGN( 1.0E-20_wp, w_comp )              )     &
                  + diss_t                                                     &
                    *   ABS( w_comp - 2.0_wp * hom(k,1,3,0)              )     &
                    / ( ABS( w_comp ) + 1.0E-20_wp                       )     &
                  ) *   weight_substep(intermediate_timestep_count)
             ENDDO
          ENDDO
       ENDDO

    END SUBROUTINE advec_u_ws
    

!------------------------------------------------------------------------------!
! Description:
! ------------
!> Advection of v - Call for all grid points
!------------------------------------------------------------------------------!
    SUBROUTINE advec_v_ws

       USE arrays_3d,                                                          &
           ONLY:  ddzw, drho_air, tend, u, v, w, rho_air_zw

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s, intermediate_timestep_count, &
                  u_gtrans, v_gtrans

       USE grid_variables,                                                     &
           ONLY:  ddx, ddy

       USE indices,                                                            &
           ONLY:  nxl, nxr, nyn, nys, nysv, nzb, nzb_max, nzt, advc_flags_1

       USE kinds

       USE statistics,                                                         &
           ONLY:  hom, sums_vs2_ws_l, sums_wsvs_ws_l, weight_substep

       IMPLICIT NONE


       INTEGER(iwp) ::  i         !< grid index along x-direction
       INTEGER(iwp) ::  j         !< grid index along y-direction
       INTEGER(iwp) ::  k         !< grid index along z-direction
       INTEGER(iwp) ::  k_mm      !< k-2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_pp      !< k+2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_ppp     !< k+3 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  nzb_max_l !< index indicating upper bound for order degradation of horizontal advection terms 
       INTEGER(iwp) ::  tn = 0    !< number of OpenMP thread
       
       REAL(wp)    ::  ibit18 !< flag indicating 1st-order scheme along x-direction
       REAL(wp)    ::  ibit19 !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)    ::  ibit20 !< flag indicating 5th-order scheme along x-direction
#ifdef _OPENACC
       REAL(wp)    ::  ibit18_l !< flag indicating 1st-order scheme along x-direction
       REAL(wp)    ::  ibit19_l !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)    ::  ibit20_l !< flag indicating 5th-order scheme along x-direction
#endif
       REAL(wp)    ::  ibit21 !< flag indicating 1st-order scheme along y-direction
       REAL(wp)    ::  ibit22 !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)    ::  ibit23 !< flag indicating 5th-order scheme along y-direction
#ifdef _OPENACC
       REAL(wp)    ::  ibit21_s !< flag indicating 1st-order scheme along y-direction
       REAL(wp)    ::  ibit22_s !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)    ::  ibit23_s !< flag indicating 5th-order scheme along y-direction
#endif
       REAL(wp)    ::  ibit24 !< flag indicating 1st-order scheme along z-direction
       REAL(wp)    ::  ibit25 !< flag indicating 3rd-order scheme along z-direction
       REAL(wp)    ::  ibit26 !< flag indicating 5th-order scheme along z-direction
       REAL(wp)    ::  diss_d !< artificial dissipation term at grid box bottom
       REAL(wp)    ::  div    !< diverence on v-grid
       REAL(wp)    ::  flux_d !< artificial 6th-order flux at grid box bottom
       REAL(wp)    ::  gu     !< Galilei-transformation velocity along x
       REAL(wp)    ::  gv     !< Galilei-transformation velocity along y
       REAL(wp)    ::  u_comp !< advection velocity along x
#ifdef _OPENACC
       REAL(wp)    ::  u_comp_l !< advection velocity along x
#endif
       REAL(wp)    ::  w_comp !< advection velocity along z
       
       REAL(wp)    ::  diss_s !< discretized artificial dissipation at southward-side of the grid box
       REAL(wp)    ::  flux_s !< discretized 6th-order flux at southward-side of the grid box
#ifndef _OPENACC
       REAL(wp), DIMENSION(nzb+1:nzt) ::  swap_diss_y_local_v !< discretized artificial dissipation at southward-side of the grid box
       REAL(wp), DIMENSION(nzb+1:nzt) ::  swap_flux_y_local_v !< discretized 6th-order flux at southward-side of the grid box
#endif
       
       REAL(wp)    ::  diss_l !< discretized artificial dissipation at leftward-side of the grid box
       REAL(wp)    ::  flux_l !< discretized 6th-order flux at leftward-side of the grid box
#ifndef _OPENACC
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn) ::  swap_diss_x_local_v !< discretized artificial dissipation at leftward-side of the grid box
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn) ::  swap_flux_x_local_v !< discretized 6th-order flux at leftward-side of the grid box
#endif
       
       REAL(wp)    ::  diss_n !< discretized artificial dissipation at northward-side of the grid box
       REAL(wp)    ::  diss_r !< discretized artificial dissipation at rightward-side of the grid box
       REAL(wp)    ::  diss_t !< discretized artificial dissipation at top of the grid box
       REAL(wp)    ::  flux_n !< discretized 6th-order flux at northward-side of the grid box
       REAL(wp)    ::  flux_r !< discretized 6th-order flux at rightward-side of the grid box
       REAL(wp)    ::  flux_t !< discretized 6th-order flux at top of the grid box
       REAL(wp)    ::  v_comp !< advection velocity along y
#ifdef _OPENACC
       REAL(wp)    ::  v_comp_s !<
#endif
!
!--    Set local version of nzb_max. At non-cyclic boundaries the order of the 
!--    advection need to be degraded near the boundary. Please note, in contrast
!--    to the cache-optimized routines, nzb_max_l is set constantly for the 
!--    entire subdomain, in order to avoid unsymmetric loops which might be 
!--    an issue for GPUs.
       IF( bc_dirichlet_l  .OR.  bc_radiation_l  .OR.                          &
           bc_dirichlet_r  .OR.  bc_radiation_r  .OR.                          &
           bc_dirichlet_s  .OR.  bc_radiation_s  .OR.                          &
           bc_dirichlet_n  .OR.  bc_radiation_n )  THEN
          nzb_max_l = nzt
       ELSE
          nzb_max_l = nzb_max
       END IF
       
       gu = 2.0_wp * u_gtrans
       gv = 2.0_wp * v_gtrans

#ifndef _OPENACC
!
!--    First compute the whole left boundary of the processor domain
       i = nxl
       DO  j = nysv, nyn
          DO  k = nzb+1, nzb_max_l

             ibit20 = REAL( IBITS(advc_flags_1(k,j,i-1),20,1), KIND = wp )
             ibit19 = REAL( IBITS(advc_flags_1(k,j,i-1),19,1), KIND = wp )
             ibit18 = REAL( IBITS(advc_flags_1(k,j,i-1),18,1), KIND = wp )

             u_comp                   = u(k,j-1,i) + u(k,j,i) - gu
             swap_flux_x_local_v(k,j) = u_comp * (                             &
                                      ( 37.0_wp * ibit20 * adv_mom_5              &
                                   +     7.0_wp * ibit19 * adv_mom_3              &
                                   +              ibit18 * adv_mom_1              &
                                      ) *                                      &
                                     ( v(k,j,i)   + v(k,j,i-1) )               &
                               -      (  8.0_wp * ibit20 * adv_mom_5              &
                                   +              ibit19 * adv_mom_3              &
                                      ) *                                      &
                                     ( v(k,j,i+1) + v(k,j,i-2) )               &
                               +      (           ibit20 * adv_mom_5              &
                                      ) *                                      &
                                     ( v(k,j,i+2) + v(k,j,i-3) )               &
                                                 )

              swap_diss_x_local_v(k,j) = - ABS( u_comp ) * (                   &
                                      ( 10.0_wp * ibit20 * adv_mom_5              &
                                   +     3.0_wp * ibit19 * adv_mom_3              &
                                   +              ibit18 * adv_mom_1              &
                                      ) *                                      &
                                     ( v(k,j,i)   - v(k,j,i-1) )               &
                               -      (  5.0_wp * ibit20 * adv_mom_5              &
                                   +              ibit19 * adv_mom_3              &
                                      ) *                                      &
                                     ( v(k,j,i+1) - v(k,j,i-2) )               &
                               +      (           ibit20 * adv_mom_5              &
                                      ) *                                      &
                                     ( v(k,j,i+2) - v(k,j,i-3) )               &
                                                           )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             u_comp                   = u(k,j-1,i) + u(k,j,i) - gu
             swap_flux_x_local_v(k,j) = u_comp * (                            &
                             37.0_wp * ( v(k,j,i) + v(k,j,i-1)   )               &
                           -  8.0_wp * ( v(k,j,i+1) + v(k,j,i-2) )               &
                           +           ( v(k,j,i+2) + v(k,j,i-3) ) ) * adv_mom_5
             swap_diss_x_local_v(k,j) = - ABS( u_comp ) * (                   &
                             10.0_wp * ( v(k,j,i) - v(k,j,i-1)   )               &
                           -  5.0_wp * ( v(k,j,i+1) - v(k,j,i-2) )               &
                           +           ( v(k,j,i+2) - v(k,j,i-3) ) ) * adv_mom_5

          ENDDO

       ENDDO
#endif

       !$ACC PARALLEL LOOP COLLAPSE(2) FIRSTPRIVATE(tn, gu, gv) &
       !$ACC PRIVATE(i, j, k, k_mm, k_pp, k_ppp) &
       !$ACC PRIVATE(ibit18, ibit19, ibit20, ibit21, ibit22, ibit23) &
       !$ACC PRIVATE(ibit18_l, ibit19_l, ibit20_l) &
       !$ACC PRIVATE(ibit21_s, ibit22_s, ibit23_s) &
       !$ACC PRIVATE(ibit24, ibit25, ibit26) &
       !$ACC PRIVATE(flux_r, diss_r, flux_l, diss_l) &
       !$ACC PRIVATE(flux_n, diss_n, flux_s, diss_s) &
       !$ACC PRIVATE(flux_t, diss_t, flux_d, diss_d) &
       !$ACC PRIVATE(div, u_comp, u_comp_l, v_comp, v_comp_s, w_comp) &
       !$ACC PRESENT(advc_flags_1) &
       !$ACC PRESENT(u, v, w) &
       !$ACC PRESENT(drho_air, rho_air_zw, ddzw) &
       !$ACC PRESENT(tend) &
       !$ACC PRESENT(hom(nzb+1:nzb_max_l,1,2:3,0)) &
       !$ACC PRESENT(weight_substep(intermediate_timestep_count)) &
       !$ACC PRESENT(sums_vs2_ws_l, sums_wsvs_ws_l)
       DO i = nxl, nxr

#ifndef _OPENACC
          j = nysv
          DO  k = nzb+1, nzb_max_l

             ibit23 = REAL( IBITS(advc_flags_1(k,j-1,i),23,1), KIND = wp )
             ibit22 = REAL( IBITS(advc_flags_1(k,j-1,i),22,1), KIND = wp )
             ibit21 = REAL( IBITS(advc_flags_1(k,j-1,i),21,1), KIND = wp )

             v_comp                 = v(k,j,i) + v(k,j-1,i) - gv
             swap_flux_y_local_v(k) = v_comp * (                              &
                                   ( 37.0_wp * ibit23 * adv_mom_5                &
                                +     7.0_wp * ibit22 * adv_mom_3                &
                                +              ibit21 * adv_mom_1                &
                                   ) *                                        &
                                     ( v(k,j,i)   + v(k,j-1,i) )              &
                            -      (  8.0_wp * ibit23 * adv_mom_5                &
                                +              ibit22 * adv_mom_3                &
                                   ) *                                        &
                                     ( v(k,j+1,i) + v(k,j-2,i) )              &
                            +      (           ibit23 * adv_mom_5                &
                                   ) *                                        &
                                     ( v(k,j+2,i) + v(k,j-3,i) )              &
                                                 )

             swap_diss_y_local_v(k) = - ABS( v_comp ) * (                     &
                                   ( 10.0_wp * ibit23 * adv_mom_5                &
                                +     3.0_wp * ibit22 * adv_mom_3                &
                                +              ibit21 * adv_mom_1                &
                                   ) *                                        &
                                     ( v(k,j,i)   - v(k,j-1,i) )              &
                            -      (  5.0_wp * ibit23 * adv_mom_5                &
                                +              ibit22 * adv_mom_3                &
                                   ) *                                        &
                                     ( v(k,j+1,i) - v(k,j-2,i) )              &
                            +      (           ibit23 * adv_mom_5                &
                                   ) *                                        &
                                     ( v(k,j+2,i) - v(k,j-3,i) )              &
                                                          )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             v_comp                 = v(k,j,i) + v(k,j-1,i) - gv
             swap_flux_y_local_v(k) = v_comp * (                              &
                           37.0_wp * ( v(k,j,i) + v(k,j-1,i)   )                 &
                         -  8.0_wp * ( v(k,j+1,i) + v(k,j-2,i) )                 &
                         +           ( v(k,j+2,i) + v(k,j-3,i) ) ) * adv_mom_5
             swap_diss_y_local_v(k) = - ABS( v_comp ) * (                     &
                           10.0_wp * ( v(k,j,i) - v(k,j-1,i)   )                 &
                         -  5.0_wp * ( v(k,j+1,i) - v(k,j-2,i) )                 &
                         +           ( v(k,j+2,i) - v(k,j-3,i) ) ) * adv_mom_5

          ENDDO
#endif

          DO  j = nysv, nyn

             flux_d    = 0.0_wp
             diss_d    = 0.0_wp

             DO  k = nzb+1, nzb_max_l

                ibit20 = REAL( IBITS(advc_flags_1(k,j,i),20,1), KIND = wp )
                ibit19 = REAL( IBITS(advc_flags_1(k,j,i),19,1), KIND = wp )
                ibit18 = REAL( IBITS(advc_flags_1(k,j,i),18,1), KIND = wp )

                u_comp = u(k,j-1,i+1) + u(k,j,i+1) - gu
                flux_r = u_comp * (                                          &
                          ( 37.0_wp * ibit20 * adv_mom_5                        &
                       +     7.0_wp * ibit19 * adv_mom_3                        &
                       +              ibit18 * adv_mom_1                        &
                          ) *                                                &
                                 ( v(k,j,i+1) + v(k,j,i)   )                 &
                   -      (  8.0_wp * ibit20 * adv_mom_5                        &
                       +              ibit19 * adv_mom_3                        &
                          ) *                                                &
                                 ( v(k,j,i+2) + v(k,j,i-1) )                 &
                   +      (           ibit20 * adv_mom_5                        &
                          ) *                                                &
                                 ( v(k,j,i+3) + v(k,j,i-2) )                 &
                                     )

                diss_r = - ABS( u_comp ) * (                                 &
                          ( 10.0_wp * ibit20 * adv_mom_5                        &
                       +     3.0_wp * ibit19 * adv_mom_3                        &
                       +              ibit18 * adv_mom_1                        &
                          ) *                                                &
                                 ( v(k,j,i+1) - v(k,j,i)  )                  &
                   -      (  5.0_wp * ibit20 * adv_mom_5                        &
                       +              ibit19 * adv_mom_3                        &
                          ) *                                                &
                                 ( v(k,j,i+2) - v(k,j,i-1) )                 &
                   +      (           ibit20 * adv_mom_5                        &
                          ) *                                                &
                                 ( v(k,j,i+3) - v(k,j,i-2) )                 &
                                              )

#ifdef _OPENACC
!
!--             Recompute the left fluxes.
                ibit20_l = REAL( IBITS(advc_flags_1(k,j,i-1),20,1), KIND = wp )
                ibit19_l = REAL( IBITS(advc_flags_1(k,j,i-1),19,1), KIND = wp )
                ibit18_l = REAL( IBITS(advc_flags_1(k,j,i-1),18,1), KIND = wp )

                u_comp_l  = u(k,j-1,i) + u(k,j,i) - gu
                flux_l    = u_comp_l * (                                     &
                                      ( 37.0_wp * ibit20_l * adv_mom_5          &
                                   +     7.0_wp * ibit19_l * adv_mom_3          &
                                   +              ibit18_l * adv_mom_1          &
                                      ) *                                    &
                                     ( v(k,j,i)   + v(k,j,i-1) )             &
                               -      (  8.0_wp * ibit20_l * adv_mom_5          &
                                   +              ibit19_l * adv_mom_3          &
                                      ) *                                    &
                                     ( v(k,j,i+1) + v(k,j,i-2) )             &
                               +      (           ibit20_l * adv_mom_5          &
                                      ) *                                    &
                                     ( v(k,j,i+2) + v(k,j,i-3) )             &
                                                 )

                 diss_l   = - ABS( u_comp_l ) * (                            &
                                      ( 10.0_wp * ibit20_l * adv_mom_5          &
                                   +     3.0_wp * ibit19_l * adv_mom_3          &
                                   +              ibit18_l * adv_mom_1          &
                                      ) *                                    &
                                     ( v(k,j,i)   - v(k,j,i-1) )             &
                               -      (  5.0_wp * ibit20_l * adv_mom_5          &
                                   +              ibit19_l * adv_mom_3          &
                                      ) *                                    &
                                     ( v(k,j,i+1) - v(k,j,i-2) )             &
                               +      (           ibit20_l * adv_mom_5          &
                                      ) *                                    &
                                     ( v(k,j,i+2) - v(k,j,i-3) )             &
                                                           )
#else
                flux_l = swap_flux_x_local_v(k,j)
                diss_l = swap_diss_x_local_v(k,j)
#endif

                ibit23 = REAL( IBITS(advc_flags_1(k,j,i),23,1), KIND = wp )
                ibit22 = REAL( IBITS(advc_flags_1(k,j,i),22,1), KIND = wp )
                ibit21 = REAL( IBITS(advc_flags_1(k,j,i),21,1), KIND = wp )

                v_comp = v(k,j+1,i) + v(k,j,i)
                flux_n = ( v_comp - gv ) * (                                 &
                          ( 37.0_wp * ibit23 * adv_mom_5                        &
                       +     7.0_wp * ibit22 * adv_mom_3                        &
                       +              ibit21 * adv_mom_1                        &
                          ) *                                                &
                                 ( v(k,j+1,i) + v(k,j,i)   )                 &
                   -      (  8.0_wp * ibit23 * adv_mom_5                        &
                       +              ibit22 * adv_mom_3                        &
                          ) *                                                &
                                 ( v(k,j+2,i) + v(k,j-1,i) )                 &
                   +      (           ibit23 * adv_mom_5                        &
                          ) *                                                &
                                 ( v(k,j+3,i) + v(k,j-2,i) )                 &
                                     )

                diss_n = - ABS( v_comp - gv ) * (                            &
                          ( 10.0_wp * ibit23 * adv_mom_5                        &
                       +     3.0_wp * ibit22 * adv_mom_3                        &
                       +              ibit21 * adv_mom_1                        &
                          ) *                                                &
                                 ( v(k,j+1,i) - v(k,j,i)  )                  &
                   -      (  5.0_wp * ibit23 * adv_mom_5                        &
                       +              ibit22 * adv_mom_3                        &
                          ) *                                                &
                                 ( v(k,j+2,i) - v(k,j-1,i) )                 &
                   +      (           ibit23 * adv_mom_5                        &
                          ) *                                                &
                                 ( v(k,j+3,i) - v(k,j-2,i) )                 &
                                                      )

#ifdef _OPENACC
!
!--             Recompute the south fluxes.
                ibit23_s = REAL( IBITS(advc_flags_1(k,j-1,i),23,1), KIND = wp )
                ibit22_s = REAL( IBITS(advc_flags_1(k,j-1,i),22,1), KIND = wp )
                ibit21_s = REAL( IBITS(advc_flags_1(k,j-1,i),21,1), KIND = wp )

                v_comp_s = v(k,j,i) + v(k,j-1,i) - gv
                flux_s   = v_comp_s * (                                      &
                                   ( 37.0_wp * ibit23_s * adv_mom_5             &
                                +     7.0_wp * ibit22_s * adv_mom_3             &
                                +              ibit21_s * adv_mom_1             &
                                   ) *                                       &
                                     ( v(k,j,i)   + v(k,j-1,i) )             &
                            -      (  8.0_wp * ibit23_s * adv_mom_5             &
                                +              ibit22_s * adv_mom_3             &
                                   ) *                                       &
                                     ( v(k,j+1,i) + v(k,j-2,i) )             &
                            +      (           ibit23_s * adv_mom_5             &
                                   ) *                                       &
                                     ( v(k,j+2,i) + v(k,j-3,i) )             &
                                                 )

                diss_s   = - ABS( v_comp_s ) * (                             &
                                   ( 10.0_wp * ibit23_s * adv_mom_5             &
                                +     3.0_wp * ibit22_s * adv_mom_3             &
                                +              ibit21_s * adv_mom_1             &
                                   ) *                                       &
                                     ( v(k,j,i)   - v(k,j-1,i) )             &
                            -      (  5.0_wp * ibit23_s * adv_mom_5             &
                                +              ibit22_s * adv_mom_3             &
                                   ) *                                       &
                                     ( v(k,j+1,i) - v(k,j-2,i) )             &
                            +      (           ibit23_s * adv_mom_5             &
                                   ) *                                       &
                                     ( v(k,j+2,i) - v(k,j-3,i) )             &
                                                          )
#else
               flux_s = swap_flux_y_local_v(k)
               diss_s = swap_diss_y_local_v(k)
#endif

!
!--             k index has to be modified near bottom and top, else array
!--             subscripts will be exceeded.
                ibit26 = REAL( IBITS(advc_flags_1(k,j,i),26,1), KIND = wp )
                ibit25 = REAL( IBITS(advc_flags_1(k,j,i),25,1), KIND = wp )
                ibit24 = REAL( IBITS(advc_flags_1(k,j,i),24,1), KIND = wp )

                k_ppp = k + 3 * ibit26
                k_pp  = k + 2 * ( 1 - ibit24  )
                k_mm  = k - 2 * ibit26

                w_comp = w(k,j-1,i) + w(k,j,i)
                flux_t = w_comp * rho_air_zw(k) * (                          &
                          ( 37.0_wp * ibit26 * adv_mom_5                        &
                       +     7.0_wp * ibit25 * adv_mom_3                        &
                       +              ibit24 * adv_mom_1                        &
                          ) *                                                &
                             ( v(k+1,j,i)   + v(k,j,i)    )                  &
                   -      (  8.0_wp * ibit26 * adv_mom_5                        &
                       +              ibit25 * adv_mom_3                        &
                          ) *                                                &
                             ( v(k_pp,j,i)  + v(k-1,j,i)  )                  &
                   +      (           ibit26 * adv_mom_5                        &
                          ) *                                                &
                             ( v(k_ppp,j,i) + v(k_mm,j,i) )                  &
                                      )

                diss_t = - ABS( w_comp ) * rho_air_zw(k) * (                 &
                          ( 10.0_wp * ibit26 * adv_mom_5                        &
                       +     3.0_wp * ibit25 * adv_mom_3                        &
                       +              ibit24 * adv_mom_1                        &
                          ) *                                                &
                             ( v(k+1,j,i)   - v(k,j,i)    )                  &
                   -      (  5.0_wp * ibit26 * adv_mom_5                        &
                       +              ibit25 * adv_mom_3                        &
                          ) *                                                &
                             ( v(k_pp,j,i)  - v(k-1,j,i)  )                  &
                   +      (           ibit26 * adv_mom_5                        &
                          ) *                                                &
                             ( v(k_ppp,j,i) - v(k_mm,j,i) )                  &
                                               )
!
!--             Calculate the divergence of the velocity field. A respective
!--             correction is needed to overcome numerical instabilities caused
!--             by a not sufficient reduction of divergences near topography.
                div = ( ( ( u_comp     + gu )                                 &
                                       * ( ibit18 + ibit19 + ibit20 )         &
                - ( u(k,j-1,i)   + u(k,j,i) )                                 &
                                       * (                                    &
                        REAL( IBITS(advc_flags_1(k,j,i-1),18,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k,j,i-1),19,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k,j,i-1),20,1), KIND = wp )  &
                                         )                                    &
                  ) * ddx                                                     &
               +  ( v_comp                                                    &
                                       * ( ibit21 + ibit22 + ibit23 )         &
                - ( v(k,j,i)     + v(k,j-1,i) )                               &
                                       * (                                    &
                        REAL( IBITS(advc_flags_1(k,j-1,i),21,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k,j-1,i),22,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k,j-1,i),23,1), KIND = wp )  &
                                         )                                    &
                  ) * ddy                                                     &
               +  ( w_comp * rho_air_zw(k)                                    &
                                       * ( ibit24 + ibit25 + ibit26 )         &
                - ( w(k-1,j-1,i) + w(k-1,j,i) ) * rho_air_zw(k-1)             &
                                       * (                                    &
                        REAL( IBITS(advc_flags_1(k-1,j,i),24,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k-1,j,i),25,1), KIND = wp )  &
                      + REAL( IBITS(advc_flags_1(k-1,j,i),26,1), KIND = wp )  &
                                         )                                    &
                   ) * drho_air(k) * ddzw(k)                                  &
                ) * 0.5_wp


                tend(k,j,i) = tend(k,j,i) - (                                 &
                       ( ( flux_r + diss_r )                                  &
                     -   ( flux_l + diss_l ) ) * ddx                          &
                     + ( ( flux_n + diss_n )                                  &
                     -   ( flux_s + diss_s ) ) * ddy                          &
                     + ( ( flux_t + diss_t )                                  &
                     -   ( flux_d + diss_d ) ) * drho_air(k) * ddzw(k)        &
                                            )  + v(k,j,i) * div

#ifndef _OPENACC
                swap_flux_x_local_v(k,j) = flux_r
                swap_diss_x_local_v(k,j) = diss_r
                swap_flux_y_local_v(k)   = flux_n
                swap_diss_y_local_v(k)   = diss_n
#endif
                flux_d                   = flux_t
                diss_d                   = diss_t

!
!--             Statistical Evaluation of v'v'. The factor has to be applied
!--             for right evaluation when gallilei_trans = .T. .
                !$ACC ATOMIC
                sums_vs2_ws_l(k,tn) = sums_vs2_ws_l(k,tn)                      &
                + ( flux_n                                                     &
                    * ( v_comp - 2.0_wp * hom(k,1,2,0)                   )     &
                    / ( v_comp - gv + SIGN( 1.0E-20_wp, v_comp - gv )    )     &
               +   diss_n                                                      &
                    *   ABS( v_comp - 2.0_wp * hom(k,1,2,0)              )     &
                    / ( ABS( v_comp - gv ) + 1.0E-20_wp                  )     &
                  ) *   weight_substep(intermediate_timestep_count)
!
!--             Statistical Evaluation of w'u'.
                !$ACC ATOMIC
                sums_wsvs_ws_l(k,tn) = sums_wsvs_ws_l(k,tn)                    &
                + ( flux_t                                                     &
                    * ( w_comp - 2.0_wp * hom(k,1,3,0)                   )     &
                    / ( w_comp + SIGN( 1.0E-20_wp, w_comp )              )     &
               +   diss_t                                                      &
                    *   ABS( w_comp - 2.0_wp * hom(k,1,3,0)              )     &
                    / ( ABS( w_comp ) + 1.0E-20_wp                       )     &
                  ) *   weight_substep(intermediate_timestep_count)

             ENDDO

             DO  k = nzb_max_l+1, nzt

                u_comp = u(k,j-1,i+1) + u(k,j,i+1) - gu
                flux_r = u_comp * (                                           &
                      37.0_wp * ( v(k,j,i+1) + v(k,j,i)   )                      &
                    -  8.0_wp * ( v(k,j,i+2) + v(k,j,i-1) )                      &
                    +           ( v(k,j,i+3) + v(k,j,i-2) ) ) * adv_mom_5

                diss_r = - ABS( u_comp ) * (                                  &
                      10.0_wp * ( v(k,j,i+1) - v(k,j,i) )                        &
                    -  5.0_wp * ( v(k,j,i+2) - v(k,j,i-1) )                      &
                    +           ( v(k,j,i+3) - v(k,j,i-2) ) ) * adv_mom_5

#ifdef _OPENACC
!
!--             Recompute the left fluxes.
                u_comp_l = u(k,j-1,i) + u(k,j,i) - gu
                flux_l   = u_comp_l * (                                       &
                             37.0_wp * ( v(k,j,i) + v(k,j,i-1)   )               &
                           -  8.0_wp * ( v(k,j,i+1) + v(k,j,i-2) )               &
                           +           ( v(k,j,i+2) + v(k,j,i-3) ) ) * adv_mom_5
                diss_l   = - ABS( u_comp_l ) * (                              &
                             10.0_wp * ( v(k,j,i) - v(k,j,i-1)   )               &
                           -  5.0_wp * ( v(k,j,i+1) - v(k,j,i-2) )               &
                           +           ( v(k,j,i+2) - v(k,j,i-3) ) ) * adv_mom_5
#else
                flux_l = swap_flux_x_local_v(k,j)
                diss_l = swap_diss_x_local_v(k,j)
#endif

                v_comp = v(k,j+1,i) + v(k,j,i)
                flux_n = ( v_comp - gv ) * (                                  &
                      37.0_wp * ( v(k,j+1,i) + v(k,j,i)   )                      &
                    -  8.0_wp * ( v(k,j+2,i) + v(k,j-1,i) )                      &
                      +         ( v(k,j+3,i) + v(k,j-2,i) ) ) * adv_mom_5

                diss_n = - ABS( v_comp - gv ) * (                             &
                      10.0_wp * ( v(k,j+1,i) - v(k,j,i)   )                      &
                    -  5.0_wp * ( v(k,j+2,i) - v(k,j-1,i) )                      &
                    +           ( v(k,j+3,i) - v(k,j-2,i) ) ) * adv_mom_5

#ifdef _OPENACC
!
!--             Recompute the south fluxes.
                v_comp_s = v(k,j,i) + v(k,j-1,i) - gv
                flux_s   = v_comp_s * (                                       &
                           37.0_wp * ( v(k,j,i) + v(k,j-1,i)   )                 &
                         -  8.0_wp * ( v(k,j+1,i) + v(k,j-2,i) )                 &
                         +           ( v(k,j+2,i) + v(k,j-3,i) ) ) * adv_mom_5
                diss_s   = - ABS( v_comp_s ) * (                              &
                           10.0_wp * ( v(k,j,i) - v(k,j-1,i)   )                 &
                         -  5.0_wp * ( v(k,j+1,i) - v(k,j-2,i) )                 &
                         +           ( v(k,j+2,i) - v(k,j-3,i) ) ) * adv_mom_5
#else
                flux_s = swap_flux_y_local_v(k)
                diss_s = swap_diss_y_local_v(k)
#endif

!
!--             k index has to be modified near bottom and top, else array
!--             subscripts will be exceeded.
                ibit26 = REAL( IBITS(advc_flags_1(k,j,i),26,1), KIND = wp )
                ibit25 = REAL( IBITS(advc_flags_1(k,j,i),25,1), KIND = wp )
                ibit24 = REAL( IBITS(advc_flags_1(k,j,i),24,1), KIND = wp )

                k_ppp = k + 3 * ibit26
                k_pp  = k + 2 * ( 1 - ibit24  )
                k_mm  = k - 2 * ibit26

                w_comp = w(k,j-1,i) + w(k,j,i)
                flux_t = w_comp * rho_air_zw(k) * (                          &
                          ( 37.0_wp * ibit26 * adv_mom_5                        &
                       +     7.0_wp * ibit25 * adv_mom_3                        &
                       +              ibit24 * adv_mom_1                        &
                          ) *                                                &
                             ( v(k+1,j,i)   + v(k,j,i)    )                  &
                   -      (  8.0_wp * ibit26 * adv_mom_5                        &
                       +              ibit25 * adv_mom_3                        &
                          ) *                                                &
                             ( v(k_pp,j,i)  + v(k-1,j,i)  )                  &
                   +      (           ibit26 * adv_mom_5                        &
                          ) *                                                &
                             ( v(k_ppp,j,i) + v(k_mm,j,i) )                  &
                                      )

                diss_t = - ABS( w_comp ) * rho_air_zw(k) * (                 &
                          ( 10.0_wp * ibit26 * adv_mom_5                        &
                       +     3.0_wp * ibit25 * adv_mom_3                        &
                       +              ibit24 * adv_mom_1                        &
                          ) *                                                &
                             ( v(k+1,j,i)   - v(k,j,i)    )                  &
                   -      (  5.0_wp * ibit26 * adv_mom_5                        &
                       +              ibit25 * adv_mom_3                        &
                          ) *                                                &
                             ( v(k_pp,j,i)  - v(k-1,j,i)  )                  &
                   +      (           ibit26 * adv_mom_5                        &
                          ) *                                                &
                             ( v(k_ppp,j,i) - v(k_mm,j,i) )                  &
                                               )
!
!--             Calculate the divergence of the velocity field. A respective
!--             correction is needed to overcome numerical instabilities caused
!--             by a not sufficient reduction of divergences near topography.
                div = ( ( u_comp + gu - ( u(k,j-1,i)   + u(k,j,i)   ) ) * ddx &
                     +  ( v_comp      - ( v(k,j,i)     + v(k,j-1,i) ) ) * ddy &
                     +  (   w_comp                      * rho_air_zw(k) -     &
                          ( w(k-1,j-1,i) + w(k-1,j,i) ) * rho_air_zw(k-1)     &
                        ) * drho_air(k) * ddzw(k)                             &
                      ) * 0.5_wp
 
                tend(k,j,i) = tend(k,j,i) - (                                 &
                       ( ( flux_r + diss_r )                                  &
                     -   ( flux_l + diss_l ) ) * ddx                          &
                     + ( ( flux_n + diss_n )                                  &
                     -   ( flux_s + diss_s ) ) * ddy                          &
                     + ( ( flux_t + diss_t )                                  &
                     -   ( flux_d + diss_d ) ) * drho_air(k) * ddzw(k)        &
                                            )  + v(k,j,i) * div

#ifndef _OPENACC
                swap_flux_x_local_v(k,j) = flux_r
                swap_diss_x_local_v(k,j) = diss_r
                swap_flux_y_local_v(k)   = flux_n
                swap_diss_y_local_v(k)   = diss_n
#endif
                flux_d                   = flux_t
                diss_d                   = diss_t

!
!--             Statistical Evaluation of v'v'. The factor has to be applied
!--             for right evaluation when gallilei_trans = .T. .
                !$ACC ATOMIC
                sums_vs2_ws_l(k,tn) = sums_vs2_ws_l(k,tn)                      &
                + ( flux_n                                                     &
                    * ( v_comp - 2.0_wp * hom(k,1,2,0)                   )     &
                    / ( v_comp - gv + SIGN( 1.0E-20_wp, v_comp - gv )    )     &
               +   diss_n                                                      &
                    *   ABS( v_comp - 2.0_wp * hom(k,1,2,0)              )     &
                    / ( ABS( v_comp - gv ) + 1.0E-20_wp                  )     &
                  ) *   weight_substep(intermediate_timestep_count)
!
!--             Statistical Evaluation of w'u'.
                !$ACC ATOMIC
                sums_wsvs_ws_l(k,tn) = sums_wsvs_ws_l(k,tn)                    &
                + ( flux_t                                                     &
                    * ( w_comp - 2.0_wp * hom(k,1,3,0)                   )     &
                    / ( w_comp + SIGN( 1.0E-20_wp, w_comp )              )     &
               +   diss_t                                                      &
                    *   ABS( w_comp - 2.0_wp * hom(k,1,3,0)              )     &
                    / ( ABS( w_comp ) + 1.0E-20_wp                       )     &
                  ) *   weight_substep(intermediate_timestep_count)

             ENDDO
          ENDDO
       ENDDO

    END SUBROUTINE advec_v_ws
    
    
!------------------------------------------------------------------------------!
! Description:
! ------------
!> Advection of w - Call for all grid points
!------------------------------------------------------------------------------!
    SUBROUTINE advec_w_ws

       USE arrays_3d,                                                          &
           ONLY:  ddzu, drho_air_zw, tend, u, v, w, rho_air

       USE control_parameters,                                                 &
           ONLY:  bc_dirichlet_l, bc_dirichlet_n, bc_dirichlet_r,              &
                  bc_dirichlet_s, bc_radiation_l, bc_radiation_n,              &
                  bc_radiation_r, bc_radiation_s, intermediate_timestep_count, &
                  u_gtrans, v_gtrans

       USE grid_variables,                                                     &
           ONLY:  ddx, ddy

       USE indices,                                                            &
           ONLY:  nxl, nxr, nyn, nys, nzb, nzb_max, nzt, advc_flags_1,         &
                  advc_flags_2

       USE kinds
       
       USE statistics,                                                         &
           ONLY:  hom, sums_ws2_ws_l, weight_substep

       IMPLICIT NONE

       INTEGER(iwp) ::  i         !< grid index along x-direction
       INTEGER(iwp) ::  j         !< grid index along y-direction
       INTEGER(iwp) ::  k         !< grid index along z-direction
       INTEGER(iwp) ::  k_mm      !< k-2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_pp      !< k+2 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  k_ppp     !< k+3 index in disretization, can be modified to avoid segmentation faults
       INTEGER(iwp) ::  nzb_max_l !< index indicating upper bound for order degradation of horizontal advection terms  
       INTEGER(iwp) ::  tn = 0    !< number of OpenMP thread
       
       REAL(wp)    ::  ibit27 !< flag indicating 1st-order scheme along x-direction
       REAL(wp)    ::  ibit28 !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)    ::  ibit29 !< flag indicating 5th-order scheme along x-direction
#ifdef _OPENACC
       REAL(wp)    ::  ibit27_l !< flag indicating 1st-order scheme along x-direction
       REAL(wp)    ::  ibit28_l !< flag indicating 3rd-order scheme along x-direction
       REAL(wp)    ::  ibit29_l !< flag indicating 5th-order scheme along x-direction
#endif
       REAL(wp)    ::  ibit30 !< flag indicating 1st-order scheme along y-direction
       REAL(wp)    ::  ibit31 !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)    ::  ibit32 !< flag indicating 5th-order scheme along y-direction
#ifdef _OPENACC
       REAL(wp)    ::  ibit30_s !< flag indicating 1st-order scheme along y-direction
       REAL(wp)    ::  ibit31_s !< flag indicating 3rd-order scheme along y-direction
       REAL(wp)    ::  ibit32_s !< flag indicating 5th-order scheme along y-direction
#endif
       REAL(wp)    ::  ibit33 !< flag indicating 1st-order scheme along z-direction
       REAL(wp)    ::  ibit34 !< flag indicating 3rd-order scheme along z-direction
       REAL(wp)    ::  ibit35 !< flag indicating 5th-order scheme along z-direction
       REAL(wp)    ::  diss_d !< artificial dissipation term at grid box bottom
       REAL(wp)    ::  div    !< divergence on w-grid
       REAL(wp)    ::  flux_d !< 6th-order flux at grid box bottom
       REAL(wp)    ::  gu     !< Galilei-transformation velocity along x
       REAL(wp)    ::  gv     !< Galilei-transformation velocity along y
       REAL(wp)    ::  u_comp !< advection velocity along x
#ifdef _OPENACC
       REAL(wp)    ::  u_comp_l !< advection velocity along x
#endif
       REAL(wp)    ::  v_comp !< advection velocity along y
#ifdef _OPENACC
       REAL(wp)    ::  v_comp_s !< advection velocity along y
#endif
       REAL(wp)    ::  w_comp !< advection velocity along z
       
       REAL(wp)    ::  diss_t !< discretized artificial dissipation at top of the grid box
       REAL(wp)    ::  flux_t !< discretized 6th-order flux at top of the grid box
       
       REAL(wp)    ::  diss_n !< discretized artificial dissipation at northward-side of the grid box
       REAL(wp)    ::  diss_r !< discretized artificial dissipation at rightward-side of the grid box
       REAL(wp)    ::  flux_n !< discretized 6th-order flux at northward-side of the grid box
       REAL(wp)    ::  flux_r !< discretized 6th-order flux at rightward-side of the grid box

       REAL(wp)    ::  diss_s !< discretized artificial dissipation at southward-side of the grid box
       REAL(wp)    ::  flux_s !< discretized 6th-order flux at southward-side of the grid box
#ifndef _OPENACC
       REAL(wp), DIMENSION(nzb+1:nzt)  ::  swap_diss_y_local_w !< discretized artificial dissipation at southward-side of the grid box
       REAL(wp), DIMENSION(nzb+1:nzt)  ::  swap_flux_y_local_w !< discretized 6th-order flux at southward-side of the grid box
#endif
       
       REAL(wp)    ::  diss_l !< discretized artificial dissipation at leftward-side of the grid box
       REAL(wp)    ::  flux_l !< discretized 6th-order flux at leftward-side of the grid box
#ifndef _OPENACC
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn) ::  swap_diss_x_local_w !< discretized artificial dissipation at leftward-side of the grid box
       REAL(wp), DIMENSION(nzb+1:nzt,nys:nyn) ::  swap_flux_x_local_w !< discretized 6th-order flux at leftward-side of the grid box
#endif
!
!--    Set local version of nzb_max. At non-cyclic boundaries the order of the 
!--    advection need to be degraded near the boundary. Please note, in contrast
!--    to the cache-optimized routines, nzb_max_l is set constantly for the 
!--    entire subdomain, in order to avoid unsymmetric loops which might be 
!--    an issue for GPUs.
       IF( bc_dirichlet_l  .OR.  bc_radiation_l  .OR.                          &
           bc_dirichlet_r  .OR.  bc_radiation_r  .OR.                          &
           bc_dirichlet_s  .OR.  bc_radiation_s  .OR.                          &
           bc_dirichlet_n  .OR.  bc_radiation_n )  THEN
          nzb_max_l = nzt
       ELSE
          nzb_max_l = nzb_max
       END IF
       
       gu = 2.0_wp * u_gtrans
       gv = 2.0_wp * v_gtrans

#ifndef _OPENACC
!
!--   compute the whole left boundary of the processor domain
       i = nxl
       DO  j = nys, nyn
          DO  k = nzb+1, nzb_max_l

             ibit29 = REAL( IBITS(advc_flags_1(k,j,i-1),29,1), KIND = wp )
             ibit28 = REAL( IBITS(advc_flags_1(k,j,i-1),28,1), KIND = wp )
             ibit27 = REAL( IBITS(advc_flags_1(k,j,i-1),27,1), KIND = wp )

             u_comp                   = u(k+1,j,i) + u(k,j,i) - gu
             swap_flux_x_local_w(k,j) = u_comp * (                             &
                                      ( 37.0_wp * ibit29 * adv_mom_5              &
                                   +     7.0_wp * ibit28 * adv_mom_3              &
                                   +              ibit27 * adv_mom_1              &
                                      ) *                                      &
                                     ( w(k,j,i)   + w(k,j,i-1) )               &
                               -      (  8.0_wp * ibit29 * adv_mom_5              &
                                   +              ibit28 * adv_mom_3              &
                                      ) *                                      &
                                     ( w(k,j,i+1) + w(k,j,i-2) )               &
                               +      (           ibit29 * adv_mom_5              &
                                      ) *                                      &
                                     ( w(k,j,i+2) + w(k,j,i-3) )               &
                                                 )

               swap_diss_x_local_w(k,j) = - ABS( u_comp ) * (                  &
                                        ( 10.0_wp * ibit29 * adv_mom_5            &
                                     +     3.0_wp * ibit28 * adv_mom_3            &
                                     +              ibit27 * adv_mom_1            &
                                        ) *                                    &
                                     ( w(k,j,i)   - w(k,j,i-1) )               &
                                 -      (  5.0_wp * ibit29 * adv_mom_5            &
                                     +              ibit28 * adv_mom_3            &
                                        ) *                                    &
                                     ( w(k,j,i+1) - w(k,j,i-2) )               &
                                 +      (           ibit29 * adv_mom_5            &
                                        ) *                                    &
                                     ( w(k,j,i+2) - w(k,j,i-3) )               &
                                                            )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             u_comp                   = u(k+1,j,i) + u(k,j,i) - gu
             swap_flux_x_local_w(k,j) = u_comp * (                             &
                            37.0_wp * ( w(k,j,i) + w(k,j,i-1)   )                 &
                          -  8.0_wp * ( w(k,j,i+1) + w(k,j,i-2) )                 &
                          +           ( w(k,j,i+2) + w(k,j,i-3) ) ) * adv_mom_5
             swap_diss_x_local_w(k,j) = - ABS( u_comp ) * (                    &
                            10.0_wp * ( w(k,j,i) - w(k,j,i-1)   )                 &
                          -  5.0_wp * ( w(k,j,i+1) - w(k,j,i-2) )                 &
                          +           ( w(k,j,i+2) - w(k,j,i-3) ) ) * adv_mom_5

          ENDDO

       ENDDO
#endif

       !$ACC PARALLEL LOOP COLLAPSE(2) FIRSTPRIVATE(tn, gu, gv) &
       !$ACC PRIVATE(i, j, k, k_mm, k_pp, k_ppp) &
       !$ACC PRIVATE(ibit27, ibit28, ibit29, ibit30, ibit31, ibit32) &
       !$ACC PRIVATE(ibit27_l, ibit28_l, ibit29_l) &
       !$ACC PRIVATE(ibit30_s, ibit31_s, ibit32_s) &
       !$ACC PRIVATE(ibit33, ibit34, ibit35) &
       !$ACC PRIVATE(flux_r, diss_r, flux_l, diss_l) &
       !$ACC PRIVATE(flux_n, diss_n, flux_s, diss_s) &
       !$ACC PRIVATE(flux_t, diss_t, flux_d, diss_d) &
       !$ACC PRIVATE(div, u_comp, u_comp_l, v_comp, v_comp_s, w_comp) &
       !$ACC PRESENT(advc_flags_1, advc_flags_2) &
       !$ACC PRESENT(u, v, w) &
       !$ACC PRESENT(rho_air, drho_air_zw, ddzu) &
       !$ACC PRESENT(tend) &
       !$ACC PRESENT(hom(nzb+1:nzb_max_l,1,3,0)) &
       !$ACC PRESENT(weight_substep(intermediate_timestep_count)) &
       !$ACC PRESENT(sums_ws2_ws_l(nzb+1:nzb_max_l,0))
       DO i = nxl, nxr

#ifndef _OPENACC
          j = nys
          DO  k = nzb+1, nzb_max_l

             ibit32 = REAL( IBITS(advc_flags_2(k,j-1,i),0,1),  KIND = wp )
             ibit31 = REAL( IBITS(advc_flags_1(k,j-1,i),31,1), KIND = wp )
             ibit30 = REAL( IBITS(advc_flags_1(k,j-1,i),30,1), KIND = wp )

             v_comp                 = v(k+1,j,i) + v(k,j,i) - gv
             swap_flux_y_local_w(k) = v_comp * (                              &
                                    ( 37.0_wp * ibit32 * adv_mom_5               &
                                 +     7.0_wp * ibit31 * adv_mom_3               &
                                 +              ibit30 * adv_mom_1               &
                                    ) *                                        &
                                     ( w(k,j,i)   + w(k,j-1,i) )              &
                             -      (  8.0_wp * ibit32 * adv_mom_5               &
                                 +              ibit31 * adv_mom_3               &
                                    ) *                                       &
                                     ( w(k,j+1,i) + w(k,j-2,i) )              &
                             +      (           ibit32 * adv_mom_5               &
                                    ) *                                       &
                                     ( w(k,j+2,i) + w(k,j-3,i) )              &
                                               )

             swap_diss_y_local_w(k) = - ABS( v_comp ) * (                     &
                                    ( 10.0_wp * ibit32 * adv_mom_5               &
                                 +     3.0_wp * ibit31 * adv_mom_3               &
                                 +              ibit30 * adv_mom_1               &
                                    ) *                                       &
                                     ( w(k,j,i)   - w(k,j-1,i) )              &
                             -      (  5.0_wp * ibit32 * adv_mom_5               &
                                 +              ibit31 * adv_mom_3               &
                                    ) *                                       &
                                     ( w(k,j+1,i) - w(k,j-2,i) )              &
                             +      (           ibit32 * adv_mom_5               &
                                    ) *                                       &
                                     ( w(k,j+2,i) - w(k,j-3,i) )              &
                                                        )

          ENDDO

          DO  k = nzb_max_l+1, nzt

             v_comp                 = v(k+1,j,i) + v(k,j,i) - gv
             swap_flux_y_local_w(k) = v_comp * (                              &
                           37.0_wp * ( w(k,j,i) + w(k,j-1,i)   )                 &
                         -  8.0_wp * ( w(k,j+1,i) +w(k,j-2,i)  )                 &
                         +           ( w(k,j+2,i) + w(k,j-3,i) ) ) * adv_mom_5
             swap_diss_y_local_w(k) = - ABS( v_comp ) * (                     &
                           10.0_wp * ( w(k,j,i) - w(k,j-1,i)   )                 &
                         -  5.0_wp * ( w(k,j+1,i) - w(k,j-2,i) )                 &
                         +           ( w(k,j+2,i) - w(k,j-3,i) ) ) * adv_mom_5

          ENDDO
#endif

          DO  j = nys, nyn

!
!--          The lower flux has to be calculated explicetely for the tendency
!--          at the first w-level. For topography wall this is done implicitely
!--          by advc_flags_1.
             k      = nzb + 1
             w_comp = w(k,j,i) + w(k-1,j,i)
             flux_d = w_comp       * ( w(k,j,i) + w(k-1,j,i) ) * adv_mom_1
             diss_d = -ABS(w_comp) * ( w(k,j,i) - w(k-1,j,i) ) * adv_mom_1

             DO  k = nzb+1, nzb_max_l

                ibit29 = REAL( IBITS(advc_flags_1(k,j,i),29,1), KIND = wp )
                ibit28 = REAL( IBITS(advc_flags_1(k,j,i),28,1), KIND = wp )
                ibit27 = REAL( IBITS(advc_flags_1(k,j,i),27,1), KIND = wp )

                u_comp = u(k+1,j,i+1) + u(k,j,i+1) - gu
                flux_r = u_comp * (                                          &
                          ( 37.0_wp * ibit29 * adv_mom_5                        &
                       +     7.0_wp * ibit28 * adv_mom_3                        &
                       +              ibit27 * adv_mom_1                        &
                          ) *                                                &
                                 ( w(k,j,i+1) + w(k,j,i)   )                 &
                   -      (  8.0_wp * ibit29 * adv_mom_5                        &
                       +              ibit28 * adv_mom_3                        &
                          ) *                                                &
                                 ( w(k,j,i+2) + w(k,j,i-1) )                 &
                   +      (           ibit29 * adv_mom_5                        &
                          ) *                                                &
                                 ( w(k,j,i+3) + w(k,j,i-2) )                 &
                                     )

                diss_r = - ABS( u_comp ) * (                                 &
                          ( 10.0_wp * ibit29 * adv_mom_5                        &
                       +     3.0_wp * ibit28 * adv_mom_3                        &
                       +              ibit27 * adv_mom_1                        &
                          ) *                                                &
                                 ( w(k,j,i+1) - w(k,j,i)  )                  &
                   -      (  5.0_wp * ibit29 * adv_mom_5                        &
                       +              ibit28 * adv_mom_3                        &
                          ) *                                                &
                                 ( w(k,j,i+2) - w(k,j,i-1) )                 &
                   +      (           ibit29 * adv_mom_5                        &
                          ) *                                                &
                                 ( w(k,j,i+3) - w(k,j,i-2) )                 &
                                              )

#ifdef _OPENACC
!
!--             Recompute the left fluxes.
                ibit29_l = REAL( IBITS(advc_flags_1(k,j,i-1),29,1), KIND = wp )
                ibit28_l = REAL( IBITS(advc_flags_1(k,j,i-1),28,1), KIND = wp )
                ibit27_l = REAL( IBITS(advc_flags_1(k,j,i-1),27,1), KIND = wp )

                u_comp_l = u(k+1,j,i) + u(k,j,i) - gu
                flux_l   = u_comp_l * (                                      &
                                      ( 37.0_wp * ibit29_l * adv_mom_5          &
                                   +     7.0_wp * ibit28_l * adv_mom_3          &
                                   +              ibit27_l * adv_mom_1          &
                                      ) *                                    &
                                     ( w(k,j,i)   + w(k,j,i-1) )             &
                               -      (  8.0_wp * ibit29_l * adv_mom_5          &
                                   +              ibit28_l * adv_mom_3          &
                                      ) *                                    &
                                     ( w(k,j,i+1) + w(k,j,i-2) )             &
                               +      (           ibit29_l * adv_mom_5          &
                                      ) *                                    &
                                     ( w(k,j,i+2) + w(k,j,i-3) )             &
                                                 )

                diss_l   = - ABS( u_comp_l ) * (                             &
                                        ( 10.0_wp * ibit29_l * adv_mom_5        &
                                     +     3.0_wp * ibit28_l * adv_mom_3        &
                                     +              ibit27_l * adv_mom_1        &
                                        ) *                                  &
                                     ( w(k,j,i)   - w(k,j,i-1) )             &
                                 -      (  5.0_wp * ibit29_l * adv_mom_5        &
                                     +              ibit28_l * adv_mom_3        &
                                        ) *                                  &
                                     ( w(k,j,i+1) - w(k,j,i-2) )             &
                                 +      (           ibit29_l * adv_mom_5        &
                                        ) *                                  &
                                     ( w(k,j,i+2) - w(k,j,i-3) )             &
                                                            )
#else
                flux_l = swap_flux_x_local_w(k,j)
                diss_l = swap_diss_x_local_w(k,j)
#endif


                ibit32 = REAL( IBITS(advc_flags_2(k,j,i),0,1),  KIND = wp )
                ibit31 = REAL( IBITS(advc_flags_1(k,j,i),31,1), KIND = wp )
                ibit30 = REAL( IBITS(advc_flags_1(k,j,i),30,1), KIND = wp )

                v_comp = v(k+1,j+1,i) + v(k,j+1,i) - gv
                flux_n = v_comp * (                                          &
                          ( 37.0_wp * ibit32 * adv_mom_5                        &
                       +     7.0_wp * ibit31 * adv_mom_3                        &
                       +              ibit30 * adv_mom_1                        &
                          ) *                                                &
                                 ( w(k,j+1,i) + w(k,j,i)   )                 &
                   -      (  8.0_wp * ibit32 * adv_mom_5                        &
                       +              ibit31 * adv_mom_3                        &
                          ) *                                                 &
                                 ( w(k,j+2,i) + w(k,j-1,i) )                 &
                   +      (           ibit32 * adv_mom_5                        &
                          ) *                                                &
                                 ( w(k,j+3,i) + w(k,j-2,i) )                 &
                                     )

                diss_n = - ABS( v_comp ) * (                                 &
                          ( 10.0_wp * ibit32 * adv_mom_5                        &
                       +     3.0_wp * ibit31 * adv_mom_3                        &
                       +              ibit30 * adv_mom_1                        &
                          ) *                                                &
                                 ( w(k,j+1,i) - w(k,j,i)  )                  &
                   -      (  5.0_wp * ibit32 * adv_mom_5                        &
                       +              ibit31 * adv_mom_3                        &
                          ) *                                                &
                                 ( w(k,j+2,i) - w(k,j-1,i) )                 &
                   +      (           ibit32 * adv_mom_5                        &
                          ) *                                                &
                                 ( w(k,j+3,i) - w(k,j-2,i) )                 &
                                              )

#ifdef _OPENACC
!
!--             Recompute the south fluxes.
                ibit32_s = REAL( IBITS(advc_flags_2(k,j-1,i),0,1),  KIND = wp )
                ibit31_s = REAL( IBITS(advc_flags_1(k,j-1,i),31,1), KIND = wp )
                ibit30_s = REAL( IBITS(advc_flags_1(k,j-1,i),30,1), KIND = wp )

                v_comp_s = v(k+1,j,i) + v(k,j,i) - gv
                flux_s   = v_comp_s * (                                      &
                                    ( 37.0_wp * ibit32_s * adv_mom_5            &
                                 +     7.0_wp * ibit31_s * adv_mom_3            &
                                 +              ibit30_s * adv_mom_1            &
                                    ) *                                      &
                                     ( w(k,j,i)   + w(k,j-1,i) )             &
                             -      (  8.0_wp * ibit32_s * adv_mom_5            &
                                 +              ibit31_s * adv_mom_3            &
                                    ) *                                      &
                                     ( w(k,j+1,i) + w(k,j-2,i) )             &
                             +      (           ibit32_s * adv_mom_5            &
                                    ) *                                      &
                                     ( w(k,j+2,i) + w(k,j-3,i) )             &
                                               )

                diss_s   = - ABS( v_comp_s ) * (                             &
                                    ( 10.0_wp * ibit32_s * adv_mom_5            &
                                 +     3.0_wp * ibit31_s * adv_mom_3            &
                                 +              ibit30_s * adv_mom_1            &
                                    ) *                                      &
                                     ( w(k,j,i)   - w(k,j-1,i) )             &
                             -      (  5.0_wp * ibit32_s * adv_mom_5            &
                                 +              ibit31_s * adv_mom_3            &
                                    ) *                                      &
                                     ( w(k,j+1,i) - w(k,j-2,i) )             &
                             +      (           ibit32_s * adv_mom_5            &
                                    ) *                                      &
                                     ( w(k,j+2,i) - w(k,j-3,i) )             &
                                                        )
#else
                flux_s = swap_flux_y_local_w(k)
                diss_s = swap_diss_y_local_w(k)
#endif

!
!--             k index has to be modified near bottom and top, else array
!--             subscripts will be exceeded.
                ibit35 = REAL( IBITS(advc_flags_2(k,j,i),3,1), KIND = wp )
                ibit34 = REAL( IBITS(advc_flags_2(k,j,i),2,1), KIND = wp )
                ibit33 = REAL( IBITS(advc_flags_2(k,j,i),1,1), KIND = wp )

                k_ppp = k + 3 * ibit35
                k_pp  = k + 2 * ( 1 - ibit33  )
                k_mm  = k - 2 * ibit35

                w_comp = w(k+1,j,i) + w(k,j,i)
                flux_t = w_comp * rho_air(k+1) * (                           &
                          ( 37.0_wp * ibit35 * adv_mom_5                        &
                       +     7.0_wp * ibit34 * adv_mom_3                        &
                       +              ibit33 * adv_mom_1                        &
                          ) *                                                &
                             ( w(k+1,j,i)  + w(k,j,i)     )                  &
                   -      (  8.0_wp * ibit35 * adv_mom_5                        &
                       +              ibit34 * adv_mom_3                        &
                          ) *                                                &
                             ( w(k_pp,j,i)  + w(k-1,j,i)  )                  &
                   +      (           ibit35 * adv_mom_5                        &
                          ) *                                                &
                             ( w(k_ppp,j,i) + w(k_mm,j,i) )                  &
                                       )

                diss_t = - ABS( w_comp ) * rho_air(k+1) * (                  &
                          ( 10.0_wp * ibit35 * adv_mom_5                        &
                       +     3.0_wp * ibit34 * adv_mom_3                        &
                       +              ibit33 * adv_mom_1                        &
                          ) *                                                &
                             ( w(k+1,j,i)   - w(k,j,i)    )                  &
                   -      (  5.0_wp * ibit35 * adv_mom_5                        &
                       +              ibit34 * adv_mom_3                        &
                          ) *                                                &
                             ( w(k_pp,j,i)  - w(k-1,j,i)  )                  &
                   +      (           ibit35 * adv_mom_5                        &
                          ) *                                                &
                             ( w(k_ppp,j,i) - w(k_mm,j,i) )                  &
                                               )
!
!--             Calculate the divergence of the velocity field. A respective
!--             correction is needed to overcome numerical instabilities caused
!--             by a not sufficient reduction of divergences near topography.
                div = ( ( ( u_comp + gu ) * ( ibit27 + ibit28 + ibit29 )      &
                  - ( u(k+1,j,i) + u(k,j,i)   )                               & 
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k,j,i-1),27,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j,i-1),28,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j,i-1),29,1), KIND = wp )     &
                                      )                                       &
                  ) * ddx                                                     &
              +   ( ( v_comp + gv ) * ( ibit30 + ibit31 + ibit32 )            &
                  - ( v(k+1,j,i) + v(k,j,i)   )                               &
                                    * (                                       &
                     REAL( IBITS(advc_flags_1(k,j-1,i),30,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_1(k,j-1,i),31,1), KIND = wp )     &
                   + REAL( IBITS(advc_flags_2(k,j-1,i),0,1),  KIND = wp )     &
                                      )                                       &
                  ) * ddy                                                     &
              +   ( w_comp * rho_air(k+1) * ( ibit33 + ibit34 + ibit35 )      &
                - ( w(k,j,i)   + w(k-1,j,i)   ) * rho_air(k)                  &
                                    * (                                       &
                     REAL( IBITS(advc_flags_2(k-1,j,i),1,1), KIND = wp )      &
                   + REAL( IBITS(advc_flags_2(k-1,j,i),2,1), KIND = wp )      &
                   + REAL( IBITS(advc_flags_2(k-1,j,i),3,1), KIND = wp )      &
                                      )                                       & 
                  ) * drho_air_zw(k) * ddzu(k+1)                              &
                ) * 0.5_wp



                tend(k,j,i) = tend(k,j,i) - (                                 &
                      ( ( flux_r + diss_r )                                   &
                    -   ( flux_l + diss_l ) ) * ddx                           &
                    + ( ( flux_n + diss_n )                                   &
                    -   ( flux_s + diss_s ) ) * ddy                           &
                    + ( ( flux_t + diss_t )                                   &
                    -   ( flux_d + diss_d ) ) * drho_air_zw(k) * ddzu(k+1)    &
                                            )  + div * w(k,j,i)

#ifndef _OPENACC
                swap_flux_x_local_w(k,j) = flux_r
                swap_diss_x_local_w(k,j) = diss_r
                swap_flux_y_local_w(k)   = flux_n
                swap_diss_y_local_w(k)   = diss_n
#endif
                flux_d                   = flux_t
                diss_d                   = diss_t

                !$ACC ATOMIC
                sums_ws2_ws_l(k,tn)  = sums_ws2_ws_l(k,tn)                    &
                      + ( flux_t                                              &
                       * ( w_comp - 2.0_wp * hom(k,1,3,0)                   ) &
                       / ( w_comp + SIGN( 1.0E-20_wp, w_comp )              ) &
                        + diss_t                                              &
                       *   ABS( w_comp - 2.0_wp * hom(k,1,3,0)              ) &
                       / ( ABS( w_comp ) + 1.0E-20_wp                       ) &
                        ) *   weight_substep(intermediate_timestep_count)

             ENDDO

             DO  k = nzb_max_l+1, nzt

                u_comp = u(k+1,j,i+1) + u(k,j,i+1) - gu
                flux_r = u_comp * (                                         &
                      37.0_wp * ( w(k,j,i+1) + w(k,j,i)   )                    &
                    -  8.0_wp * ( w(k,j,i+2) + w(k,j,i-1) )                    &
                    +           ( w(k,j,i+3) + w(k,j,i-2) ) ) * adv_mom_5

                diss_r = - ABS( u_comp ) * (                                &
                      10.0_wp * ( w(k,j,i+1) - w(k,j,i)   )                    &
                    -  5.0_wp * ( w(k,j,i+2) - w(k,j,i-1) )                    &
                    +           ( w(k,j,i+3) - w(k,j,i-2) ) ) * adv_mom_5

#ifdef _OPENACC
!
!--             Recompute the left fluxes.
                u_comp_l = u(k+1,j,i) + u(k,j,i) - gu
                flux_l   = u_comp_l * (                                     &
                            37.0_wp * ( w(k,j,i) + w(k,j,i-1)   )              &
                          -  8.0_wp * ( w(k,j,i+1) + w(k,j,i-2) )              &
                          +           ( w(k,j,i+2) + w(k,j,i-3) ) ) * adv_mom_5
                diss_l   = - ABS( u_comp_l ) * (                            &
                            10.0_wp * ( w(k,j,i) - w(k,j,i-1)   )              &
                          -  5.0_wp * ( w(k,j,i+1) - w(k,j,i-2) )              &
                          +           ( w(k,j,i+2) - w(k,j,i-3) ) ) * adv_mom_5
#else
                flux_l = swap_flux_x_local_w(k,j)
                diss_l = swap_diss_x_local_w(k,j)
#endif

                v_comp = v(k+1,j+1,i) + v(k,j+1,i) - gv
                flux_n = v_comp * (                                         &
                      37.0_wp * ( w(k,j+1,i) + w(k,j,i)   )                    &
                    -  8.0_wp * ( w(k,j+2,i) + w(k,j-1,i) )                    &
                    +           ( w(k,j+3,i) + w(k,j-2,i) ) ) * adv_mom_5

                diss_n = - ABS( v_comp ) * (                                &
                      10.0_wp * ( w(k,j+1,i) - w(k,j,i)   )                    &
                    -  5.0_wp * ( w(k,j+2,i) - w(k,j-1,i) )                    &
                    +           ( w(k,j+3,i) - w(k,j-2,i) ) ) * adv_mom_5

#ifdef _OPENACC
!
!--             Recompute the south fluxes.
                v_comp_s = v(k+1,j,i) + v(k,j,i) - gv
                flux_s   = v_comp_s * (                                     &
                           37.0_wp * ( w(k,j,i) + w(k,j-1,i)   )               &
                         -  8.0_wp * ( w(k,j+1,i) +w(k,j-2,i)  )               &
                         +           ( w(k,j+2,i) + w(k,j-3,i) ) ) * adv_mom_5
                diss_s   = - ABS( v_comp_s ) * (                            &
                           10.0_wp * ( w(k,j,i) - w(k,j-1,i)   )               &
                         -  5.0_wp * ( w(k,j+1,i) - w(k,j-2,i) )               &
                         +           ( w(k,j+2,i) - w(k,j-3,i) ) ) * adv_mom_5
#else
                flux_s = swap_flux_y_local_w(k)
                diss_s = swap_diss_y_local_w(k)
#endif

!
!--             k index has to be modified near bottom and top, else array
!--             subscripts will be exceeded.
                ibit35 = REAL( IBITS(advc_flags_2(k,j,i),3,1), KIND = wp )
                ibit34 = REAL( IBITS(advc_flags_2(k,j,i),2,1), KIND = wp )
                ibit33 = REAL( IBITS(advc_flags_2(k,j,i),1,1), KIND = wp )

                k_ppp = k + 3 * ibit35
                k_pp  = k + 2 * ( 1 - ibit33  )
                k_mm  = k - 2 * ibit35

                w_comp = w(k+1,j,i) + w(k,j,i)
                flux_t = w_comp * rho_air(k+1) * (                           &
                          ( 37.0_wp * ibit35 * adv_mom_5                        &
                       +     7.0_wp * ibit34 * adv_mom_3                        &
                       +              ibit33 * adv_mom_1                        &
                          ) *                                                &
                             ( w(k+1,j,i)  + w(k,j,i)     )                  &
                   -      (  8.0_wp * ibit35 * adv_mom_5                        &
                       +              ibit34 * adv_mom_3                        &
                          ) *                                                &
                             ( w(k_pp,j,i)  + w(k-1,j,i)  )                  &
                   +      (           ibit35 * adv_mom_5                        &
                          ) *                                                &
                             ( w(k_ppp,j,i) + w(k_mm,j,i) )                  &
                                       )

                diss_t = - ABS( w_comp ) * rho_air(k+1) * (                  &
                          ( 10.0_wp * ibit35 * adv_mom_5                        &
                       +     3.0_wp * ibit34 * adv_mom_3                        &
                       +              ibit33 * adv_mom_1                        &
                          ) *                                                &
                             ( w(k+1,j,i)   - w(k,j,i)    )                  &
                   -      (  5.0_wp * ibit35 * adv_mom_5                        &
                       +              ibit34 * adv_mom_3                        &
                          ) *                                                &
                             ( w(k_pp,j,i)  - w(k-1,j,i)  )                  &
                   +      (           ibit35 * adv_mom_5                        &
                          ) *                                                &
                             ( w(k_ppp,j,i) - w(k_mm,j,i) )                  &
                                               )
!
!--             Calculate the divergence of the velocity field. A respective
!--             correction is needed to overcome numerical instabilities caused
!--             by a not sufficient reduction of divergences near topography.
                div = ( ( u_comp + gu - ( u(k+1,j,i) + u(k,j,i)   ) ) * ddx  &
                    +   ( v_comp + gv - ( v(k+1,j,i) + v(k,j,i)   ) ) * ddy  &
                    +   (   w_comp                    * rho_air(k+1) -       &
                          ( w(k,j,i)   + w(k-1,j,i) ) * rho_air(k)           &
                        ) * drho_air_zw(k) * ddzu(k+1)                       &
                      ) * 0.5_wp

                tend(k,j,i) = tend(k,j,i) - (                                 &
                      ( ( flux_r + diss_r )                                   &
                    -   ( flux_l + diss_l ) ) * ddx                           &
                    + ( ( flux_n + diss_n )                                   &
                    -   ( flux_s + diss_s ) ) * ddy                           &
                    + ( ( flux_t + diss_t )                                   &
                    -   ( flux_d + diss_d ) ) * drho_air_zw(k) * ddzu(k+1)    &
                                            )  + div * w(k,j,i)

#ifndef _OPENACC
                swap_flux_x_local_w(k,j) = flux_r
                swap_diss_x_local_w(k,j) = diss_r
                swap_flux_y_local_w(k)   = flux_n
                swap_diss_y_local_w(k)   = diss_n
#endif
                flux_d                   = flux_t
                diss_d                   = diss_t

                !$ACC ATOMIC
                sums_ws2_ws_l(k,tn)  = sums_ws2_ws_l(k,tn)                    &
                      + ( flux_t                                              &
                       * ( w_comp - 2.0_wp * hom(k,1,3,0)                   ) &
                       / ( w_comp + SIGN( 1.0E-20_wp, w_comp )              ) &
                        + diss_t                                              &
                       *   ABS( w_comp - 2.0_wp * hom(k,1,3,0)              ) &
                       / ( ABS( w_comp ) + 1.0E-20_wp                       ) &
                        ) *   weight_substep(intermediate_timestep_count)

             ENDDO
          ENDDO
       ENDDO

    END SUBROUTINE advec_w_ws

 END MODULE advec_ws
